""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import logging
import uuid
from django.db import connection

from rest_framework.request import Request
from .rmq_utils import _publish_rmq_message
from .constants import CREATE_EVENTS, UPDATE_EVENTS, DELETE_EVENTS, EXCHANGE_NAME, ROUTING_KEY_CHANGEEVENT, PB_EVENTS

logger = logging.getLogger(__name__)


def publish_change_event(request, service, operation, entity_type, serialized_data, timestamp,
                         old_serialized_data=None, skip_audit=False, message='', metadata=None):
    """
    A standardized function to publish change events from services to RMQ across XFabric.
    Can process and publish multiple events in one function call.
    :param request: Django request object of the API request being made
    :param service: Service publishing the change event
    :param operation: Type of operation executed as part of the change event (create, update, delete, etc.)
    :param entity_type: Type of entity operated on by the change event
    :param serialized_data: Current serialized data of the object(s)/record(s) as a result of the change event
    :param timestamp: Timestamp of when the change event occurred in the following format: '%Y-%m-%dT%H:%M:%S.%fZ'
    :param old_serialized_data: Old serialized data of the object(s)/record(s) before the change event took place
                                (Optional. This is only used for update-like events where an already existing
                                record is being updated and we can compare past and present records)
    :param skip_audit: Skips audit record from getting created when set to True
    :param message: A message that can override a generic message generated by audit or any other service
    :param metadata: Extra relevant data
    :return: None. Publishes the change message to RMQ
    """
    if not metadata:
        metadata = {}
    websocket_topics = None
    if isinstance(serialized_data, dict):
        # if it's a singular event
        websocket_topics = serialized_data.get('topics')
        serialized_data = [serialized_data]
    if isinstance(old_serialized_data, dict):
        old_serialized_data = [old_serialized_data]
    if not isinstance(request, Request):
        logger.debug(f"Expected request object type :: <class 'rest_framework.request.Request'>. "
                     f'Actual request object type :: {type(request)}. '
                     f'Attempting to publish change event with given object type...')
    try:
        if hasattr(request, 'tenantuuid'):
            tenant_uuid = request.tenantuuid
        else:
            tenant = connection.get_tenant()
            tenant_uuid = str(tenant.uuid)
        # Convert missing or Nonetype teams to empty list
        if not hasattr(request, 'teams') or not request.teams:
            teams = []
        else:
            teams = request.teams
        audit_entry = {
            # tenantuuid required for websocket too
            'tenantuuid': tenant_uuid,
            'tenantid': request.tenantid,
            'service': service,
            'operation': operation,
            # entityType required for websocket too
            'entityType': entity_type,
            'timestamp': timestamp,
            'actor_uuid': str(request.useruuid) if isinstance(request.useruuid, uuid.UUID) else request.useruuid,
            'actor_username': request.username,
            'subtenantid': request.subtenantid,
            'origin': request.META.get('REMOTE_ADDR'),
            # websocketSessionID required for websocket too
            'websocketSessionID': request.headers.get('Websocket-Session-Id'),
            # Add websocket topics at top level, required by livesync service
            'topics': websocket_topics,
            'teams': teams,
            'data': _process_change_event(operation, serialized_data, old_serialized_data),
            'skip_audit': skip_audit,
            'message': message,
            'metadata': metadata
        }
    except AttributeError as e:
        logger.error(f'Request object error while publishing change event {e}')
        return
    if audit_entry:
        try:
            logger.debug(f'Attempting to publish change event for operation {operation} of entityType {entity_type} '
                         f'with the following entry')
            _publish_rmq_message(EXCHANGE_NAME, ROUTING_KEY_CHANGEEVENT, audit_entry)
        except Exception as e:
            logger.error(f'Error while publishing change event for operation {operation} of entityType {entity_type}')
            logger.exception(f'Error details :: {e}')


def _process_change_event(operation, serialized_data, old_serialized_data=None):
    """
    An internal helper function that processes serialized data of change event objects in preparation for RMQ publish
    :param operation: Type of operation executed as part of the change event (create, update, delete, etc.)
    :param serialized_data: Current serialized data of the object/record as a result of the change event
    :param old_serialized_data: Old serialized data of the object/record before the change event took place
                                (Optional. This is only used for update-like events where an already existing
                                record is being updated and we can compare past and present records)
    :return: Data portion of the change event object which holds the UUID of the record, new record/object data,
             and its old data (if applicable)
    :rtype list
    """
    audit_entry_data = []
    for entity in serialized_data:
        # serialized data sometimes doesn't contain a 'uuid' key
        entity_uuid = entity.get('uuid', entity.get('@id', '').split('/')[-1])
        entity.pop('uuid', None)
        new_data = {}
        old_data = {}
        data_to_append = {}
        if operation in CREATE_EVENTS or operation in PB_EVENTS:
            new_data = dict(entity)
        elif operation in UPDATE_EVENTS and old_serialized_data is not None:
            new_data = dict(entity)
            for old_entity in old_serialized_data:
                old_entity_uuid = old_entity.get('uuid', old_entity.get('@id', '').split('/')[-1])
                if old_entity_uuid == entity_uuid:
                    old_entity.pop('uuid', None)
                    old_data = dict(old_entity)
            #         display only the newly updated key, value pairs
            #         changed_data = {key: new_data[key] for key in new_data.keys()
            #                         if new_data[key] != old_data[key]}
        elif operation in DELETE_EVENTS:
            old_data = dict(entity)
        else:
            data_to_append ['data'] = dict(entity)
        if new_data:
            data_to_append['newData'] = new_data
        if old_data:
            data_to_append['oldData'] = old_data
        data_to_append['entityUuid'] = entity_uuid if entity_uuid else None
        audit_entry_data.append(data_to_append)
    return audit_entry_data
