######
# IMP: This file getting imported in gunicorn config file. Do not import any Django
# related file in this.
#####
import yaml
import configparser
from abc import ABC, abstractmethod
import os


def singleton(cls):
    instances = {}

    def _singleton(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]
    return _singleton


class GenericConfigReader(ABC):
    def __init__(self, *args, **kwargs):
        self.config = {}
        self.default_config_dir = kwargs.get('default_config_dir')
        self.config_dir = kwargs.get('config_dir')
        self.CONFIG_FILES_PROPERTIES = kwargs.get('CONFIG_FILES_PROPERTIES', [])
        self.CONFIG_FILES_INI = kwargs.get('CONFIG_FILES_INI', [])
        self.CONFIG_FILES_YAML = kwargs.get('CONFIG_FILES_YAML', [])
        self._load_config()

    @abstractmethod
    def _load_config(self):
        pass

    def update_dict_recursive(self, old_config, new_config):
        """
        Recursively update old_config with values from new_config.
        """
        for key, value in new_config.items():
            if key in old_config and isinstance(old_config[key], dict) and isinstance(value, dict):
                self.update_dict_recursive(old_config[key], value)
            else:
                #
                # This fix is required for ini files. Python configparser for ini provides
                # getboolean method, but since we use common parser, convert those.
                #
                if isinstance(value, str) and value.lower() == 'true':
                    old_config[key] = True
                elif isinstance(value, str) and value.lower() == 'false':
                    old_config[key] = False
                else:
                    old_config[key] = value

    def reload(self):
        self.config = {}
        self._load_config()

    def get(self, section, option=None, default=None):
        if option:
            section_data = self.config.get(section, {})
            return section_data.get(option, default)
        return self.config.get(section, default)
    
    def get_path(self, input_path):
        if self.config_dir:
            # First check if input_path present in config_dir. config_dir is env specific dir. for eg. qa, dev, prod
            abs_path = os.path.join(self.config_dir, input_path)
            if os.path.exists(abs_path):
                return abs_path
        if self.default_config_dir:
            # input_path is not present in config_dir, check it in default_config_dir
            abs_path = os.path.join(self.default_config_dir, input_path)
            if os.path.exists(abs_path):
                return abs_path
        # input_path is not present, return it as is
        return input_path

    def config_parser(self, config_file_fp, separators):
        local_config = {}
        for line in config_file_fp:
            strip_line = line.strip()
            if strip_line.startswith('#'):
                continue
            if '#' in strip_line:
                strip_line = strip_line.split('#', 1)[0].strip()
            for separator in separators:
                if separator in strip_line:
                    key_value = strip_line.split(separator, 1)
                    key = key_value[0].strip()
                    value = key_value[1].strip()
                    local_config[key] = value
                    break
        return local_config

    def reader_properties(self):
        config_files = self.CONFIG_FILES_PROPERTIES
        local_config = {}
        separators = ['=', ':', ' ']
        config_files = config_files if isinstance(config_files, list) else [config_files]
        for config_path in config_files:
            if not os.path.isfile(config_path):
                continue
            with open(config_path, 'r') as config_file_fp:
                self.update_dict_recursive(local_config, self.config_parser(config_file_fp, separators))
                config_file_fp.close()
        return local_config

    def reader_yaml(self):
        config_files = self.CONFIG_FILES_YAML
        local_config = {}
        config_files = config_files if isinstance(config_files, list) else [config_files]
        for config_path in config_files:
            if not os.path.isfile(config_path):
                continue
            with open(config_path, 'r') as config_file_fp:
                self.update_dict_recursive(local_config, yaml.load(config_file_fp, Loader=yaml.FullLoader))
                config_file_fp.close()
        return local_config

    def reader_ini(self):
        config_files = self.CONFIG_FILES_INI
        local_config = {}
        config_files = config_files if isinstance(config_files, list) else [config_files]
        parser = configparser.ConfigParser()
        #
        # https://stackoverflow.com/questions/19359556/configparser-reads-capital-keys-and-make-them-lower-case
        # By default, all keys in sections are accessible in a case-insensitive manner. E.g. for option
        # in parser["section"] yields only optionxform’ed option key names. This means lowercased keys
        # by default.
        parser.optionxform = str
        for config_path in config_files:
            ini_file_config = {}
            if not os.path.isfile(config_path):
                continue
            parser.read(config_path)
            for section in parser.sections():
                ini_file_config[section] = dict(parser.items(section))
            self.update_dict_recursive(local_config, ini_file_config)
        return local_config


@singleton
class ConfigReader(GenericConfigReader):
    def _load_config(self):
        #
        # Order of reading the files is first properties, then ini and then yaml.
        # yaml files are read last since K8s expose configs in yaml format.
        #
        self.update_dict_recursive(self.config, self.reader_properties())
        self.update_dict_recursive(self.config, self.reader_ini())
        self.update_dict_recursive(self.config, self.reader_yaml())
        