""" Copyright start
  Copyright (C) 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import logging
import datetime
import pytz
from abc import ABC, abstractmethod
from decimal import Decimal

from xfutils.exceptions import XFException

from ..constants import IGNORE_FILES


logger = logging.getLogger(__name__)


def calc_progress_tracking_vars(selections):
    file_count = 0
    precision = 1
    if isinstance(selections, int):
        file_count = selections
    else:
        for value in selections.values():
            file_count += len(value)
    if file_count > 100:
        # attempt to keep track of progress more precisely with large number of files
        precision = len(str(file_count))
    chunk_size = round(Decimal(100 / file_count), precision)  # to allow progress tracking

    return file_count, precision, chunk_size


def paths_to_object(paths):
    # Sort so that deepest paths are first
    paths = sorted(paths, key=lambda s: len(s.split('/')), reverse=True)
    tree_path = {}
    for path in paths:
        levels = path.split("/")
        file = levels.pop()
        # skip ignored files/folders or folders
        if levels[0] in IGNORE_FILES or len(file) == 0 or file in IGNORE_FILES:
            continue
        acc = tree_path
        for i, p in enumerate(levels, start=1):
            if i == len(levels):
                # Reached end of path
                # Use current terminal object if present, else use list
                acc[p] = acc[p] if p in acc else []
                if isinstance(acc[p], list):
                    # Only append if we are at a list
                    acc[p].append(path)
            else:
                # Expand with dictionary by default
                acc.setdefault(p, {})
            acc = acc[p]
    return tree_path


def generate_timestamp():
    return datetime.datetime.now(pytz.UTC).strftime("%Y-%m-%d %H:%M:%S %z")

def load_request_attrs(request):
    return {
        'tenantid': request.tenantid,
        'tenantuuid': request.tenantuuid,
        'subtenantid': request.subtenantid,
        'user_uuid': request.useruuid,
        'username': request.username,
        'teams': request.teams
    }


class PortingConfig(ABC):
    def service_request(self, porter, model_name, data, request_type='export', **kwargs):
        model_name = model_name.replace('-', '_')
        # maps to one of the class functions below based on the model name
        return getattr(self, f'{model_name}_{request_type}')(self, porter, data)


class XFSynchronousPorterActions(ABC):
    @abstractmethod
    def build_import_options(*args, **kwargs):
        pass

    @abstractmethod
    def run_import(*args, **kwargs):
        pass

    @abstractmethod
    def run_export(*args, **kwargs):
        pass
