""" Copyright start
  Copyright (C) 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import json
import logging
from io import BytesIO
from zipfile import ZipFile
import binascii

from django.core.files import File
from rest_framework import filters, status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from django.http import FileResponse
from django_filters.rest_framework import DjangoFilterBackend
from xfutils.exceptions import XFException

from .error_message import *
from .models import XFImportJob, XFExportJob
from ..constants import TEMP_LOC, KEY_SERIALIZED_REQUEST_ATTRS
from .tasks import XFPorterTask
from .utils import XFSynchronousPorterActions, load_request_attrs

FILTER_BACKENDS = (filters.OrderingFilter,
                   filters.SearchFilter,
                   DjangoFilterBackend,)

logger = logging.getLogger(__name__)


class XFImportViewSet(viewsets.ModelViewSet):
    """
    Viewset for import jobs
    """
    filter_backends = FILTER_BACKENDS
    porter_tasks_class = XFPorterTask
    import_job_class = XFImportJob

    def get_queryset(self):
        queryset = self.import_job_class.objects.all()
        return queryset

    def create(self, request, *args, **kwargs):
        try:
            request.data['tenant'] = request.tenantuuid
            import_job_serializer = self.serializer_class(data=request.data)
            import_job_serializer.is_valid(raise_exception=True)
            import_job_serializer.save()
            headers = self.get_success_headers(import_job_serializer.data)
            response_dict = import_job_serializer.data
            response_dict.pop('file', None)

            return Response(response_dict, status=status.HTTP_201_CREATED, headers=headers)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_7, e)

    def retrieve(self, request, *args, **kwargs):
        try:
            logger.info('Retrieving import job')
            serializer = self.get_serializer(self.get_object())
            response_dict = serializer.data
            response_dict.pop('file', None)
            return Response(response_dict)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_8, e)

    @action(methods=['PUT'], detail=True, url_path='generate-options')
    def generate_options(self, request, *args, **kwargs):
        try:
            logger.info('Generating import options')
            filename = request.data.get('file')
            try:
                file_data = filename.read()
            except Exception as e:
                logger.error(e)
                raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_16, e)
            try:
                serialized_request_attrs = json.loads(self.request.headers.get(KEY_SERIALIZED_REQUEST_ATTRS))
            except Exception:
                serialized_request_attrs = load_request_attrs(request)
            generate_args = {
                'import_job_uuid': kwargs.get('pk'),
                'auth': self.request.headers['Authorization'],
                'tenant': request.tenantuuid,
                'file_data': binascii.b2a_base64(file_data).decode('utf-8'),
                KEY_SERIALIZED_REQUEST_ATTRS: serialized_request_attrs
            }
            task = self.porter_tasks_class()
            headers = {
                '_tenant_id': serialized_request_attrs.get('tenantid'),
                'exec_env': serialized_request_attrs.get('exec_env')
            }
            task.build_import_options.apply_async(kwargs=generate_args, headers=headers)
            return Response({'message': 'Generating import options...',
                             'import_id': kwargs.get('pk')}, status=status.HTTP_200_OK)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_9, e)

    def update(self, request, *args, **kwargs):
        try:
            instance = self.get_object()
            serializer = self.get_serializer(instance, data=request.data, partial=True)
            serializer.is_valid(raise_exception=True)
            self.perform_update(serializer)
            response_dict = serializer.data
            response_dict.pop('file', None)
            return Response(response_dict, status=status.HTTP_200_OK)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_14, e)

    @action(methods=['PUT'], detail=True)
    def trigger(self, request, *args, **kwargs):
        try:
            logger.info('Triggering import')
            filename = request.data.get('file')
            try:
                file_data = filename.read()
            except Exception as e:
                logger.error(e)
                raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_16, e)
            options_blob = request.data.get('options')
            options = options_blob.read()
            import_job = self.get_object()
            if not options:
                # need import options to be able to trigger
                raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_10)
            try:
                serialized_request_attrs = json.loads(self.request.headers.get(KEY_SERIALIZED_REQUEST_ATTRS))
            except Exception:
                serialized_request_attrs = load_request_attrs(request)
            import_args = {
                'import_job_uuid': str(import_job.uuid),
                'options': json.loads(options),
                'auth': self.request.headers['Authorization'],
                'file_data': binascii.b2a_base64(file_data).decode('utf-8'),
                KEY_SERIALIZED_REQUEST_ATTRS: serialized_request_attrs
            }
            task = self.porter_tasks_class()
            headers = {
                '_tenant_id': serialized_request_attrs.get('tenantid'),
                'exec_env': serialized_request_attrs.get('exec_env')
            }
            task.run_import.apply_async(kwargs=import_args, headers=headers)

            return Response({'import_id': f'{import_job.uuid}'}, status=status.HTTP_200_OK)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_11, e)

    def destroy(self, request, *args, **kwargs):
        import_job = self.get_object()
        self.perform_destroy(import_job)
        return Response(status=status.HTTP_204_NO_CONTENT)


class XFExportViewSet(viewsets.ModelViewSet):
    """
    Viewset for export jobs
    """
    filter_backends = FILTER_BACKENDS
    porter_tasks_class = XFPorterTask
    export_job_class = XFExportJob

    def get_queryset(self):
        queryset = self.export_job_class.objects.all()
        return queryset

    def create(self, request, *args, **kwargs):
        logger.info('Creating export job')
        try:
            request.data['createUser'] = request.useruuid
            request.data['modifyUser'] = request.useruuid
            request.data['tenant'] = request.tenantuuid
            export_job_serializer = self.serializer_class(data=request.data)
            export_job_serializer.is_valid(raise_exception=True)
            export_job_serializer.save()
            headers = self.get_success_headers(export_job_serializer.data)

            return Response(export_job_serializer.data, status=status.HTTP_201_CREATED, headers=headers)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_3, e)
    
    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        serializer = self.get_serializer(queryset, many=True)
        data = serializer.data
        for item in data:
            del item['file']
        return Response(data)

    def retrieve(self, request, *args, **kwargs):
        try:
            logger.info('Retrieving export job')
            serializer = self.get_serializer(self.get_object())
            response_dict = serializer.data
            response_dict.pop('file', None)
            return Response(response_dict)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_2, e)

    @action(methods=['PUT'], detail=True)
    def trigger(self, request, *args, **kwargs):
        logger.info('Triggering export')
        try:
            # TODO: explore if this can be done without creating temp file and using file streams instead
            filename = f'export_{str(kwargs.get("pk"))}.zip'
            filepath = f'{TEMP_LOC}/{filename}'
            ZipFile(filepath, 'w')

            export_template = self.get_object()

            try:
                serialized_request_attrs = json.loads(self.request.headers.get(KEY_SERIALIZED_REQUEST_ATTRS))
            except Exception:
                serialized_request_attrs = load_request_attrs(request)
            export_args = {
                'export_job_uuid': str(kwargs.get('pk')),
                'filepath': filepath,
                'selections': export_template.selections,
                'auth': self.request.headers['Authorization'],
                KEY_SERIALIZED_REQUEST_ATTRS: serialized_request_attrs
            }
            task = self.porter_tasks_class()
            headers = {
                '_tenant_id': serialized_request_attrs.get('tenantid'),
                'exec_env': serialized_request_attrs.get('exec_env')
            }
            task.run_export.apply_async(kwargs=export_args, headers=headers)

            return Response({'export_id': kwargs.get('pk')}, status=status.HTTP_200_OK)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_4, e)

    @action(methods=['GET'], detail=True)
    def download(self, *args, **kwargs):
        logger.info('Downloading export zip')
        export_job = self.get_object()
        if export_job.status == 'Complete':
            file_content = BytesIO(export_job.file)
            response = FileResponse(File(file_content), content_type='application/zip')
            response['Content-Disposition'] = f'attachment: filename=export.zip'
            return response
        else:
            return Response({'message': xf_porter_1[1],
                             'status': export_job.status,
                             'progressPercent': export_job.progressPercent}, status.HTTP_200_OK)

    def destroy(self, request, *args, **kwargs):
        export_job = self.get_object()
        self.perform_destroy(export_job)
        return Response(status=status.HTTP_204_NO_CONTENT)


class XFSynchronousExportViewSet(viewsets.ViewSet):
    
    porter_actions_class = XFSynchronousPorterActions

    @action(methods=['PUT'], detail=True)
    def export(self, request, *args, **kwargs):
        logger.info('Triggering export')
        try:
            filename = f'sync_export_{str(kwargs.get("pk"))}.zip'
            filepath = f'{TEMP_LOC}/{filename}'
            ZipFile(filepath, 'w')
            try:
                serialized_request_attrs = json.loads(self.request.headers.get(KEY_SERIALIZED_REQUEST_ATTRS))
            except Exception:
                serialized_request_attrs = load_request_attrs(request)
            export_args = {
                'filepath': filepath,
                'selections': request.data.get('selections'),
                KEY_SERIALIZED_REQUEST_ATTRS: serialized_request_attrs
            }
            task = self.porter_actions_class()
            results = task.run_export(**export_args)

            return Response(results, status=status.HTTP_200_OK)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_4, e)
        

class XFSynchronousImportViewSet(viewsets.ViewSet):
    porter_actions_class = XFSynchronousPorterActions

    def import_options(self, request, *args, **kwargs):
        try:
            logger.info('Generating import options')
            filename = request.data.get('file')
            try:
                file_data = filename.read()
            except Exception as e:
                logger.error(e)
                raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_16, e)
            generate_args = {
                'tenant': request.tenantuuid,
                'file_data': json.loads(file_data)
            }
            actions = self.porter_actions_class()
            results = actions.build_import_options(**generate_args)
            return Response(results, status=status.HTTP_200_OK)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_9, e)

    def trigger_import(self, request, *args, **kwargs):
        try:
            logger.info('Triggering import')
            file_blob = request.data.get('file')
            options_blob = request.data.get('options')
            file_data = file_blob.read()
            options = options_blob.read()
            if not options:
                # need import options to be able to trigger
                raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_10)
            try:
                serialized_request_attrs = json.loads(self.request.headers.get(KEY_SERIALIZED_REQUEST_ATTRS))
            except Exception:
                serialized_request_attrs = load_request_attrs(request)
            import_args = {
                'options': json.loads(options),
                'file_data': json.loads(file_data),
                KEY_SERIALIZED_REQUEST_ATTRS: serialized_request_attrs
            }
            actions = self.porter_actions_class()
            results = actions.run_import(**import_args)

            return Response(results, status=status.HTTP_200_OK)
        except Exception as e:
            logger.error(e)
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_porter_11, e)
