""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import jwt
import logging
import uuid
from django.utils.decorators import sync_only_middleware
from urllib.parse import parse_qs

from .context import tenantid

logger = logging.getLogger(__name__)

def validate_guid(original_guid):
    #
    # When micro services internally calls metrics service using Service-Token, it don't have tenantuuid.
    # But it doesn't mean we will never pass tenantuuid when calling using Service-Token.
    # For current case, workflow service need to call audit service to start collection of storage size for
    # all tenants, hence no need to pass tenantuuid, as that call is for all tenants.
    #
    if not original_guid:
        logger.error("No input UUID")
        return False
    try:
        return bool(uuid.UUID(original_guid, version=4).hex)
    except ValueError:
        logger.error("Invalid UUID")
        return False

async def asgi_validate_guid(original_guid):
    try:
        return bool(uuid.UUID(original_guid, version=4).hex)
    except ValueError:
        logger.error("Invalid UUID")
        return False

def _get_token_claim(token):
    try:
        return jwt.decode(token, algorithms=['RS256'], options={'verify_signature': False})
    except Exception:
        pass
    return None

async def _asgi_get_token_claim(token):
    try:
        return jwt.decode(token, algorithms=['RS256'], options={'verify_signature': False})
    except Exception:
        pass
    return None

def _process_claim(claim):
    #
    # Controlplane token has user_uuid as the only identifier
    # For platform token, there are user_uuid and tenantuid as two claims
    # tenantuuid is deployment uuid for platform

    tenantid_value = claim.get('user_uuid')
    if claim.get('tenantuuid'):
        #
        # Initially we thought of adding userid@tenantuuid format. It was to know which user
        # facing issue for tenant. Now, lets first see if we really need it or not.
        #
        #tenantid_value = '{0}@{1}'.format(tenantid_value, claim.get('tenantuuid'))
        tenantid_value = claim.get('tenantuuid')
    elif claim.get('tenantid'):
        tenantid_value = claim.get('tenantid')
    if validate_guid(tenantid_value):
        tenantid.set(tenantid_value)

async def _asgi_process_claim(claim):
    #
    # Controlplane token has user_uuid as the only identifier
    # For platform token, there are user_uuid and tenantuid as two claims
    # tenantuuid is deployment uuid for platform

    tenantid_value = claim.get('user_uuid')
    if claim.get('tenantuuid'):
        #
        # Initially we thought of adding userid@tenantuuid format. It was to know which user
        # facing issue for tenant. Now, lets first see if we really need it or not.
        #
        #tenantid_value = '{0}@{1}'.format(tenantid_value, claim.get('tenantuuid'))
        tenantid_value = claim.get('tenantuuid')
    if await asgi_validate_guid(tenantid_value):
        tenantid.set(tenantid_value)

async def asgi_process_token(token):
    tenantid.set(None)
    claim = await _asgi_get_token_claim(token)
    if not claim:
        return
    await _asgi_process_claim(claim)

def process_incoming_request(request):
    tenantid.set(None)
    try:
        token = request.headers.get('Authorization')
        # Token is in form 'Bearer <token>' or 'Service-Token <token>'
        token = token.split(' ')[1]
    except Exception:
        return
    claim = _get_token_claim(token)
    if not claim:
        return
    _process_claim(claim)


def process_outgoing_request(response, request):
    pass

@sync_only_middleware
def tenantid_middleware(get_response):
    def middleware(request):
        process_incoming_request(request=request)
        #
        # Print request method and URL. This is helpful to trace with API is called by which tenant.
        # Since before this, correlation_id and tenantuuid is already set in logger, below info
        # statement will print details including correlation_id and tenantuuid.
        #
        logger.info("%s %s" % (request.method, request.path))
        response = get_response(request)
        process_outgoing_request(response=response, request=request)
        return response
    return middleware

# live-sync is ASGI service. Need ASGI middleware for it.
class ASGITenantIDMiddleware:
    def __init__(self, app):
        # Initialize the middleware with the inner ASGI application
        self.app = app

    async def __call__(self, scope, receive, send):
        if scope['type'] in ('http', 'websocket'):
            try:
                # For live-sync, we get token in query params
                query_string = scope.get("query_string").decode()
                query_params = parse_qs(query_string)
                token = query_params.get('token')[0]
                await asgi_process_token(token)
                request_path = scope.get('path')
                # WS means its WebSocket request
                request_method = scope['type']
                if scope['type'] == 'http':
                    request_method = scope.get('method')
                #
                # Print request method and URL. This is helpful to trace with API is called by which tenant.
                # Since before this, correlation_id and tenantuuid is already set in logger, below info
                # statement will print details including correlation_id and tenantuuid.
                #
                logger.info("%s %s" % (request_method, request_path))
            except Exception:
                pass
        await self.app(scope, receive, send)