import logging
from django.conf import settings
from django.http import HttpResponse, HttpResponseServerError
from django.db import connections
from django.core.cache import cache
from asgiref.sync import iscoroutinefunction, markcoroutinefunction
logger = logging.getLogger(__name__)

#
# Because a lot of Django middleware such as Django’s AuthenticationMiddleware use the database,
# implementing startup, liveness and readiness probes as a simple view wouldn’t work. When your app fails
# to access the database, Django generates an exception and returns a 500 error page long before
# Django executes your view. This doesn’t provide the best developer experience for Kubernetes.
# Hence use this middleware to implement probes.
#
class ProbeMiddleware(object):
    # live-sync is asgi service. Need to add below.
    sync_capable = True
    async_capable = False

    def __init__(self, get_response):
        self.get_response = get_response
        # live-sync is asgi service. Need to add below.
        if iscoroutinefunction(self.get_response):
            markcoroutinefunction(self)

    def __call__(self, request):
        if (not request.method == "GET") or (not request.path.startswith('/probe/')):
            return self.get_response(request)
        ############
        # startup probe
        ########
        # Sometimes, you have to deal with legacy applications that might require an additional startup
        # time on their first initialization.
        # If a container fails a startup probe, Kubernetes will kill the container and restart it.
        # Startup probes are useful for detecting when an application needs more time to start up than expected.
        #
        if request.path == "/probe/startup":
            return self.startup(request)

        ############
        # liveness probe
        ############
        # Kubernetes uses liveness probes to know when to restart a container.
        # Many applications running for long periods of time eventually transition to broken states,
        # and cannot recover except by being restarted. Kubernetes provides liveness probes to detect
        # and remedy such situations. liveness probe failure only affects the failed container, and
        # only the failed container will be restarted.
        #
        elif request.path == "/probe/alive":
            return self.alive(request)

        #####################
        # readiness probe
        #####################
        # Kubernetes uses readiness probes to decide when the container is available for accepting traffic.
        # Sometimes, applications are temporarily unable to serve traffic. For example, an application might
        # need to load large data or configuration files during startup, or depend on external services after
        # startup. In such cases, you don't want to kill the application, but you don't want to send it requests
        # either. Kubernetes provides readiness probes to detect and mitigate these situations. A pod with
        # containers reporting that they are not ready does not receive traffic through Kubernetes Services.
        # For a readiness probe, giving up means not routing traffic to the pod, but the pod is not restarted.
        # Readiness probes runs on the container during its whole lifecycle.
        # Note, any container’s readiness probe failure will prevent traffic to the whole pod temporarily.
        #
        elif request.path == "/probe/ready":
            return self.ready(request)

    def is_database_working(self):
        try:
            #
            # Connect to database
            # Here, specifically connecting only to the 'replica' database. Controlplane service is multi-region
            # and it should serve traffic even if only able to connect to the 'replica' i.e read only database.
            # Controlplane service container don't necessary need to connect to the write database i.e 'default' 
            # to serve traffic.
            # For other service, this may not be true, as those services are not multi-region.
            #
            cursor = connections[settings.REPLICA_DATABASE].cursor()
            return True
        except Exception as e:
            logger.error(str(e))
        return False

    # redis
    def is_cache_working(self):
        try:
            client = cache.client.get_client()
            client.ping()
            return True
        except Exception as e:
            logger.error(str(e))
        return False

    # TODO Below code will work only for live-sync. Need to add is_broker_up in xf-utils directly.
    def is_queue_running(self):
        try:
            from postman.utils.utility import is_broker_up
            if is_broker_up():
                return True
        except Exception as e:
            logger.error(str(e))
        return False

    def startup(self, request):
        return HttpResponse("OK")
    
    def alive(self, request):
        return HttpResponse("OK")

    def ready(self, request):
        return HttpResponse("OK")
        # TODO see below on how to integrate resource checking for readiness probe
        if not self.is_database_working():
            return HttpResponseServerError("Cannot connect to database")
        if not self.is_cache_working():
            return HttpResponseServerError("Cannot connect to cache")
        # LiveSync is not dependent on database or Redis.
        # It is only dependent on RabbitMQ
        if not self.is_queue_running():
            return HttpResponseServerError("Cannot connect to rabbitmq")
        return HttpResponse("OK")