import logging
from django.db import connection
from django_tenants.utils import schema_context
from tenant.models import Tenant
from .metrics_service import notify_metrics

logger = logging.getLogger(__name__)

def _get_tenant_schema_sizes():
    tenant_data = []
    with schema_context('public'):
        with connection.cursor() as cursor:
            cursor.execute("""
                SELECT sum(pg_total_relation_size(t.oid))::bigint AS schema_size,
                       s.nspname AS schema_name
                FROM pg_class AS t
                RIGHT JOIN pg_namespace AS s
                ON t.relnamespace = s.oid
                WHERE NOT s.nspname LIKE ANY (ARRAY['pg\\_catalog','pg\\_toast%','information\\_schema','pg\\_temp%'])
                AND t.relkind = 'r'
                GROUP BY s.nspname;
            """)
            schema_sizes = cursor.fetchall()
            tenants = Tenant.objects.all()
            tenant_info = {tenant.schema_name: tenant for tenant in tenants}
            for schema_size, schema_name in schema_sizes:
                if schema_name in tenant_info:
                    tenant = tenant_info[schema_name]
                    if hasattr(tenant, 'tenantid'):
                        tenantid = tenant.tenantid
                    else:
                        tenantid = tenant.tenant_id
                    tenant_data.append({
                        'schema_name': schema_name,
                        'tenantuuid': str(tenant.uuid),
                        'tenantid': tenantid,
                        'schema_size': int(schema_size)
                    })
    return tenant_data

def monitor_storage(service_name, metrics_host='metrics:443'):
    tenant_data = _get_tenant_schema_sizes()
    metrics = []
    for tenant in tenant_data:
        metrics.append({
            "tenantid": tenant['tenantid'],
            "tenantuuid": tenant['tenantuuid'],
            "service_name": service_name,
            "ttl": 3600, # In secs
            "name": "storage_usage_bytes",
            "unit": "B", # bytes
            "publish": True,
            "value": tenant['schema_size']
        })
    payload = {
        "metrics": metrics,
        "formats": {
            "storage_usage_bytes": [
                "xf_dplane_tenant_storage_usage_bytes{tenantid=\"${tenantid}\", service=\"${service_name}\"} ${value}"
            ]
        }
    }
    notify_metrics(payload, metrics_host)