""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import logging
from functools import reduce
from operator import and_, or_

from django.db.models import Count
from django.db.models import Q
from django.utils.text import slugify

logger = logging.getLogger(__name__)


def get_query(query, queryset):
    if query.get('filters'):
        queryset = query_filters(queryset, query.get('filters'), query.get('logic', 'AND'))

    if query.get('aggregates'):
        queryset = query_aggregates(queryset, query)

    if query.get('sort'):
        queryset = query_sort(queryset, query)

    if query.get('distinct'):
        fields = query.get('distinct').get('field')
        queryset = queryset.order_by(*fields).distinct(*fields)
    
    if query.get('limit'):
        queryset = queryset[:int(query.get('limit'))]

    if query.get('fields'):
        queryset = queryset.only(*query.get('fields'))

    return queryset


def query_filters(queryset, filters, logic):
    nested_condition = []
    _operators = ['icontains', 'lt', 'gt', 'lte', 'gte', 'isnull', 'in', 'iendswith', 'istartswith']

    def _condition_join(condition, logic):
        if logic.lower() == 'or':
            return reduce(or_, condition)
        elif logic.lower() == 'and':
            return reduce(and_, condition)

    def _condition(filters, logic):
        condition = []
        for filter in filters:
            if filter.get('filters'):
                _condition(filter.get('filters'), filter.get('logic'))
            else:
                field = filter.get('field').replace('.', '__')
                operator = filter.get('operator')
                value = filter.get('value')
                # Filter on the tag slug rather than the name to preserve
                if field == 'tags':
                    field = 'tags__slug'
                    value = [slugify(item) for item in value]

                if filter.get('operator') == 'nicontains':
                    operator = operator[1:]
                    condition.append(~Q(**{'{}__{}'.format(field, operator): value}))

                if filter.get('operator') == 'neq':
                    operator = operator[1:]
                    condition.append(~Q(**{'{}'.format(field): value}))

                if filter.get('operator') == 'nin':
                    operator = operator[1:]
                    condition.append(~Q(**{'{}__{}'.format(field, operator): value}))

                if filter.get('operator') in _operators:
                    if filter.get('type') == "array":
                        condition.append(Q(**{'{}__i{}'.format(field, operator): value}))
                    else:
                        condition.append(Q(**{'{}__{}'.format(field, operator): value}))

                if filter.get('operator') == 'eq':
                    condition.append(Q(**{'{}'.format(field): value}))
        if len(condition) != 0:
            condition = _condition_join(condition, logic)
            nested_condition.append(condition)

    _condition(filters, logic)

    if len(nested_condition) > 1:
        nested_condition = _condition_join(nested_condition, logic)
        queryset = queryset.filter(nested_condition).distinct()
    elif len(nested_condition) == 1:
        queryset = queryset.filter(nested_condition[0]).distinct()
    return queryset


def query_aggregates(queryset, query):
    for aggregate in query.get('aggregates', []):
        field = aggregate.get("field")
        if aggregate.get("operator") == 'count':
            queryset = queryset.values(field).annotate(total=Count(field)).order_by('-total')

    return queryset


def query_sort(queryset, query):
    for s in query.get('sort', []):
        if s.get("direction").lower() == 'asc':
            queryset = queryset.order_by(s.get("field"))
        elif s.get("direction").lower() == 'desc':
            queryset = queryset.order_by("-{}".format(s.get("field")))

    return queryset
