import logging
from abc import ABC
from django.conf import settings
from django.core.cache import caches

logger = logging.getLogger(__name__)

#
# Introducing AbstractContext which will be passed to the RateLimiter.
# This is required as RateLimiter is not aware on what all rates for metrics from
# licensing which RateLimiter is monitoring. AbstractContext helps to get rates from License for metrics.
# Also AbstractContext provides cache details to RateLimiter.
#
class AbstractContext(ABC):
    def __init__(self, service_config=None, entitlements_ratemap=None, cache_name='default'):
        self.service_config = service_config
        self.entitlements_ratemap = entitlements_ratemap
        self.cache_client = None
        self.raw_cache_client = None
        if self.is_cache_present(cache_name):
            #
            # This check is required as integration service need redis cache in XF, but not when
            # running as agent.
            #
            self.cache_client = self.get_cache_client(cache_name)
            self.raw_cache_client = self.get_raw_cache_client(cache_name)

    def is_cache_present(self, cache_name):
        try:
            caches[cache_name]
            return True
        except Exception as e:
            logger.warning(f"Cache '{cache_name}' does not exist: {str(e)}")
            return False

    def get_cache_key(self, tenantid, metric_name, is_raw_client=False):
        if is_raw_client:
            cache_key = '{0}:{1}:{2}:{3}'.format(settings.CACHE_CONFIG.get('prefix'),
                                                 settings.CACHE_CONFIG.get('version'),
                                                 tenantid,
                                                 metric_name)
        else:
            # Django's cache API add prefix and version out of the box as per defined in settings.py
            cache_key = '{0}:{1}'.format(tenantid,
                                         metric_name)
        return cache_key

    def get_cache_client(self, cache_name):
        #
        # When you use caches[cache_name], you are interacting with Django's cache API,
        # which includes methods such as add, incr, get, etc.
        #
        client = caches[cache_name]
        return client
    
    #
    # When adding json string, Django's cache API(normal cache_client) adding extra dummy characters.
    # Those extra dummy characters don't allow to change JSON directly in redis when trying via UI tool(redisinsight).
    # Hence adding get_raw_cache_client for json string kind of keys.
    # Since we are using redis, it will return django_redis.client.DefaultClient.
    #
    def get_raw_cache_client(self, cache_name):
        #
        # When you use caches[cache_name].client.get_client(), you get the raw client used by the backend
        # (e.g., a Redis or Memcached client). This client follows its own API,
        # which might be different from Django's cache API.
        #
        raw_client = caches[cache_name].client.get_client()
        return raw_client

class GenericContext(AbstractContext):
    def __init__(self,
                service_config=None,
                entitlements_ratemap=None,
                cache_name='default'):
        super().__init__(service_config, entitlements_ratemap, cache_name)