import logging
import json
from django.conf import settings
from .core import AbstractContext
from xfutils.utils import generate_interservice_jwt_token, api_call

logger = logging.getLogger(__name__)

def _get_license_entitlements(tenantid, context: AbstractContext):
    logger.info(f"Fetching license entitlements for tenant {tenantid}")
    token = generate_interservice_jwt_token(context.service_config, tenantid=tenantid)
    header = {
        'Accept': 'application/json',
        'Content-Type': 'application/json',
    }
    header['Authorization'] = f"Bearer {token}"
    auth_host = context.service_config.get('application', 'AUTH_HOST')
    endpoint = f"https://{auth_host}/api/auth/licenses/"
    response_json = None
    #
    # Keeping only 1 attempt, but in case going ahead if we need retries, increase below.
    #
    attempt = 1
    while attempt > 0:
        try:
            logger.info(f"Calling {endpoint}")
            response_json = api_call(endpoint, method='GET', headers=header, verify=False)
        except Exception as e:
            logger.error(str(e))
        if response_json:
            break
        else:
            attempt -= 1
    if not response_json:
        logger.warning(f"Unable to fetch license entitlements for tenant {tenantid}")
        return None
    license_entitlements = response_json.get('hydra:member')[0]
    if not license_entitlements:
        logger.warning(f"Unable to fetch license entitlements for tenant {tenantid}")
        return None
    return license_entitlements.get('info')

def format_license_entitlements(license_entitlements, context: AbstractContext):
    for entitlement, entitlement_format in context.entitlements_ratemap.items():
        if entitlement in license_entitlements.keys():
            license_entitlements[entitlement] = entitlement_format.format(license_entitlements[entitlement])

def get_license_entitlements(tenantid, context: AbstractContext):
    #
    # When adding json string, Django's cache API(normal cache_client) adding extra dummy characters.
    # Those extra dummy characters don't allow to change JSON directly in redis when trying via UI tool(redisinsight).
    # Hence adding is_raw_client for json string kind of keys.
    # Since we are using redis, it will return django_redis.client.DefaultClient.
    #
    cache_key = context.get_cache_key(tenantid, 'license_entitlements', is_raw_client=True)
    license_entitlements_json_string = context.raw_cache_client.get(cache_key)
    if license_entitlements_json_string:
        # license_entitlements found in cache, return it
        license_entitlements = json.loads(license_entitlements_json_string)
    else:
        license_entitlements = _get_license_entitlements(tenantid, context)
        if not license_entitlements:
            return
        # Set it in cache
        license_entitlements_json_string = json.dumps(license_entitlements)
        context.raw_cache_client.set(cache_key, license_entitlements_json_string, ex=settings.CACHE_CONFIG.get('license_entitlement_cache_ttl'))
    return license_entitlements