import jwt
import logging
import datetime
import importlib
from django.conf import settings
from rest_framework import authentication, status
from rest_framework import permissions
from xfutils.exceptions import XFException

logger = logging.getLogger(__name__)

def _validate_jwt(token, public_key, algorithms=[settings.SERVICE_TOKEN_AUTH.get('ALGORITHM')]):
    try:
        return jwt.decode(token, public_key, algorithms=algorithms)
    except jwt.ExpiredSignatureError as e:
        error = 'Token has expired'
        logger.error("{0}. Recieved error: {1}".format(error, str(e)))
        raise XFException(status.HTTP_401_UNAUTHORIZED, error)
    except jwt.InvalidTokenError as e:
        error = 'Token is invalid'
        logger.error("{0}. Recieved error: {1}".format(error, str(e)))
        raise XFException(status.HTTP_401_UNAUTHORIZED, error)

def _is_service_token(request):
    auth_header = request.META.get('HTTP_AUTHORIZATION')
    auth_header_type = settings.SERVICE_TOKEN_AUTH.get('AUTH_HEADER_TYPE')
    if (not auth_header) or (not auth_header.startswith(auth_header_type + ' ')):
        return False
    return True

def prepare_service_token(payload={}):
    # Define token payload
    iat = datetime.datetime.utcnow()
    payload['iat'] = iat
    payload['exp'] = iat + datetime.timedelta(minutes=1)
    return jwt.encode(payload,
                      settings.SERVICE_TOKEN_AUTH.get('SIGNING_KEY').strip(),
                      algorithm=settings.SERVICE_TOKEN_AUTH.get('ALGORITHM'))

def _execute_hook(hook_path, drf_request, verified_claim):
    *module_path, function_name = hook_path.split('.')
    module_name = '.'.join(module_path)
    module = importlib.import_module(module_name)
    function_to_call = getattr(module, function_name)
    if function_to_call:
        return function_to_call(drf_request, verified_claim)

class XFServiceUser:
    def __init__(self):
        self.is_authenticated = True
    def __str__(self):
        return "XFServiceUser"

class IsServiceTokenAuthentication(permissions.BasePermission):
    def has_permission(self, request, view):
        if not request.user.is_authenticated:
            return False
        return _is_service_token(request)

class ServiceTokenAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        meta = request.META
        auth_header = meta.get('HTTP_AUTHORIZATION')
        # Move to next authentication scheme if 'Service-Token' is missing
        auth_header_type = settings.SERVICE_TOKEN_AUTH.get('AUTH_HEADER_TYPE')
        if (not auth_header) or (not auth_header.startswith(auth_header_type + ' ')):
            logger.error(f"Service token authentication header missing")
            return None
        public_key = settings.SERVICE_TOKEN_AUTH.get('VERIFYING_KEY').strip()
        if not public_key:
            error = "Service private key could not be loaded for validation"
            logger.error(error)
            raise XFException(status.HTTP_401_UNAUTHORIZED, error)
        token = auth_header.split(' ')[1]
        # validate_jwt raises an exception if decoding fails
        verified_claim = _validate_jwt(token, public_key)
        post_auth_hook = settings.SERVICE_TOKEN_AUTH.get('POST_AUTH_HOOK')
        if post_auth_hook:
            result = _execute_hook(post_auth_hook, request, verified_claim)
            if result:
                return result
        return XFServiceUser(), None