import ast
import inspect
import json
import jwt
import logging
import math
from gunicorn.glogging import Logger
import requests
from django_guid import get_guid

logger = logging.getLogger(__name__)

def normal_round(n):
    if n - math.floor(n) < 0.5:
        return math.floor(n)
    return math.ceil(n)


def api_call(url, method='GET', params=None, body='', headers=None, auth=None, verify=True,
             username='', password='', auth_config=None, timeout=None, auth_type='interservice', *args, **kwargs):
    if auth_config:
        auth = (auth_config['username'], auth_config['password'])
    elif username or password:
        auth = (username, password)

    verify = _convert_verify(verify)

    # build **args for requests call
    request_args = {
        'verify': verify,
    }
    if timeout:
        request_args['timeout'] = timeout
    if auth:
        request_args['auth'] = auth
    if params:
        request_args['params'] = params

    if not headers:
        headers = {}
    headers.update({'content-type': 'application/json'})

    if 'Correlation-ID' not in headers.keys():
        headers.update({'Correlation-ID': get_guid()})

    if kwargs.get('appliance_auth', False):
        if auth_type == 'interservice':
            token = generate_interservice_jwt_token(kwargs.get('all_config', {}), tenantid=kwargs.get('tenant', ''))
        elif auth_type == 'agent_appliance':
            token = generate_jwt_token(kwargs.get('all_config', {}).get('xf.instance.applianceUuid'),
                                       kwargs.get('all_config', {}).get('xf.instance.applianceKey'),
                                       tenantid=kwargs.get('tenant_id'), type='appliance')
        else:
            raise ValueError("Keyword argument auth_type must be either 'interservice' or 'agent_appliance' "
                             "when making an authenticated call with api_call()")
        headers.update({"Authorization": "Bearer {0}".format(token)})


    if headers:
        request_args['headers'] = headers

    # get rid of the body on GET/HEAD requests
    bodyless_methods = ['head', 'get']
    if method.lower() not in bodyless_methods:
        request_args['data'] = _convert_body(body)

    response = requests.request(method, url, **request_args)
    return maybe_json_or_raise(response, *args, **kwargs)


def generate_interservice_jwt_token(application_config, **kwargs):
    pvt_key_file = application_config.get('appliance', {}).get('private_key_filename')
    with open(pvt_key_file) as f:
        pvt_key = f.read()
    token = generate_jwt_token(application_config.get('appliance', {}).get('uuid'),
        pvt_key, type='appliance', **kwargs)
    return token

def maybe_json_or_raise(response, *args, **kwargs):
    """
    Helper function for processing request responses

    Returns any json found in the response. Otherwise, it will extract the
    response as text, or, failing that, as bytes.

    :return: the response from the request
    :rtype: dict or str or bytes
    :raises: :class:`requests.HTTPError` if status code was 4xx or 5xx
    """
    if response.ok:
        try:
            logger.info('Processing request responses.')
            return response.json(strict=False)
        except Exception:
            if not kwargs.get('content'):
                return response.text or response.content
            else:
                return response.content or response.text
    else:
        msg = ''
        try:
            msg = response.json()
        except Exception:
            pass
        if not msg:
            msg = response.text

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            # add any response content to the error message
            msg = f'{e} :: {msg}'
            raise requests.exceptions.HTTPError(msg, response=response)
        

def _convert_verify(verify):
    if verify and isinstance(verify, str):
        try:
            verify = ast.literal_eval(verify.title())
        except Exception as e:
            logger.warning(f'Str verification failed. {e}')
    if not isinstance(verify, bool):
        # just default to true
        return True
    return verify


def _convert_body(body):
    if body and isinstance(body, str):
        try:
            logger.info(f'converting body into json {body}')
            body = json.loads(body, strict=False)
        except Exception:
            logger.warning('Json conversion failed.')

    if body and not isinstance(body, str):
        body = json.dumps(body)
    return body


def generate_jwt_token(jti, private_key, **kwargs):
    auth_payload = kwargs
    auth_payload['jti'] = jti

    auth_header = {
        'alg': 'RS256'
    }
    token = jwt.encode(auth_payload, private_key, algorithm='RS256', headers=auth_header)
    try:
        token = token.decode('utf-8')
    except:
        pass
    return token


#
# Gunicorn logger prints in below format
# [2024-04-03 03:58:58 +0000] [14] [DEBUG] POST /api/audit/query/audit-logs/
# Need to update its format to standard logging format, hence use GunicornCustomLogger.
#
class GunicornCustomLogger(Logger):
    def setup(self, cfg):
        super().setup(cfg)
        #
        # [None] [None] is correlation_id and tenantuuid. Since gunicorn is very first thing which does logging,
        # printing it as None. Note that, later, when django middleware runs, it prints correlation_id and tenantuuid.
        #
        formatter = logging.Formatter('[%(asctime)s] [%(process)d] [%(levelname)s] [None] [None] [%(name)s:%(lineno)d] %(message)s', datefmt='%d/%b/%Y %H:%M:%S %Z')
        if self.error_log.handlers:
            self.error_log.handlers[0].setFormatter(formatter)
        if self.access_log.handlers:
            self.access_log.handlers[0].setFormatter(formatter)
