import datetime
import os
import ssl
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django_tenants.utils import tenant_context
from pika import SSLOptions, ConnectionParameters, BlockingConnection
from pika.credentials import ExternalCredentials
from utils.config_parser import all_config
from xfutils.utils import normal_round, GunicornCustomLogger, generate_jwt_token, api_call

from agents.models import Agent
from tenant.models import Tenant


class Command(BaseCommand):
    help = 'Sends an update to XFabric if the database is working and RabbitMQ connectivity is present'

    def handle(self, *args, **options):
        agent_id = all_config.get('xf.instance.agentId')
        tenant_id = os.environ.get('TENANT_ID')
        try:
            self.stdout.write("Checking connectivity to database...")
            tenant = Tenant.objects.first()
            with tenant_context(tenant):
                agent = Agent.objects.first()
            self.stdout.write("Database connectivity confirmed")
        except Exception as e:
            self.stderr.write(self.style.ERROR(f"Could not establish connection with database. Error is as follows:\n{str(e)}"))
            raise CommandError("Error connecting to database")
        try:
            self.stdout.write("Checking connectivity to RabbitMQ cluster...")
            self._check_rmq_connectivity(all_config.get('rmq_public', {}).get('host', ''), agent_id)
            self.stdout.write("RabbitMQ conectivity confirmed")
        except Exception as e:
            self.stderr.write(self.style.ERROR(f"Could not establish connection with RabbitMQ cluster. Error is as follows:\n{repr(e)}"))
            raise CommandError("Error reaching RabbitMQ cluster")
        
        

    def _check_rmq_connectivity(self, rmq_host, vhost):
        credentials = ExternalCredentials()
        heartbeat=120
        retry_delay=5
        connection_attempts=2
        chainfile = '/opt/xf/configs/ca_chain_bundle.pem'
        context = ssl.create_default_context(cafile=chainfile)
        clientcert = '/opt/xf/configs/client_cert.pem'
        clientkey = '/opt/xf/configs/client_key.pem'
        context.load_cert_chain(clientcert, clientkey)
        ssl_options = SSLOptions(context, server_hostname=rmq_host)

        parameters=ConnectionParameters(rmq_host, 5671, vhost, credentials, heartbeat=heartbeat,  retry_delay=retry_delay, connection_attempts=connection_attempts, ssl_options=ssl_options)
        connection = BlockingConnection(parameters)
        return True
        