import json
from time import sleep

from django.conf import settings
from django.core.management.base import BaseCommand
from django_tenants.utils import schema_context

from agents.models import Agent

from agents.utils.connector_utils import is_connector_installable_on_agent


class Command(BaseCommand):
    help = 'Creates a Tenant in the database'

    def add_arguments(self, parser):
        schema_name = parser.add_argument('--schema_name', required=True)
        filepath = parser.add_argument('--filepath', required=True)
        max_retries = parser.add_argument('--max_retries', type=int, default=10)


    def handle(self, *args, **options):
        try:
            with schema_context(options.get('schema_name')):
                if not self._agent_exists(max_retries=options.get('max_retries')):
                    return self.stdout.write(self.style.ERROR('Error: Cannot install Connector(s) because Agent object does not exist'))
                self.install_selected_agent_connectors(filepath=options.get('filepath'))
        except Exception as e:
            return self.stdout.write(self.style.ERROR('Error: Failed to install Connectors: %s' % str(e)))
        finally:
            pass
        return self.stdout.write(self.style.SUCCESS('Successfully installed Connectors'))


    def install_selected_agent_connectors(self, filepath='/opt/integration/connectors/agent_connectors.json'):
        try:
            with open(filepath, 'rb') as conn_file:
                connectors_json = json.loads(conn_file.read())
            if connectors_json:
                self.stdout.write(f"Installing the following {len(connectors_json)} Connectors: {[conn.get('name') for conn in connectors_json]}")
            for connector in connectors_json:
                installable, message = is_connector_installable_on_agent(connector.get('name'), connector.get('version'), settings.SELF_ID)
                if not installable:
                    self.stdout.write(message)
                    continue
                self.stdout.write(f"Installing connector {connector.get('name')}...")
                data_dict = {
                    'name': connector.get('name'),
                    'version': connector.get('version'),
                    'agent': settings.SELF_ID,
                    'replace': True,
                    'request_data': {
                        'tenantid': settings.TENANT_ID,
                        'username': settings.SELF_ID,
                        'useruuid': settings.ALL_CONFIG.get('xf.instance.applianceUuid'),
                        #TODO: proper logic for subtenant mapping
                        'subtenantid': settings.TENANT_ID,
                        #TODO: proper logic for teams mapping?
                        'teams': [],
                        #if the audit message has no origin element audit servuce throws error
                        'META': {"REMOTE_ADDR": f"agent_{settings.SELF_ID}"}
                    },
                    'trust_api': False, # remove here and in install_connector()
                    'rbac_info': None # remove here and in install_connector()
                }
                from connectors.connector_self_operations import SelfOperations
                ops_obj = SelfOperations()
                result = ops_obj.install_connector(data_dict)
                self.stdout.write(f"Finished installing Connector {connector.get('name')}")
            # If all connectors installed successfully, delete the file so that future startups will not need to try and install them
            #self.stdout.write("Removing Agent Connectors file")
            #os.remove("/opt/integration/connectors/agent_connectors.json")
            self.stdout.write("Finished installing Connectors")
        except FileNotFoundError as e:
            self.stdout.write('Returning after FileNotFound exception')
            return

    def _agent_exists(self, iteration=1, max_retries=10):
        exists = Agent.objects.filter(agentId=settings.SELF_ID).exists()
        if exists:
            return True
        elif iteration < max_retries:
            sleep(1)
            return self._agent_exists(iteration=iteration+1, max_retries=max_retries)
        else:
            return False