import json
import logging
import os
import shutil

import yaml
from django.conf import settings
from django.db import connection
from django.utils.text import slugify

from agents.constants import *
from agents.exceptions import TarOperationFailed
from agents.models import Agent
from agents.utils.cert_utils import generate_keys_for_agent, get_pem_text
from connectors.models import Connector
from connectors.serializers import ConnectorConfigurationAgentSerializer

logger = logging.getLogger(__name__)


def is_connector_added(name, connectors):
    for connector in connectors:
        if name == connector.get('name'):
            return True
    return False


def copy_agent_directory(agent_id):
    path = f'{AGENT_INSTALLER_DESTINATION_PATH}/{agent_id}'
    if os.path.exists(path):
        shutil.rmtree(path)
    shutil.copytree(AGENT_INSTALLER_SOURCE_PATH, path)
    return path


def prepare_connectors_json(path, agent_obj, connectors, include_last_known_configurations):
    """
    Compile the list of Connectors to include with the Agent installer

    The API request to download the installer will include a list
    of connectors to install on the Agent. In addition to these,
    there are certain mandatory connectors, such as the Utilities
    connector, which are mandatory. Additionally the user can elect
    to include any connectors already installed on the Agent as well

    Arguments:
    path -- Filepath of the connectors.json file which will be bundled
    with the installer
    agent_obj -- an Object representing the Agent whose installer is being
    generated
    connectors -- The list of Connectors to be included with the installer
    include_last_known_configurations -- Boolean flag for whether Connectors
    already installed on the Agent should be included
    """
    # Add Connectors already installed on the Agent to the list, if the
    # user chose to do so
    if include_last_known_configurations:
        connectors_on_agent = Connector.objects.filter(agent_id=agent_obj.agentId)

        for connector in connectors_on_agent:
            config_serializer = ConnectorConfigurationAgentSerializer(connector.configuration.all(),
                                                                      many=True)
            configurations = config_serializer.data
            connector_data = {
                'name': connector.name,
                'version': connector.version,
                'rpm_full_name': connector.rpm_full_name,
                'configurations': configurations,
                'rpm_installed': connector.metadata.get('rpm_installed', False)
            }
            # Currently the "include last known configurations" option is
            # not compatible with providing a list of other connectors to
            # include. If that were to change in the future, instead of 
            # appending here, we would need to loop through the connectors
            # list and if the same connector appears there, replace it with
            # connector_data
            connectors.append(connector_data)
    mandatory_connectors = ['cyops_utilities']
    mandatory_connectors_qs = Connector.objects.filter(name__in=mandatory_connectors,
                                                       agent_id=connection.schema_name)
    for mandatory_connector in mandatory_connectors_qs:
        if not is_connector_added(mandatory_connector.name, connectors):
            connectors.append({
                'name': mandatory_connector.name,
                'version': mandatory_connector.version,
                'rpm_full_name': mandatory_connector.rpm_full_name,
                'rpm_installed': mandatory_connector.metadata.get('rpm_installed', False)
            })
    with open(os.path.join(path, 'connectors.json'), 'w') as f:
        json.dump(connectors, f)


def replace_common_values_in_installer(filepath, agent_obj, tenantid, hostname):
    all_config = settings.ALL_CONFIG
    with open(filepath) as f:
        fcontents = f.read()

    from xfutils.passwordModule import manage_password
    db_pw = all_config.get('database', {}).get('db_password', '')
    db_pw = manage_password(db_pw, 'encrypt')
    fcontents = fcontents.replace('@encrypted_db_password@', db_pw)

    fcontents = fcontents.replace('@vhost@', agent_obj.agentId)
    fcontents = fcontents.replace('@user@', agent_obj.agentId)
    fcontents = fcontents.replace('@agentName@', slugify(agent_obj.name))
    fcontents = fcontents.replace('@agentId@', agent_obj.agentId)
    fcontents = fcontents.replace('@tenantid@', tenantid)
    fcontents = fcontents.replace('@rmqPublicHost@', all_config.get('rmq_public', {}).get('host'))
    fcontents = fcontents.replace('@rmqPublicSni@', all_config.get('rmq_public', {}).get('sni'))
    fcontents = fcontents.replace('@schema@', 'schema' + tenantid.translate({ord(c): "" for c in "!@#$%^&*()[]{};:,./<>?\|`~-=_+"}))
    fcontents = fcontents.replace('@hostname@', hostname)
    fcontents = fcontents.replace('@applianceUuid@', str(agent_obj.applianceUuid))
    #TODO: get XF version and build number from helmcharts
    agent_version = settings.RELEASE_VERSION
    build_number = settings.BUILD_NUMBER
    fcontents = fcontents.replace('@agentVersion@', agent_version)
    fcontents = fcontents.replace('@agentBuildNumber@', build_number)
    return fcontents


def populate_docker_agent_script(agent_obj, tenantid, hostname, connectors=list):
    src_filepath = f"{AGENT_INSTALLER_ARCHIEVED_PATH}/source/deploy_docker_agent.sh"
    dest_filepath = f"{AGENT_INSTALLER_ARCHIEVED_PATH}/destination/deploy-docker-agent-{agent_obj.agentId}.sh"
    yaml_content = populate_docker_compose_file(agent_obj, tenantid, hostname, connectors=connectors, write=False)
    yaml_content = yaml_content.replace('"', '\\"')
    with open(src_filepath) as f:
        shcontents = f.read()
    shcontents = shcontents.replace('@dockercompose@', yaml_content)
    shcontents = shcontents.replace('@agentId@', agent_obj.agentId)
    shcontents = shcontents.replace('@agentName@', slugify(agent_obj.name))
    shcontents = shcontents.replace('@agentVersion@', settings.RELEASE_VERSION)
    shcontents = shcontents.replace('@agentBuildNumber@', settings.BUILD_NUMBER)
    with open(dest_filepath, 'w') as f:
        f.write(shcontents)
    return dest_filepath


def populate_docker_compose_file(agent_obj, tenantid, hostname, connectors=list, write=True):
    all_config = settings.ALL_CONFIG
    src_filepath = f"{AGENT_INSTALLER_ARCHIEVED_PATH}/docker_template/docker-compose-agent-source.yaml"
    dest_filepath = f"{AGENT_INSTALLER_ARCHIEVED_PATH}/destination/docker-compose-agent-{agent_obj.agentId}.yaml"

    dockerfile = replace_common_values_in_installer(src_filepath, agent_obj, tenantid, hostname)

    # TODO: Might need changes here to account for dev vs prod vs demo etc environments
    extra_hosts = all_config.get('xf.agent.extraHosts', None)
    if extra_hosts:
        extra_hosts_string = "extra_hosts:\n"
        for host in extra_hosts:
            extra_hosts_string  = f"{extra_hosts_string}      - {host}\n"
        dockerfile = dockerfile.replace('@extrahosts@', extra_hosts_string)
    else:
        dockerfile = dockerfile.replace('@extrahosts@', '')
    appliance_key = agent_obj.applianceKey.replace('\n', '\n        ')
    dockerfile = dockerfile.replace('@applianceKey@', appliance_key)

    if connectors:
        connectors_config_definition_string = f"connectors_config:\n    content: |\n      {json.dumps(connectors)}"
        dockerfile = dockerfile.replace('@connectorsConfigDefinition@', connectors_config_definition_string)
        connectors_config_mapping_string = f"- source: connectors_config\n        target: /opt/integration/connectors/agent_connectors.json"
        dockerfile = dockerfile.replace('@connectorsConfigMapping@', connectors_config_mapping_string)
    else:
        dockerfile = dockerfile.replace('@connectorsConfigDefinition@', '')
        dockerfile = dockerfile.replace('      @connectorsConfigMapping@\n', '')
    dockerfile = add_certs_to_compose_string(dockerfile, agent_obj)

    if write:
        with open(dest_filepath, 'w') as f:
            f.write(dockerfile)
        return dest_filepath
    else:
        return dockerfile

def add_certs_to_compose_string(compose_content, agent_obj):
    all_config = settings.ALL_CONFIG
    dockerfile_with_certs = '\n'.join(
        [compose_content] + \
        get_compose_formatted_cert(all_config.get('rmq_public', {}).get('caCertPath'), 'ca_cert') + \
        get_compose_formatted_cert(all_config.get('rmq_public', {}).get('caChainBundle'), 'ca_chain_bundle') + \
        get_compose_formatted_cert(agent_obj.clientCertificate, 'client_cert', isFile=False) + \
        get_compose_formatted_cert(agent_obj.clientKey, 'client_key', isFile=False)
    )
    return dockerfile_with_certs

def get_compose_formatted_cert(cert, cert_title, isFile=True):
    cert_strings = [f'  {cert_title}:', '    content: |']
    if isFile:
        with open(cert) as f:
            cert_strings += string_to_compose_config_format(f)  
    else:
        cert_strings += string_to_compose_config_format(cert.split('\n'))    
    return cert_strings


def string_to_compose_config_format(s):
    lines = []
    for line in s:
        if line == "\n" or not line.strip():
            continue
        lines.append(f'      {line.strip()}')
    return lines


def populate_installer_sh_file(agent_obj, tenantid, hostname, connectors=list):
    all_config = settings.ALL_CONFIG
    src_filepath = f"{AGENT_INSTALLER_ARCHIEVED_PATH}/source/agent_install.sh"
    dest_filepath = f"{AGENT_INSTALLER_ARCHIEVED_PATH}/destination/agent_install-{agent_obj.agentId}.sh"

    sh_content = replace_common_values_in_installer(src_filepath, agent_obj, tenantid, hostname)
    appliance_key = agent_obj.applianceKey.replace('\n', '\n  ')
    sh_content = sh_content.replace('@applianceKey@', appliance_key)
    with open(all_config.get('rmq_public', {}).get('caCertPath')) as cf:
        sh_content = sh_content.replace('@caCert@', cf.read())
    with open(all_config.get('rmq_public', {}).get('caChainBundle')) as chain_bundle_file:
        sh_content = sh_content.replace('@caChainBundle@', chain_bundle_file.read())
    sh_content = sh_content.replace('@clientCert@', agent_obj.clientCertificate)
    sh_content = sh_content.replace('@clientKey@', agent_obj.clientKey)
    sh_content = sh_content.replace('@connectorsJson@', json.dumps(connectors))
    extra_hosts = all_config.get('xf.agent.extraHosts', None)
    if extra_hosts:
        for host in extra_hosts:
            splithost=host.split(':')
            host = f"{splithost[1]} {splithost[0]}"
        extra_hosts_string = '\n'.join(extra_hosts)
        sh_content = sh_content.replace('@extraHosts@', extra_hosts_string, 1)
    with open(dest_filepath, 'w') as f:
        f.write(sh_content)
    return dest_filepath

def yaml_read_and_replace(path, agent_obj):
    all_config = settings.ALL_CONFIG
    rmq_host = all_config.get('rmq_public', {}).get('host')
    rmq_port = all_config.get('rmq_public', {}).get('port')
    version = all_config.get('xf.agent.version')
    with open(f'{path}/agent_config.yml') as f:
        all_config = settings.ALL_CONFIG
        doc = yaml.safe_load(f)

        doc['rmq_public']['host'] = rmq_host
        doc['rmq_public']['port'] = rmq_port
        doc['rmq_public']['vhost'] = agent_obj.agentId
        doc['rmq_public']['authType'] = agent_obj.authType
        # doc['rmq_public']['user'] = agent_obj.
        doc['rmq_public']['agentId'] = agent_obj.agentId
        doc['rmq_public']['masterId'] = Agent.objects.get(isLocal=True).agentId
        doc['postgres.pg_password'] = agent_obj.agentId

    with open(f'{path}/agent_config.yml', 'w') as f:
        yaml.dump(doc, f)

    with open(f'{path}/setup.sh') as f:
        data = f.read()
    data = data.replace('@product_version@', str(version))
    data = data.replace('@pg_password@', agent_obj.agentId)
    data = data.replace('@agent_id@', agent_obj.agentId)
    data = data.replace('@secure_message_exchange_hostname@', rmq_host)

    with open(f'{path}/agent_config.yml') as f:
        data = data.replace('@agent_config@', f.read())

    with open(path + '/pg_hba.conf') as f:
        pg_hba = f.read()
    data = data.replace('@pg_hba@', pg_hba)

    with open(path + '/connectors.json') as f:
        connectors = f.read()
    data = data.replace('@connectors@', connectors)

    with open(path + '/ca_cert.pem') as f:
        ca_cert = f.read()
    data = data.replace('@ca_cert@', ca_cert)

    # TODO: throw error if cert or key is empty
    data = data.replace('@client_cert@', agent_obj.clientCertificate or "")
    data = data.replace('@client_key@', agent_obj.clientKey or "")

    with open(path + '/setup.sh', 'w') as file:
        file.write(data)


def get_cert_from_file(filepath):
    with open(filepath) as f:
        return f.read()