from cryptography import x509
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import hashes, serialization
from datetime import datetime, timedelta
import logging

from constants.error_message import *
from agents.exceptions import *

from django.conf import settings

logger = logging.getLogger('__name__')


def generate_keys_for_agent(agent_id):
    try:
        private_key = rsa.generate_private_key(
            key_size=2048,
            public_exponent=65537
        )

        csr = get_client_cert_signing_request(agent_id, private_key)
        ca_cert, ca_key = get_ca_data()

        client_cert = get_client_certificate(ca_cert, ca_key, csr)
        return client_cert, private_key
    except Exception as e:
        message = xf_integration_35.format(agent_id)
        message = f'{message} ERROR :: {str(e)}'
        raise CertificateGenerationFailed(message)


def get_ca_data():
    all_config = settings.ALL_CONFIG
    ca_cert_path = all_config.get('rmq_public', {}).get('caCertPath')
    with open(ca_cert_path, 'rb') as f:
        ca_cert = x509.load_pem_x509_certificate(f.read())
    ca_key_path = all_config.get('rmq_public', {}).get('caPrivateKey')
    with open(ca_key_path, 'rb') as f:
        ca_key = serialization.load_pem_private_key(f.read(), password=None)
    return ca_cert, ca_key


def get_pem_text(pem, pem_type='public'):
    if pem_type == 'public':
        return pem.public_bytes(serialization.Encoding.PEM).decode('utf-8')
    else:
        return pem.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        ).decode('utf-8')


def get_client_cert_signing_request(common_name, private_key):
    csr = x509.CertificateSigningRequestBuilder().subject_name(x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'client')
    ])).add_extension(
        x509.BasicConstraints(ca=False, path_length=None),
        critical=False
    ).add_extension(
        x509.KeyUsage(
            digital_signature=True,
            crl_sign=True,
            content_commitment=False,
            key_encipherment=False,
            data_encipherment=False,
            key_agreement=False,
            key_cert_sign=False,
            encipher_only=False,
            decipher_only=False
        ),
        critical=False
    ).add_extension(
        x509.ExtendedKeyUsage([ExtendedKeyUsageOID.CLIENT_AUTH]),
        critical=False
    ).sign(private_key, hashes.SHA256())

    return csr


def get_client_certificate(ca_cert, ca_key, csr):
    client_cert = x509.CertificateBuilder().subject_name(
        csr.subject
    ).issuer_name(
        ca_cert.subject
    ).public_key(
        csr.public_key()
    ).serial_number(
        x509.random_serial_number()
    ).not_valid_before(
        datetime.utcnow()
    ).not_valid_after(
        datetime.utcnow() + timedelta(days=365)
    ).sign(ca_key, hashes.SHA256())

    return client_cert
