# Copyright start
# Copyright (C) 2008 - 2023 Fortinet Inc.
# All rights reserved.
# FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
# Copyright end
import logging

import requests
from django.conf import settings
from requests.auth import HTTPBasicAuth
from xfutils.utils import maybe_json_or_raise

from connectors.environment import expand

all_config = settings.ALL_CONFIG
logger = logging.getLogger('__name__')


class RMQ_Management():
    def __init__(self, agent_id, operation='spinup'):
        self.host = all_config.get('rmq_public', {}).get('host')
        self.port = all_config.get('rmq_public', {}).get('mgtport')
        self.base_url = f'http://{self.host}:{self.port}/api'
        self.auth = HTTPBasicAuth(all_config.get('rmq_public', {}).get('user'), all_config.get('rmq_public', {}).get('password'))
        self.headers = {"content-type": "application/json"}
        self.agent_id = agent_id
        if operation == 'spinup':
            self.agent_spinup()
        elif operation == 'spindown':
            self.agent_spindown()
        elif operation == 'connections':
            self.connections = self.get_connections()

    def create_rmq_vhost_for_agent(self):
        vhost_url = f'{self.base_url}/vhosts/{self.agent_id}'
        vhost_resp = requests.put(vhost_url, auth=self.auth, headers=self.headers)
        vhost_resp = maybe_json_or_raise(vhost_resp)
        return vhost_resp

    def delete_rmq_vhost(self,):
        vhost_url = f'{self.base_url}/vhosts/{self.agent_id}'
        try:
            vhost_resp = requests.delete(vhost_url, auth=self.auth, headers=self.headers)
            vhost_resp = maybe_json_or_raise(vhost_resp)
            return vhost_resp
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return
            else:
                raise e

    def create_exchange_for_agent(self, exchange_name='xf.agent'):
        exchanges = all_config.get('rabbitmq_cyops').get('exchange', [])
        for exchange in exchanges:
            name = expand({'self_id': self.agent_id}, exchange['exchangeName'])
            req_body = exchange.get('exchangeOptions', {})
            req_body['type'] = exchange.get('exchangeType', 'topic')

            url = f'{self.base_url}/exchanges/{self.agent_id}/{name}'
            response = requests.put(url, auth=self.auth, headers=self.headers, json=req_body)
            response = maybe_json_or_raise(response)

        return response

    def create_rmq_user_for_agent(self):
        url = f'{self.base_url}/users/{self.agent_id}'
        body = {
            "tags": "agent_user"
        }
        response = requests.put(url, json=body, auth=self.auth, headers=self.headers)
        response = maybe_json_or_raise(response)
        return response

    def delete_rmq_user(self):
        url = f'{self.base_url}/users/{self.agent_id}'
        try:
            response = requests.delete(url, auth=self.auth, headers=self.headers)
            response = maybe_json_or_raise(response)
            return response
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return
            else:
                raise e

    def add_vhost_permissions_for_user(self, configure=".*", read=".*", write=".*"):
        url = f'{self.base_url}/permissions/{self.agent_id}/{self.agent_id}'
        body = {
            "configure": configure,
            "read": read,
            "write": write
        }
        response = requests.put(url, json=body, auth=self.auth, headers=self.headers)
        response = maybe_json_or_raise(response)
        return response

    def agent_spinup(self):
        self.create_rmq_vhost_for_agent()
        self.create_exchange_for_agent()
        self.create_rmq_user_for_agent()
        self.add_vhost_permissions_for_user()
    
    def agent_spindown(self):
        self.delete_rmq_user()
        self.delete_rmq_vhost()

    def get_connections(self):
        url = f'{self.base_url}/vhosts/{self.agent_id}/connections'
        response = requests.get(url, auth=self.auth, headers=self.headers)
        response = maybe_json_or_raise(response)
        return response
    
    def get_exchanges(self):
        url = f'{self.base_url}/exchanges/{self.agent_id}'
        response = requests.get(url, auth=self.auth, headers=self.headers)
        response = maybe_json_or_raise(response)
        return response

    def get_queues(self):
        url = f'{self.base_url}/queues/{self.agent_id}'
        response = requests.get(url, auth=self.auth, headers=self.headers)
        response = maybe_json_or_raise(response)
        return response