import datetime
import logging
import shutil
import uuid
import os.path

from django.conf import settings
from django.db import connection, IntegrityError
from django.http import HttpResponse
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.views import APIView
from xfutils.audit import publish_change_event

from agents.constants import REMOTE_AGENT_OPERATIONS
from agents.exceptions import *
from agents.filtersets import AgentFilterSet
from agents.models import Agent
from agents.serializers import AgentSerializer, AgentReadOnlySerializer
from agents.utils.agent_installer_utils import prepare_connectors_json, yaml_read_and_replace, copy_agent_directory, \
     populate_docker_compose_file, populate_installer_sh_file, populate_docker_agent_script
from agents.utils.agent_count import has_room_for_agent
from agents.utils.appliance_utils import create_appliance_for_agent
from agents.utils.cert_utils import generate_keys_for_agent, get_pem_text
from agents.utils.connector_utils import create_connector_object
from agents.utils.rmq_mgmt import RMQ_Management
from connectors.models import Connector
from connectors.utils import broadcast_to_remote, extract_audit_data_from_request
from utils.connector_op_utility import get_executable_class


logger = logging.getLogger(__name__)
ENTITY_TYPES = {'agents': 'agents'}


# Create your views here.
class AgentViewSet(viewsets.ModelViewSet):
    queryset = Agent.objects.filter(isLocal=False).order_by('-createDate')
    serializer_class = AgentReadOnlySerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    http_method_names = ['get', 'post', 'put', 'delete']
    ordering_fields = ('name', 'healthStatus', 'enabled', 'createDate')
    filter_class = AgentFilterSet

    # Instead of transaction.atomic, set config to in-progress and then do the rmq status
    # if RMQ setup fails, set health status to failed
    def create(self, request, *args, **kwargs):
        self.serializer_class = AgentSerializer
        try:
            input_data = request.data
            if len(input_data['name']) > 100:
                msg = "Agent name should have a maximum of 100 characters"
                logger.error(msg)
                raise AgentCreationFailed(msg)
            if input_data.get('uuid'):
                input_data['uuid'] = uuid.UUID(input_data.get('uuid'))
            #@TODO: count will fail if we delete add agent
            try:
                has_room_for_agent()
            except AgentLimitReached as e:
                logger.error(e)
                raise e
            agent_id = str(uuid.uuid4())
            input_data['agentId'] = agent_id
            input_data['isLocal'] = False
            input_data['authType'] = 'certificate'
            # TODO: default to admin team once there is a good way
            # to differentiate between Teams objects
            input_data['owners'] = request.teams
            # Generate client certs and keys
            client_cert, client_key = generate_keys_for_agent(agent_id)
            input_data['clientCertificate'] = get_pem_text(client_cert)
            input_data['clientKey'] = get_pem_text(client_key, pem_type='private')
            input_data['healthStatus'] = {
                'status': 'awaiting',
                'message': 'Ready for Agent deployment'
            }
            # Create XF Appliance user for agent to authenticate as
            appliance = create_appliance_for_agent(agent_id, input_data['name'], request.tenantid, request.tenantuuid)
            input_data['applianceKey'] = appliance.get('appliance', {}).get('private_key')
            input_data['applianceUuid'] = appliance.get('appliance', {}).get('uuid')
            input_data['version'] = settings.RELEASE_VERSION
            agent_serializer = self.get_serializer(data=input_data)
            agent_serializer.is_valid(raise_exception=True)

            rmq_manager = RMQ_Management(agent_id)

            agent_obj = agent_serializer.save()
            headers = self.get_success_headers(agent_serializer.data)

            publish_change_event(request, 'integrations', 'create', ENTITY_TYPES['agents'],
                                 agent_serializer.data, agent_serializer.data['createDate'])
            return Response(agent_serializer.data, status=status.HTTP_201_CREATED, headers=headers)
        except CertificateGenerationFailed as e:
            logger.exception(e)
            raise e
        except ApplianceCreationFailed as e:
            raise e
        except Exception as e:
            logger.exception(e, exc_info=True)
            raise AgentCreationFailed(e.args)

    def destroy(self, request, *args, **kwargs):
        self.serializer_class = AgentSerializer
        try:
            obj = self.get_object()
            deleted_at = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            serializer_data = self.get_serializer(obj, context={'request': request}).data

            rmq_manager = RMQ_Management(obj.agentId, operation='spindown')

            obj.delete()
            publish_change_event(request, 'integrations', 'delete', ENTITY_TYPES['agents'],
                                 serializer_data, deleted_at)
            return Response(status=status.HTTP_204_NO_CONTENT)
        except Exception as e:
            logger.exception(e, exc_info=True)
            raise AgentDeleteFailed(e.args)

    def update(self, request, *args, **kwargs):
        self.serializer_class = AgentSerializer
        try:
            input_data = request.data
            obj = self.get_object()
            was_active = obj.active
            serializer = self.get_serializer(obj, data=input_data, partial=True)
            # Validate user-submitted data before continuing
            serializer.is_valid(raise_exception=True)
            
            require_health_check = False
            
            if was_active and not input_data.get('active', was_active):
                input_data['healthStatus'] = {
                    'status': 'inactive',
                    'message': 'Agent is inactive'
                }
                
            elif not was_active and input_data.get('active', was_active):
                rmq_handler = RMQ_Management(obj.agentId)
                input_data['healthStatus'] = {
                    'status': 'disconnected',
                    'message': 'No active Agent connections'
                }
                require_health_check = True
            # Need a new serializer since the validated one can no longer be modified
            serializer = self.get_serializer(obj, data=input_data, partial=True)
            serializer.is_valid()
            obj = serializer.save()

            if require_health_check:
                message_json = {
                    "data": {"agent": obj.agentId},
                    "sourceId": settings.SELF_ID,
                    "destinationId": obj.agentId,
                    "action": REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'),
                    "method": "POST",
                    "tenant": request.tenantid
                }
                message_json['data']['request_data'] = extract_audit_data_from_request(request)
                logger.error(f"Caling broadcast function to publish healthcheck message")
                broadcast_to_remote(message_json, obj.agentId, REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'), 'receiveConnectorInstructionRequest', 'POST')

            headers = self.get_success_headers(serializer.data)
            if getattr(obj, '_prefetched_objects_cache', None):
                # If 'prefetch_related' has been applied to a queryset, we need to
                # forcibly invalidate the prefetch cache on the instance.
                obj._prefetched_objects_cache = {}
            publish_change_event(request, 'integrations', 'update', ENTITY_TYPES['agents'],
                                 serializer.data, serializer.data['modifyDate'])
            return Response(serializer.data, status=status.HTTP_200_OK, headers=headers)
        except Exception as e:
            logger.exception(e, exc_info=True)
            raise AgentUpdateFailed(e.args)


    def bulk_delete(self, request, *args, **kwargs):
        agent_ids = request.data.get('agents', [])
        agent_objs = Agent.objects.filter(agentId__in=agent_ids)
        spundown_agents = []
        for agent in agent_objs:
            try:
                rmq = RMQ_Management(agent.agentId, operation='spindown')
                spundown_agents.append(agent.agentId)
            except Exception as e:
                message = f"Could not delete Agent {agent} :: ERROR: {e.message}"
                logger.exception(message)
                if spundown_agents:
                    spundown_agents = Agent.objects.filter(agentId__in=spundown_agents)
                    spundown_agents.delete()
                raise AgentDeleteFailed(message)
        agent_objs.delete()
        return Response({"message": "success"}, status=status.HTTP_200_OK)
    
    def trigger_healthcheck(self, request, *args, **kwargs):
        agent_id = request.data.get('agent', None)
        if not agent_id:
            agent_id = request.GET.get('agent', None)

        agent_obj = Agent.objects.get(agentId=agent_id)

        # First use RMQ management to see if there are any consumers on the agent vhost
        rmq_connections = RMQ_Management(agent_id=agent_id, operation='connections').connections
        if len(rmq_connections) == 0:
            if agent_obj.healthStatus['status'] != 'awaiting':
                agent_obj.healthStatus = {
                    'status': 'disconnected',
                    'message': 'No active Agent connections'
                }
                agent_obj.save()
            return Response(agent_obj.healthStatus, status=status.HTTP_200_OK)
        rmq_data = {
            'agent': agent_id
        }
        rmq_data['token'] = request.headers.get('Authorization')
        message_json = {
            "data": rmq_data,
            "sourceId": settings.SELF_ID,
            "destinationId": agent_obj.agentId,
            "action": REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'),
            "method": "POST",
            "token": rmq_data.get('token'),
            "tenant": request.tenantid
        }
        message_json['data']['request_data'] = extract_audit_data_from_request(request)
        agent_obj.healthStatus = {
                'status': 'in progress',
                'message': 'Agent health check in progress'
        }
        agent_obj.save()
        broadcast_to_remote(message_json, agent_id, REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'), 'receiveConnectorInstructionRequest', 'POST')
        return Response(agent_obj.healthStatus, status=status.HTTP_200_OK)
    
    def health_status(self, request, *args, **kwargs):
        agent_id = request.GET.get('agent', None)
        agent_obj = Agent.objects.get(agentId=agent_id)
        return Response(agent_obj.healthStatus, status=status.HTTP_200_OK)
    

class AgentInstallerView(APIView):
    http_method_names = ['post']

    def post(self, request, *args, **kwargs):
        request_data = request.data
        agent = request_data.get('agent')

        try:
            agent_obj = Agent.objects.get(agentId=agent)
        except Agent.DoesNotExist:
            return Response({'message': f'Agent {agent} does not exist.'}, status=status.HTTP_404_NOT_FOUND)

        hostname = request.get_host()
        connectors = request.data.get('connectors', [])
        connector_names = [c['name'] for c in connectors]
        if 'cyops_utilities' not in connector_names:
            installed_utils_connector = Connector.objects.get(name='cyops_utilities', agent_id=settings.SELF_ID, installed=True)
            connectors.insert(0, {
                'name': 'cyops_utilities',
                'version': installed_utils_connector.version
            })
        
        install_method = request_data.get('deployment', 'docker')
        if install_method == 'docker':
            file_func = populate_docker_agent_script
            filename = f'deploy-docker-agent-{agent_obj.agentId}.sh'
        elif install_method == 'rpm':
            file_func = populate_installer_sh_file
            filename = 'install-agent-{agent_obj.agentId}.sh'

        try:
            install_file = file_func(agent_obj, tenantid=request.tenantid, hostname=hostname, connectors=connectors)
            with open(install_file, 'rb') as instFile:
                content = instFile.read()
            response = HttpResponse(content)
            response['content_type'] = 'application/text'
            response['Content-Disposition'] = f'attachment; filename={filename}'
            return response
        except Exception as e:
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        finally:
            if os.path.isfile(install_file):
                shutil.rmtree(install_file, ignore_errors=True)

        # if install_method == 'docker':
        #     docker_file = populate_docker_compose_file(agent_obj, tenantid=request.tenantid)

        #     try:
        #         with open(docker_file, 'rb') as shfile:
        #             content = shfile.read()
        #         response = HttpResponse(content)
        #         response['content_type'] = 'application/text'
        #         response['Content-Disposition'] = 'attachment; filename=docker-compose-agent.yaml'
        #         shutil.rmtree(docker_file, ignore_errors=True)
        #         return response
        #     except Exception as e:
        #         return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        # elif install_method == 'rpm':
        #     sh_file = populate_installer_sh_file(agent_obj, tenantid=request.tenantid)