from datetime import datetime
import json
import requests

from django.conf import settings
from django.db import connection
from rest_framework.exceptions import APIException

from agents.constants import REMOTE_AGENT_OPERATIONS
from agents.models import Agent

from connectors.core.constants import REMOTE_CONNECTOR_OPERATIONS, REMOTE_STATUS
from connectors.core.connector import logger, ConnectorError
from connectors.models import Connector, Configuration
from connectors.serializers import ConnectorConfigurationSerializer, ConnectorDetailSerializer
from connectors.utils import create_request_from_ack_payload, insert_connector
from constants.error_message import *
from xfutils.audit import publish_change_event
from xfutils.utils import api_call, generate_interservice_jwt_token

all_config = settings.ALL_CONFIG


def connector_install_ack(data):
    status = data.get('status')
    config = data.pop('config', [])
    agent = data.get('agent')
    request_data = data.pop('request_data', {})
    request_obj = create_request_from_ack_payload(request_data)
    timestamp = data.pop('timestamp', datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ'))

    # Connector is successful on remote
    if status == REMOTE_STATUS.get('finished'):
        try:
            insert_connector(data, config=config, raise_exception=True, ack_call=True)
            data_to_publish = {
                'name': data.get('name'),
                'version': data.get('version'),
                'status': data.get('status'),
                'agent': data.get('agent'),
                'timestamp': data.get('timestamp')
            }
            publish_change_event(request_obj, 'integration', 'install', 'connectors', data_to_publish, timestamp)
        except Exception as e:
            error_message = 'Error occurred while updating the remote connector {0} of agent {1}'.format(
                data.get('name'), agent)
            logger.exception('{0} ERROR:: {1}'.format(error_message, str(e)))
            return error_message
    else:
        # Connector is failed on remote
        try:
            conn_instance = Connector.objects.get(name=data.get('name', None),
                                                    version=data.get('version', None),
                                                    agent=data.get('agent', settings.SELF_ID))
            if status == REMOTE_STATUS.get('failed') and conn_instance.remote_status and \
                    conn_instance.remote_status.get('status') == REMOTE_STATUS.get('installation_in_progress'):
                status = REMOTE_STATUS.get('installation_failed')

            conn_instance.remote_status = {'status': status}
            conn_instance.status = REMOTE_STATUS.get('failed')
            conn_instance.save()
        except Connector.DoesNotExist:
            # A connector which was installed during agent startup will not have a corresponding Connector object on
            # XF, so it must be created
            conn_data = {
                'name': data.get('name'),
                'version': data.get('version'),
                'agent': data.get('agent'),
                'label': data.get('name'),  # info.get('label', info.get('name', None)),
                'status': REMOTE_STATUS.get('failed'),
                'remote_status': {'status': REMOTE_STATUS.get('installation_failed')},
            }
            serializer = ConnectorDetailSerializer(data=conn_data)
            serializer.is_valid(raise_exception=True)
            serializer.save()
        data_to_publish = {
                'name': data.get('name'),
                'version': data.get('version'),
                'status': REMOTE_STATUS.get('installation_failed'),
                'agent': data.get('agent'),
                'timestamp': data.get('timestamp')
            }
        publish_change_event(request_obj, 'integration', 'install', 'connectors', data_to_publish, timestamp)

    return "connector status updated successfully"


def connector_uninstall_ack(data):
    status = data.get('status')
    name = data.get('name')
    version = data.get('version')
    agent = data.get('agent')
    connector_query = Connector.objects.filter(name=name, version=version, agent=agent)
    request_data = data.pop('request_data', None)
    request_obj = create_request_from_ack_payload(request_data)
    timestamp = data.pop('timestamp')
    if connector_query.exists():
        try:
            if status == REMOTE_STATUS.get('uninstall_finished'):
                # connector_instance = connector_query.first()
                # connector_instance.metadata.update({'upgrade_in_progress': data.get('upgrade_in_progress')})
                # connector_instance.save(update_fields=['metadata'])
                connector_query.delete()
                publish_change_event(request_obj, 'integration', 'uninstall', 'connectors', data, timestamp)
                return "Connector deleted successfully"
            elif status == REMOTE_STATUS.get('failed'):
                conn_instance = connector_query.first()
                if status == REMOTE_STATUS.get('failed') and conn_instance.remote_status and \
                        conn_instance.remote_status.get('status') == REMOTE_STATUS.get(
                    'uninstall_in_progress'):
                    status = REMOTE_STATUS.get('uninstall_failed')
                conn_instance.remote_status = {'status': status}
                conn_instance.status = 'installed'
                conn_instance.save()
                publish_change_event(request_obj, 'integration', 'uninstall', 'connectors', data, timestamp)
                return "connector status updated successfully"
        except Exception as e:
            error_message = xf_integration_31.format(name, version, agent)
            logger.error('{0} Error:: {1}'.format(error_message, str(e)))
            return error_message
    else:
        error_message = xf_integration_31.format(name, version, agent)
        logger.error(f'{error_message} Error:: Received uninstall ack request for a connector which is not installed')
        return error_message


def connector_update_ack(data):
    connector = None
    try:
        connector = Connector.objects.filter(name=data.get('name', ''),
                                             version=data.get('version', ''),
                                             agent=data.get('agent', settings.SELF_ID)).first()
        connector.remote_status = {'status': REMOTE_STATUS.get('status_update_finished')}
        connector.status = {'status': REMOTE_STATUS.get('status_update_finished')}
        connector.active = data.get('active', True)
        connector.save()
        return 'Connector updated successfully'
    except Exception as e:
        logger.exception(e)
        error_message = xf_integration_38[1].format(e)
        if connector:
            connector.remote_status = {'status': REMOTE_STATUS.get('status_update_failed')}
            connector.status = {'status': REMOTE_STATUS.get('status_update_failed')}
            connector.save()
        return error_message


def connector_config_ack(data):
    status = data.get('status')
    config_id = data.get('config_id')
    agent = data.get('agent')
    name = data.get('name')
    operation = data.pop('operation')
    request_data = data.pop('request_data', None)
    request_obj = create_request_from_ack_payload(request_data)
    timestamp = data.pop('timestamp')

    config_qs = Configuration.objects.filter(config_id=config_id, agent=agent)
    try:
        config_instance = config_qs.first()
        conn_obj = config_instance.connector
        config_data = {
            'name': name,
            'default': data.get('default', config_instance.default),
            'remote_status': {"status": status},
            'status': 0
        }
        configuration_serializer = ConnectorConfigurationSerializer(config_instance, data=config_data, partial=True)

        if data.get('default', False):
            Configuration.objects.filter(
                connector=conn_obj.id,
                agent_id=agent,
                default=True).update(default=False)

        configuration_serializer.is_valid(raise_exception=True)
        configuration_serializer.save()
        conn_obj.config_count = conn_obj.config_count + 1
        conn_obj.save()
        publish_change_event(request_obj, 'integration', operation, 'configurations', configuration_serializer.data,
                             timestamp)
        return "Configuration status updated successfully"
    except Exception as e:
        error_message = xf_integration_8.format(agent)
        logger.error('{0} ERROR:: {1}'.format(error_message, str(e)))
        return error_message


def connector_config_delete_ack(data):
    request_data = data.pop('request_data', None)
    request_obj = create_request_from_ack_payload(request_data)
    timestamp = data.pop('timestamp')

    conf_obj = Configuration.objects.get(agent_id=data.get('agent'), config_id=data.get('config_id'))
    conf_obj.delete()
    publish_change_event(request_obj, 'integration', 'delete', 'configurations', data, timestamp)
    return "Connector configuration deleted successfully"


def connector_execution_ack(data):
    request_data = data.pop('request_data', None)
    if request_data:
        request_obj = create_request_from_ack_payload(request_data)
    timestamp = data.pop('timestamp', None)
    exec_data = data.get('data', {})
    # @TODO: Update remote status in Database
    if exec_data.get('env', {}).get('poll'):
        return _resume_workflow_poll_trigger(data)
    if exec_data.get('env', {}).get('wf_id'):
        output = _send_workflow_execution_response(data)
        #publish_change_event(request_obj, 'integration', 'execute', 'connectors', data, timestamp)
        return output
    else:
        message = "Execution result received"
        return message


def connector_healthcheck_ack(data):
    request_data = data.pop('request_data', None)
    request_obj = create_request_from_ack_payload(request_data)
    timestamp = data.pop('timestamp')

    config_obj = Configuration.objects.get(config_id=data.get('config_id'), connector__name=data.get('name'))
    status = {
        'message': data.get('message'),
        'status': data.get('status'),
        'last_checked': str(datetime.now())
    }
    config_obj.health_status = status
    config_obj.save()
    publish_change_event(request_obj, 'integration', 'healthcheck', 'connectors', data, timestamp, skip_audit=True)
    return "Healthcheck output received"


def agent_healthcheck_ack(data):
    request_data = data.pop('request_data', None)
    # Request data won't be present if this is coming from an agent during
    # startup or shutdown
    if data.get('healthStatus').get('message') in ['Agent has been connected', 'Agent cannot connect to RabbitMQ', 'Agent has been shut down']:
        tenant = connection.get_tenant()
        request_data['tenantuuid'] = str(tenant.uuid)
        request_data['subtenantid'] = None
        request_data['teams'] = []

    request_obj = create_request_from_ack_payload(request_data)
    timestamp = data.pop('timestamp')
    agent_id = data.pop('agent')
    # data['uuid'] = agent_id

    agent_obj = Agent.objects.get(agentId=agent_id)
    data['uuid'] = str(agent_obj.uuid)
    agent_obj.healthStatus = data.get('healthStatus')
    # Update the stored Agent version if a version is included with the payload
    if data.get('version'):
        agent_obj.version = data.get('version')

    agent_obj.save()

    publish_change_event(request_obj, 'integration', 'healthcheck', 'agents', data, timestamp, skip_audit=True)
    return "Agent health status updated successfully"


def _send_workflow_execution_response(message):
    logger.info(f"Received playbook step connector execution result:\n{message}")
    data = message.get('data', {})
    peer_wf_id = data.get('env', {}).get('wf_id')
    result = data
    remote_status = data.get('remote_status', 'finished') if data.get('status',
                                                                      '').lower() == 'success' else 'failed'
    step_id = data.get('env').get('step_id')
    child_step_id = data.get('env').get('child_step_id')
    obj = {'remote_result': result, 'remote_status': remote_status, 'step_id': step_id,
           'child_step_id': child_step_id}
    
    token = generate_interservice_jwt_token(all_config, tenantid=connection.get_tenant().tenant_id)
    headers = {"Authorization": f"Bearer {token}"}
    try:
        url = f"https://{settings.ALL_CONFIG.get('application', 'wf_host')}/api/workflow/workflows/{str(peer_wf_id)}/resume/?format=json"
        response = api_call(url, method='POST', body=obj, headers=headers, verify=False)
        return 'Successfully resumed the awaiting workflow'
    except Exception as e:
        logger.exception(e)
        source_id = message.get('sourceId')
        # logger.exception('Error: %s while resuming workflow with workflow id: %s from source %s, ', str(e),
        #                    peer_wf_id, source_id)
        wf_terminate_url = 'https://' + settings.ALL_CONFIG.get('application',
                                                                'wf_host') + '/api/workflow/revoke/tasks/?format=json'
        obj = {'wfLogId': peer_wf_id, 'wf_failure_message': xf_integration_21.format(source_id, str(e))}
        try:
            response = api_call(wf_terminate_url, method='POST', body=obj, headers=headers, verify=False)
            return "Successfully terminated the awaiting workflow"
        except Exception as e:
            message = xf_integration_32.format(peer_wf_id)
            logger.exception(f'{message} ERROR: {str(e)}')
            return message


def _resume_workflow_poll_trigger(message):
    data = message.get('data', {})
    peer_wf_id = data.get('env', {}).get('wf_id', '')
    if message.get('status', '').lower() == 'failed':
        logger.error(f'message: {message}')
        data['message'] = message.get('message', '')

    token = generate_interservice_jwt_token(all_config, tenantid=connection.get_tenant().tenant_id)
    headers = {"Authorization": f"Bearer {token}"}
    try:
        url = "https://" + settings.ALL_CONFIG.get('application', 'wf_host') + "/api/workflow/workflows/" + str(
            peer_wf_id) + "/resume-poll-task/?format=json"
        response = api_call(url, method='POST', body=data, headers=headers, verify=False)
        return 'Successfully resumed the awaiting poll trigger task'
    except requests.exceptions.HTTPError as http_err:
        if http_err.response.status_code == 429:
            # @TODO: Do nothing if HTTP error is 429 Too Many Requests
            logger.error("Do nothing for HTTP error is 429 Too Many Requests")
            pass
        else:
            # Handle other HTTP errors
            logger.exception(http_err)
    except Exception as e:
        logger.exception(f'response: {response}  ERROR: {e} ')
        # TODO: handle failure
        try:
            wf_terminate_url = ('https://' + settings.ALL_CONFIG.get('application', 'wf_host')
                                + ('/api/workflow/revoke/tasks'
                                   '/?format=json'))
            obj = {'wfLogId': peer_wf_id, 'wf_failure_message': 'Resume Poll Task Failed'}
            response = api_call(wf_terminate_url, method='POST', body=obj, headers=headers, verify=False)
            return "Successfully terminated the awaiting workflow"
        except Exception as e:
            message = xf_integration_32.format(peer_wf_id)
            logger.exception(f'{message} ERROR: ')
            return message


ack_operations = {
    REMOTE_CONNECTOR_OPERATIONS.get('connector_install'): connector_install_ack,
    REMOTE_CONNECTOR_OPERATIONS.get('connector_uninstall'): connector_uninstall_ack,
    REMOTE_CONNECTOR_OPERATIONS.get('connector_update'): connector_update_ack,
    REMOTE_CONNECTOR_OPERATIONS.get('config_create'): connector_config_ack,
    REMOTE_CONNECTOR_OPERATIONS.get('config_delete'): connector_config_delete_ack,
    REMOTE_CONNECTOR_OPERATIONS.get('connector_execution'): connector_execution_ack,
    REMOTE_CONNECTOR_OPERATIONS.get('connector_healthcheck'): connector_healthcheck_ack,
    REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'): agent_healthcheck_ack,
}
