import uuid
from django.conf import settings
from rest_framework import status
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from xfutils.exceptions import XFException

from connectors.models import Connector, Configuration
from connectors.serializers import ConnectorDetailSerializer, ConnectorConfigurationSerializer
from connectors.core.constants import REMOTE_STATUS
from connectors.utils import broadcast_to_remote, encrypt_password, get_or_create_team_or_role, extract_audit_data_from_request, validate_agent_health
from connectors.core.connector import logger, ConnectorError
from connectors.core.constants import REMOTE_CONNECTOR_OPERATIONS
from connectors.connector_operations import ConnectorOperation
from constants.error_message import *


class RemoteOperations(ConnectorOperation):
    def install_connector(self, data, request=None):
        # Initial entry in DB for connector install
        name = data.get('name')
        version = data.get('version')
        agent_id = data.get('agent')
        agent_obj = validate_agent_health(agent_id)

        conn_data = {
            'name': name,
            'version': version,
            'agent': agent_id,
            'label': name,  # info.get('label', info.get('name', None)),
            'status': 'in-progress',
            'remote_status': {'status': REMOTE_STATUS.get('installation_in_progress')},
            # 'tags': tags,
        }
        serializer = ConnectorDetailSerializer(data=conn_data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        conn_data = serializer.data
        # with open(connector_tgz_path, 'rb') as tgz_file:
        #     conn_data['file_content'] = base64.b64encode(tgz_file.read())
        #     conn_data['filename'] = filename
        if request:
            request_data = extract_audit_data_from_request(request)
            data['request_data'] = request_data
        message_json = {
            "data": data,
            "status": "initial",
            "sourceId": settings.SELF_ID,
            "destinationId": agent_id,
            "action": REMOTE_CONNECTOR_OPERATIONS.get('connector_install'),
            "method": "POST",
            "tenant": data.get('tenant')
        }
        broadcast_to_remote(message_json, agent_id, REMOTE_CONNECTOR_OPERATIONS.get('connector_install'),
                            'receiveConnectorInstructionRequest', 'POST')
        return conn_data

    def uninstall_connector(self, data, request=None):
        name = data.get('name')
        version = data.get('version')
        agent_id = data.get('agent')
        agent_obj = validate_agent_health(agent_id)

        instance = Connector.objects.filter(name=name, version=version, agent_id=agent_id).first()
        instance.remote_status = {'status': REMOTE_STATUS.get('uninstall_in_progress')}
        instance.status = {'status': REMOTE_STATUS.get('uninstall_in_progress')}
        instance.save()
        if request:
            request_data = extract_audit_data_from_request(request)
            data['request_data'] = request_data
        message_json = {
            "data": data,
            "status": REMOTE_STATUS.get('uninstall_in_progress'),
            "sourceId": settings.SELF_ID,
            "destinationId": agent_id,
            "action": REMOTE_CONNECTOR_OPERATIONS.get('connector_uninstall'),
            "method": "POST",
            "tenant": data.get('tenant')
        }

        broadcast_to_remote(message_json, agent_id, REMOTE_CONNECTOR_OPERATIONS.get('connector_uninstall'),
                            'receiveConnectorInstructionRequest', 'POST')
        return data

    def connector_detail_update(self, data, request=None):
        agent_id = data.get('agent')
        instance = Connector.objects.filter(name=data.get('name'), version=data.get('version'),
                                                agent_id=agent_id).first()
        data['name'] = instance.name
        data['version'] = instance.version
        instance.remote_status = {'status': REMOTE_STATUS.get('status_update_in_progress')}
        instance.status = {'status': REMOTE_STATUS.get('status_update_in_progress')}
        message_json = {
            'data': data,
            'status': REMOTE_STATUS.get('status_update_in_progress'),
            'sourceId': settings.SELF_ID,
            'destinationId': agent_id,
            'action': REMOTE_CONNECTOR_OPERATIONS.get('connector_update'),
            'method': 'PUT',
            'tenant': data.get('tenant')
        }
        broadcast_to_remote(message_json, agent_id, REMOTE_CONNECTOR_OPERATIONS.get('connector_update'),
                            'receiveConnectorInstructionRequest', 'PUT')
        return Response(data, status=status.HTTP_200_OK)

    def create_connector_config(self, data, request=None, operation='create'):
        name = data.get('name', '')
        conn_name = data.get('connector_name')
        conn_version = data.get('connector_version')
        agent = data.get('agent')
        agent_obj = validate_agent_health(agent)

        teams = data.pop('teams', [])
        tags = data.pop('tags', [])
        conn_instance = Connector.objects.filter(name=conn_name, version=conn_version, agent_id=agent).first()

        if not conn_instance:
            raise APIException('Connector Does not exist', status.HTTP_400_BAD_REQUEST)

        if conn_instance.remote_status.get('status') in ['in-progress', 'deletion-in-progress']:
            raise APIException('Connector installation is in progress, please try after completion', status.HTTP_400_BAD_REQUEST)

        if conn_instance.remote_status.get('status') in ['failed']:
            raise APIException('Connector installation is failed, please retry the installation', status.HTTP_500_INTERNAL_SERVER_ERROR)

        config_id = data.get('config_id')
        config_instance = None
        remote_status = {
            'status': REMOTE_STATUS.get('in_progress'),
            'message': 'Configuration has been published to agent {0}'.format(agent)
        }
        if config_id:
            config_instance = Configuration.objects.filter(config_id=config_id).first()

        if config_instance and data.get('remote_status', {}).get('status') == REMOTE_STATUS.get('failed'):
            conn_config_obj = {
                'remote_status': remote_status
            }
        else:
            config = data.get('config')
            encrypt_password(conn_instance.config_schema, config)
            conn_config_obj = {
                'config_id': data.get('config_id', str(uuid.uuid4())),
                'name': name,
                'default': data.get('default', False),
                'config': config,
                'connector': conn_instance.id,
                'agent': agent,
                'remote_status': remote_status,
                'status': REMOTE_STATUS.get('in_progress'),
                'tags': tags
            }
            if teams:
                team_ids = get_or_create_team_or_role('team', teams)
                conn_config_obj.update({'team_ids': team_ids})
        try:
            serializer = ConnectorConfigurationSerializer(config_instance, data=conn_config_obj)
            serializer.is_valid(raise_exception=True)
            serializer.save()
            conn_config_obj.update({
                'teams':teams, 'connector_name': conn_name,
                'connector_version': conn_version, "tenant": data.get('tenant')
            })
            if request:
                request_data = extract_audit_data_from_request(request)
                conn_config_obj['request_data'] = request_data
            message_json = {
                "data": conn_config_obj,
                "status": "initial",
                "sourceId": settings.SELF_ID,
                "destinationId": agent,
                "action": REMOTE_CONNECTOR_OPERATIONS.get('config_create'),
                "method": "POST",
            }
            broadcast_to_remote(message_json, agent, REMOTE_CONNECTOR_OPERATIONS.get(f'config_{operation}'),
                                'receiveConnectorInstructionRequest', 'POST')
            return serializer.data
        except Exception as e:
            message = 'Connector configuration failed: {0}'.format(str(e))
            raise APIException(message, status.HTTP_500_INTERNAL_SERVER_ERROR)

    def update_connector_config(self, data, request=None):
        self.create_connector_config(data, request=request, operation='update')
    
    def delete_connector_config(self, data, request=None):
        agent = data.get('agent')
        agent_obj = validate_agent_health(agent)

        config_id = data.get('config_id')
        config_obj = Configuration.objects.get(agent=agent, config_id=config_id)

        if not config_obj:
            raise APIException('Connector Does not exist', status.HTTP_400_BAD_REQUEST)

        if config_obj.remote_status.get('status') in ['deletion-in-progress']:
            raise APIException('Connector deletion is in progress, please try after completion', status.HTTP_500_INTERNAL_SERVER_ERROR)
        try:
            remote_status = {
                'status': REMOTE_STATUS.get('deletion_in_progress'),
                'message': 'Deleting connector configuration on Agent {0}'.format(agent)
            }
            serializer = ConnectorConfigurationSerializer(config_obj, data=remote_status, partial=True)
            serializer.is_valid(raise_exception=True)
            serializer.save()

            if request:
                request_data = extract_audit_data_from_request(request)
                data['request_data'] = request_data
            message_json = {
                "data": data,
                "status": "initial",
                "sourceId": settings.SELF_ID,
                "destinationId": agent,
                "action": REMOTE_CONNECTOR_OPERATIONS.get('config_delete'),
                "method": "POST"
            }
            broadcast_to_remote(message_json, agent, REMOTE_CONNECTOR_OPERATIONS.get('config_delete'),
                                'receiveConnectorInstructionRequest', 'POST')
        except Exception as e:
            raise APIException(f'Failed to delete Connector Configuration. ERROR :: {repr(e)}', status.HTTP_500_INTERNAL_SERVER_ERROR)
        

    def connector_execution(self, data, request=None):
        agent = data.get('agent')
        agent_obj = validate_agent_health(agent)

        if hasattr(request, 'tenantuuid'):
            request_data = extract_audit_data_from_request(request)
            data['request_data'] = request_data
        message_json = {
            'data': data,
            'status': 'initial',
            'sourceId': settings.SELF_ID,
            'destinationId': agent,
            'action': REMOTE_CONNECTOR_OPERATIONS.get('connector_execution'),
            'method': 'POST'
        }

        broadcast_to_remote(message_json, agent, REMOTE_CONNECTOR_OPERATIONS.get('connector_execution'),
                            'receiveConnectorExecutionRequest', 'POST')
        # @TODO : create execute action entry with remote status in progress
        return {"remote_status": REMOTE_STATUS.get('awaiting')}, False
    
    def connector_healthcheck(self, data, request=None):
        agent = data.get('agent', settings.SELF_ID)
        agent_obj = validate_agent_health(agent)
        
        config = data.get('configuration')
        action = REMOTE_CONNECTOR_OPERATIONS.get('connector_healthcheck')
        if request:
            request_data = extract_audit_data_from_request(request)
            data['request_data'] = request_data
        message_json = {
            'data': data,
            'sourceId': settings.SELF_ID,
            'destinationId': agent,
            'action': action,
            'method': 'POST',
            "tenant": data.get('tenant')
        }
        broadcast_to_remote(message_json, agent, action, 'receiveConnectorInstructionRequest', 'POST')
        config_instance = Configuration.objects.filter(config_id=config.get('config_id')).first()
        status = {'status':REMOTE_STATUS.get('in_progress'), 'message': 'Connector health check in progress'}
        config_instance.health_status = status
        config_instance.save()
        return status