import base64
import datetime
import os
import shutil
import tempfile
import uuid
from threading import Thread
import traceback

import requests
from django.conf import settings
from rest_framework import status
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from urllib3.exceptions import InsecureRequestWarning

from agents.constants import REMOTE_AGENT_OPERATIONS

from connectors.core.connector import logger, ConnectorError
from connectors.serializers import ConnectorConfigurationSerializer
from constants.error_message import *
from connectors.core.constants import INSTALL_SOURCE_API_UPLOAD, REMOTE_CONNECTOR_OPERATIONS, REMOTE_STATUS
from connectors.models import Connector, Configuration
from connectors.utils import (untar_connector_tgz, get_validated_info_file, validate_config, broadcast_to_remote,
                              import_connector, install_connector_files, delete_connector, send_connector_operations_ack,
                              encrypt_password, get_or_create_team_or_role, omit_password, update_pwd_fields,
                              get_connector_instance, get_operation, get_configuration, is_operation_param_valid,
                              get_parsed_config, update_env_files, download_connector_tgz, delete_events, set_connector_file_permissions)
from connectors.connector_operations import ConnectorOperation
from xfutils.audit import publish_change_event
from xfutils.exceptions import XFException


class SelfOperations(ConnectorOperation):
    def install_connector(self, data, request=None, custom_connector=False):
        name = data.get('name')
        version = data.get('version')
        agent_id = data.get('agent')
        replace = data.get('replace')
        trust_api = data.get('trust_api')
        rbac_info = data.get('rbac_info')
        connector_tgz_path = ""
        error_message = ""

        if not settings.LW_AGENT and name in settings.NOT_ALLOWED_CONN_SELF:
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_integration_37, name)

        try:
            if data.get('file'):
                # Write the tgz file in file system
                connector_tgz_path = data.get('file')
            elif settings.LW_AGENT:
                # Download the file in case of light weight agent
                # In global the files will be already there
                logger.info(f'Downloading TGZ file')
                connector_tgz_path = download_connector_tgz(name, version)

            # Un-tar connector tgz file
            logger.info(f'Decompressing TGZ file')
            connector_path = untar_connector_tgz(connector_tgz_path, conn_name=name, conn_version=version)
            info = get_validated_info_file(connector_path, replace, name, version)
            info.update(data.get('extra_data', dict()))
            logger.info(f'Importing connector')
            result = import_connector(
                info, connector_path=connector_path,
                replace=replace, rbac_info=rbac_info,
                install_source=INSTALL_SOURCE_API_UPLOAD,
                tenant=data.get('tenant'),
                agent=agent_id,
                config=data.get('config', dict()),
                request_data=data.get('request_data'),
                omit_password= not settings.LW_AGENT
            )
            
            if settings.LW_AGENT:
                logger.info(f'Installing connector files')
                install_connector_files(name, version, connector_path)
                if custom_connector:
                    set_connector_file_permissions(name, version)
                info.update({
                    'status': REMOTE_STATUS.get('finished'),
                    'agent': agent_id,
                    'icon_small': result.get('icon_small'),
                    'icon_large': result.get('iconLarge'),
                    'timestamp': result.get('modified'),
                    'config': result.get('configuration', []),
                })
                if data.get('request_data', None):
                    info['request_data'] = data.get('request_data')
                logger.info(f'Sending status update message to XF...')
                self.handle_response(info, REMOTE_STATUS.get('finished'), REMOTE_CONNECTOR_OPERATIONS.get('connector_install'))
                logger.info(f'Status update message sent')
            else:
                data_to_publish = {
                        'name': name,
                        'version': version,
                        'status': REMOTE_STATUS.get('finished'),
                        'agent': agent_id,
                        'timestamp': result.get('modified')
                }
                publish_change_event(request, 'integration', 'install', 'connectors', data_to_publish, result.get('modified'))
            return result
        except ConnectorError as exp:
            error_message = 'From the connector import exception handling %s' % str(exp)
            logger.exception(error_message)
            try:
                result.status = REMOTE_STATUS.get('installation_failed')
                result.save()
            except NameError:
                pass
            if custom_connector:
                raise exp
            else:
                self.handle_response(data, REMOTE_STATUS.get('failed'), REMOTE_CONNECTOR_OPERATIONS.get('connector_install'), error_message)
        except Exception as exp:
            logger.exception('{0} ERROR :: {1}'.format(xf_integration_1, str(exp)))
            try:
                result.status = REMOTE_STATUS.get('installation_failed')
                result.save()
            except NameError:
                pass
            if custom_connector:
                raise exp
            else:
                self.handle_response(data, REMOTE_STATUS.get('failed'), REMOTE_CONNECTOR_OPERATIONS.get('connector_install'), error_message)
        finally:
            pass
            # Todo uncomment with remove of install files on self
            # if os.path.exists(os.path.join(tempfile.gettempdir(), 'connectors', name)):
            #     shutil.rmtree(os.path.join(tempfile.gettempdir(), 'connectors', name))
            # if os.path.exists(connector_tgz_path):
            #     os.remove(connector_tgz_path)

    def uninstall_connector(self, data, request=None):
        message_json = {
            'data': {
                'name': data.get('name'),
                'version': data.get('version'),
            },
            'sourceId': settings.SELF_ID,
            'destinationId': data.get('agent'),
            'action': 'connectorDelete',
            'method': 'DELETE'
        }
        delete_connector(data, request)
        return {'message': "Connector deleted"}

    def connector_detail_update(self, data, request=None):
        try:
            instance = Connector.objects.filter(name=data.get('name'), version=data.get('version')).first()
            data.update({
                'name': instance.name,
                'version': instance.version,
                'active': data.get('active', True),
                'operation': 'update'
            })
            instance.active = data.get('active', True)
            instance.save(update_fields=['active'])
            response = {
                'name': instance.name,
                'version': instance.version,
                'active': instance.active,
            }
            if settings.LW_AGENT:
                data.update({'status': REMOTE_STATUS.get('status_update_finished')})
                send_connector_operations_ack(data, REMOTE_CONNECTOR_OPERATIONS.get('connector_update'),
                                              data.get('tenant'))
            return Response(response, status=status.HTTP_200_OK)
        except Exception as e:
            if settings.LW_AGENT:
                data.update({'status': REMOTE_STATUS.get('status_update_failed')})
                send_connector_operations_ack(data, REMOTE_CONNECTOR_OPERATIONS.get('connector_update'),
                                              data.get('tenant'))
            raise XFException(status.HTTP_400_BAD_REQUEST, xf_integration_38, e)

    def update_connector_config(self, data, request=None):
        try:
            return self.create_or_update_connector_config(data, request=request, operation='update')
        except Exception as e:
            if settings.LW_AGENT:
                data.update({
                            'status': REMOTE_STATUS.get('failed'),
                            'timestamp': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
                            'operation': 'update'
                            })
                logger.exception(f"Sending Exception response on update connector config operation: {data}")
                send_connector_operations_ack(data, REMOTE_CONNECTOR_OPERATIONS.get('config_create'), data.get('tenant'))
                data.update({'status': REMOTE_STATUS.get('failed')})
            else:
                raise e

    def create_connector_config(self, data, request=None):
        try:
            return self.create_or_update_connector_config(data, request=request)
        except Exception as e:
            if settings.LW_AGENT:
                data.update({
                            'status': REMOTE_STATUS.get('failed'),
                            'timestamp': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
                            'operation': 'create'
                            })
                logger.exception(f"Sending Exception response on create connector config operation: {data}")
                send_connector_operations_ack(data, REMOTE_CONNECTOR_OPERATIONS.get('config_create'), data.get('tenant'))
            else:
                raise e

    def create_or_update_connector_config(self, data, request=None, operation='create'):
        teams = data.get("teams", [])
        config = data.get('config', {})
        conn_name = data.get('connector_name')
        conn_version = data.get('connector_version')
        agent = data.get('agent', settings.SELF_ID)
        try:
            conn_obj_list = Connector.objects.filter(name=conn_name, version=conn_version, agent_id=agent)
            if not conn_obj_list:
                raise XFException(status.HTTP_400_BAD_REQUEST, xf_integration_55, conn_name, conn_version)
            else:
                conn_obj = conn_obj_list.first()

            if not conn_obj.metadata.get('multi_configuration_allowed', True):
                if conn_obj.config_count:
                    raise Exception(
                        'Connector "{0}" does not support more than one configuration'.format(conn_obj.name))

            agent_obj = conn_obj.agent

            if not data.get('config_id', None):
                data['config_id'] = str(uuid.uuid4())
            validate_config(config, conn_obj.config_schema)

            # Todo this will be used when on add config will be called
            # decrypted_config = copy.deepcopy(config)
            # decrypted_config['config_id'] = data.get('config_id')
            # decrypted_config['name'] = data.get('name')

            encrypt_password(conn_obj.config_schema, config)

            # Update other config to default False
            if data.get('default', False):
                Configuration.objects.filter(
                    connector=data.get('connector'),
                    agent_id=agent_obj.agentId,
                    default=True).update(default=False)

            create_data = {
                'config_id': data.get('config_id'),
                'name': data.get('name', ''),
                'default': data.get('default', False),
                'config': config,
                'connector': conn_obj.id,
                'agent': data.get('agent', agent_obj.agentId),
                'remote_status': {'status': REMOTE_STATUS.get('finished')},
                'status': 0,
                'tags':  data.get('tags', [])
            }

            if teams:
                team_ids = get_or_create_team_or_role('team', teams)
                create_data.update({"team_ids": team_ids})
            if operation=='update':
                config_instance = Configuration.objects.filter(agent_id=agent, config_id=data.get('config_id')).first()
                update_pwd_fields(create_data.get('config'), config_instance.config, config_instance.connector.config_schema)
                configuration_serializer = ConnectorConfigurationSerializer(config_instance, data=create_data)
            else:
                configuration_serializer = ConnectorConfigurationSerializer(data=create_data)
            try:
                configuration_serializer.is_valid(raise_exception=True)
                configuration_serializer.save()
            except Exception as e:
                error_message = f'Bad request data for configuration {operation}'
                index = str(e).find('string=')
                if index:
                    start_index = index + len('string=\'')
                    end_index = str(e).find('\'', start_index)
                    error_message = str(e)[start_index:end_index]
                else:
                    error_message = 'Data missing or Unique constraint violated'
                raise XFException(status.HTTP_400_BAD_REQUEST, xf_integration_55, operation, error_message)

            # TODO: how config add events will be supported in SaaS
            # conn_instance = get_connector_instance(conn_obj.name, conn_obj.version, conn_obj)
            # try:
            #     conn_instance.on_add_config(decrypted_config, conn_obj.active)
            # except Exception as e:
            #     response_status = status.HTTP_207_MULTI_STATUS
            #     logger.error('Configuration add listener for the connector {0} failed with Error :: {1}'.format(
            #         conn_obj.name, str(e)))

            result = configuration_serializer.data
            if operation=='create':
                conn_obj_list.update(config_count=conn_obj.config_count + 1)
            omit_password(result)
            if settings.LW_AGENT:
                data.update({
                    'config_id': configuration_serializer.data.get('config_id'),
                    'status': REMOTE_STATUS.get('finished'),
                    'timestamp': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
                    'operation': operation
                })
                logger.warn(f"Finished configuration {operation} successfully, sending operation acknowledgement")
                send_connector_operations_ack(data, REMOTE_CONNECTOR_OPERATIONS.get('config_create'), data.get('tenant'))
            else:
                publish_change_event(request, 'integration', operation, 'configurations', result,
                    datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ'))
            return result
        except Connector.DoesNotExist:
            error_message = xf_integration_6.format(data.get('connector'))
            logger.error(error_message)
            raise XFException(status.HTTP_400_BAD_REQUEST, ('xf_integration_6', xf_integration_6), data.get('connector'))
        except ConnectorError as e:
            error_message = str(e)
            logger.error(error_message)
            raise APIException(error_message, status.HTTP_500_INTERNAL_SERVER_ERROR)
        except Exception as e:
            error_message = xf_integration_7.format(conn_obj.name, conn_obj.version)
            logger.exception('{0} ERROR:: {1}'.format(error_message, str(e)))
            raise XFException(status.HTTP_500_INTERNAL_SERVER_ERROR, ('xf_integration_7', xf_integration_7), conn_obj.name, conn_obj.version)
            

    def delete_connector_config(self, data, request=None):
        agent = settings.SELF_ID
        config_id = data.get('config_id')
        response_data = {
            "status": "success",
            "config_id": config_id,
            "agent": agent,
            "timestamp": datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        }
        config_object = Configuration.objects.get(agent_id=agent, config_id=config_id)
        connector_object = config_object.connector
        config_object.delete()
        if connector_object.config_count > 0:
            connector_object.config_count = connector_object.config_count - 1
            connector_object.save()
        if settings.LW_AGENT:
            if data.get('request_data', None):
                response_data['request_data'] = data.get('request_data')
            send_connector_operations_ack(response_data, REMOTE_CONNECTOR_OPERATIONS.get('config_delete'), data.get('tenant'))
        else:
            publish_change_event(request, 'integration', 'delete', 'configurations', data, response_data['timestamp'])
            return response_data

    def connector_execution(self, data, request=None):
        try:
            name = data.get('name')
            if not name:
                name = data.get('connector')
            version = data.get('version')
            agent_id = data.get('agent')
            operation = data.get('operation')
            params = data.get('params', {})
            env = data.pop('env', {})

            conn_obj = Connector.objects.get(name=name, version=version, agent_id=agent_id)
            connector = get_connector_instance(name, version, conn_obj)

            # Validate the operation params
            data['params'] = is_operation_param_valid(operation, params)
            secret_key = data.get('secret_key', '')

            # If it is direct call evaluate jinja
            if not secret_key == settings.SEALAB_SECRET_KEY:
                try:
                    data['params'] = get_parsed_config(data.get('params'))
                except Exception as e:
                    pass

            connector_execute_data = {
                "params": data.get('params', {}),
                "operation": operation.get('operation'),
                "config":  data.get('config', {})
            }
            response, is_binary = connector._handle_operation(connector_execute_data, env=env, request=request)
            update_env_files(env)
            # Check why request_data is sent in send_connector_operations_ack
            # if data.get('request_data', None):
            #     response_data['request_data'] = data.get('request_data')
            response_data = {
                "status": "complete",
                "data": response,
                "is_binary": is_binary,
                "timestamp": datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
            }
            self.handle_response(response_data, "complete",
                                 REMOTE_CONNECTOR_OPERATIONS.get('connector_execution'),
                                 tenant=data.get('tenant'))
            return response, is_binary
        except Exception as e:
            logger.exception(f'\n\n{e}\n\n')
            # if data.get('request_data', None):
            #     response_data['request_data'] = data.get('request_data')
            response_data = {
                'data': {
                    'status': 'failed',
                    'env': env,
                    'data': str(e),
                    'timestamp': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
                }
            }
            self.handle_response(response_data, "failed", REMOTE_CONNECTOR_OPERATIONS.get('connector_execution'), message=str(e), tenant=data.get('tenant'))

    def connector_healthcheck(self, data, request=None):
        agent = data.get('agent', settings.SELF_ID)
        config = data.get('configuration')
        config_id = config.get('config_id')
        name = data.get('connector')
        version = data.get('version')
        health_status = REMOTE_STATUS.get('disconnected')
        message = ""
        try:
            conn_obj = Connector.objects.get(name=name, version=version, agent_id=agent)
            connector = get_connector_instance(name, version, conn_obj)
            response = connector.verify_health(config, conn_obj.active)
            if response.get('status'):
                health_status = REMOTE_STATUS.get(response.get('status').lower(), response.get('status'))
            else:
                health_status = REMOTE_STATUS.get('available')
            message = response.get('message', message)
        except Connector.DoesNotExist:
            message = 'Connector {0} with version {1} not found'.format(name, version)
            logger.error(message)
        except ConnectorError as e:
            message = 'Error on health check of the connector {0} version {1}: {2}'.format(name, version, str(e))
            logger.exception('{0} ERROR {1}'.format(message, str(e)))
        except Exception as e:
            message = 'Error on health check of the connector {0} version {1}'.format(name, version)
            logger.exception('{0} ERROR {1}'.format(message, str(e)))

        config_instance = Configuration.objects.filter(config_id=config_id).first()
        config_instance.health_status = {'status': health_status, 'message': message}
        config_instance.save()

        ack_data = {
            'name': data.get('connector'),
            'version': data.get('version'),
            'config_id': config_id,
            'status': health_status,
            'message': message,
            'timestamp': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        }

        if settings.LW_AGENT:
            if data.get('request_data', None):
                ack_data['request_data'] = data.get('request_data')
            send_connector_operations_ack(ack_data, REMOTE_CONNECTOR_OPERATIONS.get('connector_healthcheck'), data.get('tenant'))
        else:
            publish_change_event(request, 'integration', 'healthcheck', 'connectors', ack_data, ack_data['timestamp'], skip_audit=True)
        return {'status': health_status, 'message': message}

    def agent_healthcheck(self, data, request=None):
        if not settings.LW_AGENT:
            return Response(status=status.HTTP_200_OK)
        response = {
            'agent': settings.SELF_ID,
            'version': settings.RELEASE_VERSION,
            'healthStatus': {
                'status': 'connected',
                'message': 'Agent is connected'
            },
            'timestamp': datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        }
        if data.get('request_data', None):
            response['request_data'] = data.get('request_data')
        send_connector_operations_ack(response, REMOTE_AGENT_OPERATIONS.get('agent_healthcheck'), data.get('tenant'))
