""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import ast
import json
import logging
import time

import requests
from django.conf import settings
from django.core.cache import cache
from xfutils.utils import api_call

from connectors.core.connector import ConnectorError
from connectors.core.constants import SSL_VALIDATION_ERROR, REQUEST_READ_TIMEOUT, CONNECTION_TIMEOUT, \
    INVALID_URL_OR_CREDENTIALS, UNAUTHORIZED
from connectors.core.constants import TIM_SKU_ENABLED
from integrations.crudhub import make_request

logger = logging.getLogger(__name__)

def api_health_check(url, method='GET', params=None, body='', headers=None, verify=False,
                     username='', password='', auth_config=None, *args, **kwargs):
    """
    This method can be used to make any generic api call to check health of the api.

    """
    return api_call(url, method, params, body, headers, None,
                    verify, username, password, auth_config, *args, **kwargs)


api_health_check.__str__ = lambda: 'Makes Simple Http API Call'


def update_connnector_config(connector_name=None, version=None, updated_config={}, configId=None, agent=None):
    from connectors.models import Configuration, Connector
    from connectors.serializers import ConnectorConfigurationSerializer
    try:
        if not updated_config:
            message = "Invalid input. Provide the configuration to be updated."
            raise ValueError(message)
        if Configuration.objects.filter(config_id=configId).exists():
            config_obj_instance = Configuration.objects.filter(config_id=configId)
            config_obj = config_obj_instance.first()
        elif connector_name and version:
            agent_id = agent if agent else settings.SELF_ID
            connector_object = Connector.objects.filter(name=connector_name, version=version, agent=agent_id).first()
            config_obj_instance = Configuration.objects.filter(connector=connector_object.id, default=True)
            config_obj = config_obj_instance.first()
        else:
            message = "Invalid inputs. Please provide the connector configuration id or the connector name and version for the configuration is to be updated"
            raise ValueError(message)
        config_serializer = ConnectorConfigurationSerializer(config_obj)
        configurations = config_serializer.data
        for key in ['name', 'status', 'config_id']:
            updated_config.pop(key, None)
        new_config = get_updated_config(configurations, updated_config, configId)
        config_obj_instance.update(config=new_config.get('config', configurations.get('config')))
    except ValueError as e:
        logger.exception(str(e))
        raise ConnectionError(str(e))
    except Exception as e:
        logger.exception("Error occurred while updating the configuration of connector {0} inline. ERROR :: {1}".format(
            connector_name, str(e)))
        raise ConnectionError(
            "Error occurred while updating the configuration of connector {0} inline. ERROR :: {1}".format(
                connector_name, str(e)))


def get_updated_config(configurations, updated_config, configId):
    prev_config = configurations.get('config', None)
    if prev_config:
        newly_added_configs = {k: v for k, v in updated_config.items() if k not in prev_config}
        removed_config = {k: v for k, v in prev_config.items() if k not in updated_config}
        update_config = {k: v for k, v in updated_config.items() if k in prev_config}
        newly_added_configs.update(removed_config)
        newly_added_configs.update(update_config)
        configurations['config'] = newly_added_configs
        return configurations
    else:
        raise ValueError("No configuration found to update. Please config_id or mark a configuration to default")

def get_info_from_cache(key=TIM_SKU_ENABLED):
    if cache.has_key(key):
        return cache.get(key)
    if key == TIM_SKU_ENABLED:
        url = settings.CRUD_HUB_URL + \
              '/api/auth/license/?param=license_details'
        retry_count = 1
        while retry_count < settings.CH_MAX_RETRY_COUNT:
            try:
                response = make_request(url=url, method='GET')
                logger.info('Successfully fetched SKU information from license server')
                is_tim_sku_enabled = response.get('details', {}).get('contracts', {}).get('has_tim_subscription', False)
                cache.set(TIM_SKU_ENABLED, is_tim_sku_enabled, timeout=7200)
                return cache.get(TIM_SKU_ENABLED)
            except Exception as e:
                logger.error('Error fetching SKU information from DAS. Retrying after 10 seconds.: %s', str(e))
                time.sleep(10)
            retry_count += 1
    return None
