""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import json
from django.db.models import Q
from rest_framework import serializers

from connectors.models import Connector, Operation, Configuration, Role, Team
from agents.models import Agent
from query.serializers import DynamicFieldsModelSerializer
from taggit.models import Tag
from taggit.serializers import (TagListSerializerField,
                                TaggitSerializer)


class RoleSerializer(serializers.ModelSerializer):
    class Meta:
        model = Role
        fields = '__all__'


class TeamSerializer(serializers.ModelSerializer):
    class Meta:
        model = Team
        fields = '__all__'


class DynamicFieldSerializer(serializers.ModelSerializer):
    def __init__(self, *args, **kwargs):
        context = kwargs.get('context', None)
        fields = None
        if context:
            fields = context.get('fields', None)

        super().__init__(*args, **kwargs)

        if fields is not None:
            allowed = set(fields)
            existing = set(self.fields)
            for field_name in existing - allowed:
                self.fields.pop(field_name)


class ConnectorListConfigurationSerializer(DynamicFieldSerializer):
    class Meta:
        model = Configuration
        fields = ('id', 'config_id', 'name', 'default')


class ConnectorListSerializer(serializers.ModelSerializer):
    configuration = ConnectorListConfigurationSerializer(many=True, read_only=True)

    class Meta:
        model = Connector
        fields = ('id', 'name', 'version', 'label', 'category', 'active', 'system', 'icon_small', 'iconLarge',
                  'description', 'config_count', 'status', 'install_result', 'configuration', 'ingestion_supported',
                  'tags', 'agent', 'remote_status', 'development', 'created', 'modified', 'metadata', 'rpm_full_name',
                  'help_online', 'publisher', 'type', 'latestAvailableVersion', 'installed')

    def to_representation(self, instance):
        ret = super().to_representation(instance)
        ret['category'] = ret['category'].strip('[]').split(',') if ret['category'] else []
        ret['rpm_installed'] = ret['metadata'].get('rpm_installed')
        ret.pop('metadata', None)
        return ret

class ConnectorConfigurationQuerySerializer(TaggitSerializer, DynamicFieldsModelSerializer):
    tags = TagListSerializerField()

    def to_representation(self, obj):
        obj = super().to_representation(obj)
        return obj

    class Meta:
        model = Configuration
        fields = '__all__'

class ConnectorConfigurationSerializer(TaggitSerializer, serializers.ModelSerializer):
    team_ids = serializers.PrimaryKeyRelatedField(many=True, queryset=Team.objects.all(), source='teams',
                                                  required=False, allow_null=True)
    tags = TagListSerializerField()
    class Meta:
        model = Configuration
        exclude = ('teams',)

    def to_representation(self, instance):
        ret = super().to_representation(instance)
        ret['teams'] = ret.pop('team_ids', [])
        return ret


class ConnectorConfigurationAgentSerializer(serializers.ModelSerializer):
    class Meta:
        model = Configuration
        fields = ('id', 'config_id', 'name', 'default', 'config', 'connector', 'status', 'agent')


class ConnectorOperationSerializer(serializers.ModelSerializer):
    role_ids = serializers.PrimaryKeyRelatedField(many=True, queryset=Role.objects.all(), source='roles',
                                                  required=False, allow_null=True)

    class Meta:
        model = Operation
        exclude = ('roles',)

    def to_representation(self, instance):
        ret = super().to_representation(instance)
        ret['roles'] = ret.pop('role_ids', [])
        metadata_fields = ret['metadata']
        for field, value in metadata_fields.items():
            if not ret.get(field, None):
                ret[field] = value
        ret.pop('metadata', None)
        return ret


class ConnectorLimitedSerializer(serializers.ModelSerializer):
    class Meta:
        model = Connector
        fields = ('name', 'version','system', 'remote_status', 'status', 'active', 'latestAvailableVersion', 'config_count')

class AgentConnectorSerializer(serializers.ModelSerializer):
    connector = serializers.SerializerMethodField('get_connector_serializer')
    #connectors = ConnectorLimitedSerializer(many=True, read_only=True)
    def __init__(self, *args, **kwargs):
        self.connector_name = kwargs.pop('connector_name')
        self.connector_version = kwargs.pop('connector_version', None)
        self.configured = kwargs.pop('configured', False)
        super().__init__(*args, **kwargs)

    class Meta:
        model = Agent
        fields = ('name', 'version','agentId', 'active', 'connector', 'isLocal', 'healthStatus')

    # def to_representation(self, instance):
    #     ret = super().to_representation(instance)
    #     connectors = ret.pop('connectors', [])
    #     if connectors:
    #         ret['connector'] = connectors[0]
    #     return ret

    def get_connector_serializer(self, obj):
        query_filter = {"name": self.connector_name, "agent_id": obj.agentId}
        if self.connector_version:
            query_filter.update({"version": self.connector_version})
        if self.configured:
            queryset = Connector.objects.filter(~Q(config_count=0), **query_filter).order_by('version').first()
        else:
            queryset = Connector.objects.filter(**query_filter).order_by('version').first()
        if queryset:
            serializer = ConnectorLimitedSerializer(queryset, read_only=True)
            return serializer.data


class ConnectorDetailSerializer(serializers.ModelSerializer):
    configuration = serializers.SerializerMethodField('get_config_serializer')
    operations = ConnectorOperationSerializer(many=True, read_only=True)

    def __init__(self, *args, **kwargs):
        self.teams = kwargs.pop('teams', [])
        super().__init__(*args, **kwargs)

    def get_config_serializer(self, obj):
        if isinstance(self.teams, str):
            self.teams = json.loads(self.teams)
        if self.teams:
            queryset = Configuration.objects.filter(Q(teams__isnull=True) | Q(teams__in=self.teams),
                                                    connector=obj).distinct()
            serializer = ConnectorConfigurationSerializer(queryset, many=True, read_only=True)
            return serializer.data
        else:
            queryset = Configuration.objects.filter(connector=obj)
            serializer = ConnectorConfigurationSerializer(queryset, many=True, read_only=True)
            return serializer.data

    def to_internal_value(self, data):
        category = data.get('category', None)
        if isinstance(category, list):
            data['category'] = (',').join(data['category']) if data['category'] else None
        return super().to_internal_value(data)

    def to_representation(self, instance):
        ret = super().to_representation(instance)
        ret['category'] = ret['category'].strip('[]').split(',') if ret['category'] else []
        ret.pop('configuration_old', None)
        ret.pop('migrate', None)
        metadata_fields = ret['metadata']
        for field, value in metadata_fields.items():
            if not ret.get(field, None):
                ret[field] = value
        ret.pop('metadata', None)
        return ret

    class Meta:
        model = Connector
        fields = '__all__'
        read_only_fields = ()

class TagSerializer(DynamicFieldsModelSerializer):
    class Meta:
        model = Tag
        fields = ('name',)