""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import copy
import importlib
import inspect
import json
import logging
import jmespath
import datetime

import os
import sys
import base64
import re
import shlex
import shutil
import tarfile
import tempfile
import uuid
import zipfile
from glob import glob
import subprocess
from packaging.version import Version

import requests
import simplejson
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import Q
from django.http.request import HttpRequest
from django_filters import FilterSet, CharFilter
from rest_framework import status
from urllib3.exceptions import InsecureRequestWarning
from django_tenants.utils import schema_context

from agents.models import Agent
from connectors.core.base_connector import Connector, ConnectorError
from connectors.core.constants import SEALAB_SECRET_KEY, REMOTE_STATUS, REMOTE_CONNECTOR_OPERATIONS
from connectors.environment import _expand
from connectors.models import Configuration as ConfigurationModel, Team, Role
from connectors.models import Connector as ConnectorModel
from connectors.models import Operation as OperationModel
from connectors.serializers import ConnectorDetailSerializer, ConnectorConfigurationSerializer, \
    ConnectorOperationSerializer
from constants.error_message import *
from contenthub.models import ContentHub
from data_import.models import DataImport
from integrations.crudhub import make_request
from postman.core.publisher import Publisher
from postman.utils.constants import execute_action_ack
from xfutils.audit import publish_change_event
from xfutils.utils import maybe_json_or_raise, api_call
from xfutils.passwordModule import manage_password
from postman.utils.utility import get_peer_publish_action, get_tenant_publish_action

from django.db.models import Count, Q, F
from taggit.models import Tag
from taggit.managers import TaggableManager


logger = logging.getLogger('connectors')


def execute_cmd(cmd, use_shell=False):
    if not use_shell:
        cmd = shlex.split(cmd)
    process = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=use_shell)
    ret = process.returncode
    std_out = process.stdout.strip().decode("utf-8")
    std_err = process.stderr.strip().decode("utf-8")
    return (ret, std_out, std_err)


def reload_connector_modules(root_dir, connector_name):
    connector_path = os.path.join(root_dir, connector_name)
    connector_module = connector_name
    for dir in os.listdir(connector_path):
        file_details = os.path.splitext(dir)
        connector_module_file = '%s.%s' % (connector_module, file_details[0])
        if file_details[1] == ".py":
            file_module = importlib.import_module(connector_module_file)
            file_module = importlib.reload(file_module)


def get_connector(item, version, info_json=None):
    try:
        if settings.CONNECTOR_DEVELOPMENT_VERSION_POSTFIX in version:
            root_dir = settings.DEV_CONNECTOR_INSTALLATION_DIR
        else:
            root_dir = settings.CONNECTOR_INSTALLATION_DIR

        # connector_name = '%s_%s' % (item, version.replace('.', '_'))
        connector_name = '%s' % (item)

        if root_dir == settings.DEV_CONNECTOR_INSTALLATION_DIR:
            connector_path = settings.DEV_CONNECTOR_INSTALLATION_DIR
            sys.path.append(connector_path)
            reload_connector_modules(root_dir, connector_name)
        else:
            connector_path = settings.INTEGRATION_DIR
            sys.path.append(connector_path)

        connector_module_path = connector_name

        connector_module = importlib.import_module('.connector', '%s.%s' % (settings.CONNECTOR_FILES_PATH, connector_module_path))
        
        for name, obj in inspect.getmembers(connector_module):
            if inspect.isclass(obj) and issubclass(obj, Connector) and not inspect.isabstract(obj):
                try:
                    connector_instance = obj(info_json=info_json)
                except:
                    connector_instance = obj()
                return {'instance': connector_instance}
    except Exception as e:
        return {'instance': None, 'message': str(e)}


def get_connector_instance(conn_name, conn_version, conn_obj=None):
    info_json = {}
    if conn_obj:
        connector_serializer = ConnectorDetailSerializer(conn_obj)
        info_json = connector_serializer.data
    connector = get_connector(conn_name, conn_version, info_json)
    if connector.get('instance'):
        return connector.get('instance')
    else:
        logger.exception(
            'Exception occurred while getting the connector instance ERROR :: {0}'.format(connector.get('message')))
        raise ConnectorError(connector.get('message'))
    
def get_connector_obj(name, version, agent_id):
        connector_qs = ConnectorModel.objects.filter(name=name, version=version, agent_id=agent_id, active=True)
        if not connector_qs.exists():
            version = get_connector_version_or_latest(name, version, agent_id=agent_id)
            connector_qs = ConnectorModel.objects.filter(name=name, version=version, agent_id=agent_id, active=True)

        return connector_qs.first()


def get_connector_version_or_latest(name, version, status='installed', agent_id=None):
    if not agent_id: agent_id = settings.SELF_ID
    if status:
        connector_instance = ConnectorModel.objects.filter(name=name, version=version, status=status,
                                                           agent_id=agent_id)
    else:
        connector_instance = ConnectorModel.objects.filter(name=name, version=version, agent_id=agent_id)
    if not connector_instance.exists() and (version and settings.CONNECTOR_DEVELOPMENT_VERSION_POSTFIX in version):
        raise ConnectorError("Connector {0}  with version {1} doesn't exists".format(name, version))
    if not connector_instance.exists():
        version_latest = None
        if status:
            connectors = ConnectorModel.objects.filter(name=name, status='installed', agent_id=agent_id,
                                                       development=False)
        else:
            connectors = ConnectorModel.objects.filter(name=name, agent_id=agent_id, development=False)
        for each_connector in connectors:
            version_current = each_connector.version
            if not version_latest:
                version_latest = version_current
            else:
                if tuple(map(int, (version_current.split(".")))) > tuple(map(int, (version_latest.split(".")))):
                    version_latest = version_current
        if version_latest:
            version = version_latest
    return version


def get_connector_or_latest(name, version, status='installed', agent_id=None):
    if not agent_id:
        agent_id = settings.SELF_ID
    if status:
        connector_instance = ConnectorModel.objects.filter(name=name, version=version, status='installed',
                                                           agent=agent_id)
    else:
        connector_instance = ConnectorModel.objects.filter(name=name, version=version, agent=agent_id)
    if not connector_instance.exists():
        version_latest = None
        connector_instance_latest = {}
        if status:
            connectors = ConnectorModel.objects.filter(name=name, status='installed', agent=agent_id)
        else:
            connectors = ConnectorModel.objects.filter(name=name, agent=agent_id)
        for each_connector in connectors:
            version_current = each_connector.version
            if not version_latest:
                version_latest = version_current
                connector_instance_latest = each_connector
            else:
                if tuple(map(int, (version_current.split(".")))) > tuple(map(int, (version_latest.split(".")))):
                    version_latest = version_current
                    connector_instance_latest = each_connector
        if version_latest:
            connector_instance = connector_instance_latest
    else:
        connector_instance = connector_instance.first()
    return connector_instance


def get_connector_path(name, version):
    connector_root_dir = settings.CONNECTOR_INSTALLATION_DIR
    if settings.CONNECTOR_DEVELOPMENT_VERSION_POSTFIX in version:
        connector_root_dir = settings.DEV_CONNECTOR_INSTALLATION_DIR
    if not os.path.exists(connector_root_dir):
        os.mkdir(connector_root_dir)
    return os.path.join(connector_root_dir, name)


def call_connectors_on_app_start_func(conn_id=None):
    try:
        from connectors.models import Connector
        filter = {'status': 'Completed', 'agent_id': settings.SELF_ID}
        if conn_id:
            filter.update({'id': conn_id})
        connectors = Connector.objects.filter(**filter)
        for connector in connectors:
            try:
                logger.info('Calling on_app_start() for connector %s', connector.name)
                conn_instance = get_connector_instance(connector.name, connector.version)
                config = get_configuration(conn_id=connector.id, decrypt=True,
                                           config_schema=connector.config_schema, convert_to_dict=True)
                conn_instance.on_app_start(config, connector.active)
            except Exception as e:
                logger.warn('Error while calling connector %s on_app_start function: %s', connector.name,
                            str(e))
    except:
        pass


def iterate_password_on_change_fields(each_field, fields_with_type_pwd, max_depth):
    if max_depth == 0:
        return fields_with_type_pwd
    for key, values in each_field.get('onchange').items():
        for each_on_change_field in values:
            if each_on_change_field.get('onchange'):
                max_depth = max_depth - 1
                iterate_password_on_change_fields(each_on_change_field, fields_with_type_pwd, max_depth)
            if each_on_change_field.get('type') == 'password':
                fields_with_type_pwd.append(each_on_change_field.get('name'))
    return fields_with_type_pwd


def encrypt_password(config_schema, config):
    fields_with_type_pwd = []
    for each_field in config_schema.get('fields', []):
        if each_field.get('onchange'):
            fields_with_type_pwd = iterate_password_on_change_fields(each_field, fields_with_type_pwd, max_depth=20)
        if each_field.get('type') == 'password':
            fields_with_type_pwd.append(each_field.get('name'))
    fields_with_type_pwd = [*set(fields_with_type_pwd)]
    for field in fields_with_type_pwd:
        # for update call if password is NULL string it means there is no change in password field
        if config.get(field) and config.get(field) != 'NULL':
            config[field] = manage_password(data=config[field], action='encrypt')


def decrypt_password(config, config_schema):
    fields_with_type_pwd = []
    for each_field in config_schema.get('fields', []):
        if each_field.get('onchange'):
            fields_with_type_pwd = iterate_password_on_change_fields(each_field, fields_with_type_pwd, max_depth=20)
        if each_field.get('type') == 'password':
            fields_with_type_pwd.append(each_field.get('name'))
    fields_with_type_pwd = [*set(fields_with_type_pwd)]
    for field in fields_with_type_pwd:
        if config.get(field):
            config[field] = manage_password(data=config[field], action='decrypt')


def omit_password(result):
    configurations = result.get('configuration', [])
    config_schema = result.get('config_schema', {})
    fields_with_type_pwd = [each_field.get('name') for each_field in config_schema.get('fields', []) if
                            each_field.get('type') == 'password']
    if isinstance(configurations, list):
        for v in configurations:
            config = v.get('config', {})
            for field in fields_with_type_pwd:
                if field in config:
                    config[field] = 'NULL'
    else:
        config = configurations.get('config', {})
        for field in fields_with_type_pwd:
            if field in config:
                config[field] = 'NULL'


def update_pwd_fields(new_config, old_config, config_schema):
    fields_with_type_pwd = [each_field.get('name') for each_field in config_schema.get('fields', []) if
                            each_field.get('type') == 'password']

    for field in fields_with_type_pwd:
        # if the password field received from UI is NULL take the saved password in DB
        # else take the received password
        if new_config.get(field) and new_config.get(field) == 'NULL':
            new_config[field] = manage_password(data=old_config[field], action='decrypt')
        elif new_config.get(field):
            new_config[field] = manage_password(data=new_config[field], action='decrypt')


def on_connector_delete(conn_name, conn_version, config):
    connector_instance = get_connector_instance(conn_name, conn_version)
    if conn_name == "phishing-classifier" and Version(conn_version) < Version('1.1.0'):
        # Workaround for phishing classifier bug which deletes workspace on uninstall
        return
    else:
        connector_instance.teardown(config)


def on_active_status_update(conn_name, conn_version, old_status, new_status, config):
    conn_instance = get_connector_instance(conn_name, conn_version)
    if new_status:
        conn_instance.on_activate(config)
    else:
        conn_instance.on_deactivate(config)


def _process_images(connector_path, image_file):
    image_path = os.path.join(connector_path, 'images', image_file)
    with open(image_path, 'rb') as image_file:
        image = 'data:image/jpeg;base64,' + (base64.b64encode(image_file.read())).decode()
        return image


def process_images(connector_path, data):
    icon_small, icon_large = '', ''
    try:
        if data.get('icon_small_name'):
            icon_small = _process_images(connector_path, data.get('icon_small_name'))
    except Exception as e:
        logger.warn('Error while saving icon_small jpeg image %s', str(e))
    try:
        if data.get('icon_large_name'):
            icon_large = _process_images(connector_path, data.get('icon_large_name'))
    except Exception as e:
        logger.warn('Error while saving icon_small jpeg image %s', str(e))

    return icon_small, icon_large


def get_parsed_config(config, config_schema={}):
    '''
    This function will parse if config is in the form of jinja template.
    :param config:
    :return:
    '''
    try:
        field_type_mapping = {each_field.get('name'): each_field.get('type') for each_field in
                              config_schema.get('fields', [])}
        return {key: _expand({}, value, field_type_mapping.get(key)) for key, value in config.items()}
    except Exception as e:
        logger.error('Error while evaluating jinja template for config %s', str(e))
        raise e


def get_previous_version_details(info, replace=True):
    # this will be used to import the previous version config in new version if config schema is same
    try:
        operations_role_mapping = {}
        prev_config = []
        latest_version = get_connector_version_or_latest(info.get('name'),
                                                         info.get('version'))
        if latest_version and ConnectorModel.objects.filter(name=info.get('name'), version=latest_version,
                                                            status='installed', agent_id=settings.SELF_ID).exists():
            connector_detail = \
                ConnectorModel.objects.filter(name=info.get('name'), version=latest_version,
                                              agent_id=settings.SELF_ID).first()
            connector_serializer = ConnectorDetailSerializer(connector_detail)
            connector_detail = connector_serializer.data
            operations = connector_detail.get('operations')
            if bool(info.get('configuration')):
                configuration = connector_detail.get('configuration')
                prev_config = import_prev_config(info.get('configuration'), connector_detail.get('config_schema'),
                                                 configuration, connector_detail.get('id'), replace)
            if operations:
                for operation in operations:
                    operations_role_mapping[operation.get('operation')] = operation.get('roles')
        return prev_config, operations_role_mapping
    except Exception as e:
        logger.error("Error occurred while importing the previous configuration ERROR :: {0}".format(str(e)))
    return {}


def symlink_system_connector(connector_name, connector_version):
    create_symlink = True
    # adding the development filter here because when updating a system connector for which there is a 
    # _dev version, the _dev version was causing an error here which led to the connector being stuck
    # in an "In Progress" state and never moving to "Completed"
    connectors = ConnectorModel.objects.filter(name=connector_name, development=False)
    if len(connectors) > 0:
        for each_connector in connectors:
            version_current = each_connector.version
            if tuple(map(int, (version_current.split(".")))) > tuple(map(int, (connector_version.split(".")))):
                create_symlink = False
                break

    if not create_symlink:
        return

    symlink_dir = os.path.join(settings.CONNECTORS_DIR, connector_name)
    # remove previous symlink if any
    try:
        os.unlink(symlink_dir)
    except:
        logger.info('No previous symlink for the connector: %s' % symlink_dir)

    # create the new symlink
    connector_path = get_connector_path(connector_name, connector_version)
    if settings.LW_AGENT:
        logger.info('Creating the symlink for agent')
        os.symlink(connector_path, symlink_dir)
        return
    #
    # From 7.2.2, docker enterprise support is added. Before to 7.2.2, when system connector gets installed,
    # it used to create the symlink in /opt/cyops-integration/integration/connectors. Since from specfile
    # of system connectors, manage.py is run as 'sudo -u nginx', symlink owner is set to 'nginx'. Since nginx
    # was creating symlink in /opt/cyops-integration/integration/connectors, ACL permission was required to this
    # dir. Because of the ACL, it was causing issue on docker having backing filesystem as 'xfs', while restoring
    # these symlinks. Hence create symlinks using 'fsr-integration' user and remove the ACL from
    # /opt/cyops-integration/integration/connectors for 'nginx'. User 'nginx' don't need ACL on mentioned dir,
    # as connector installation, uninstallation is done at dir /opt/cyops/configs/integration/connectors from 7.2.2.
    # When user is 'nginx', it will be creating symlink as 'sudo -u fsr-integration' from integration-sdk.
    #

    home_dir = os.path.basename(os.getenv('HOME'))
    if home_dir != 'nginx':
        os.symlink(connector_path, symlink_dir)
        return
    validate_cmd_arg(connector_path)
    validate_cmd_arg(symlink_dir)
    cmd = "sudo -u {0} /bin/ln -fs {1} {2}".format(settings.USER_INTEGRATIONS, connector_path, symlink_dir)
    execute_cmd(cmd)


def install_connector_packages(connector_name, connector_version, raise_ex=False):
    file = get_connector_path(connector_name, connector_version)
    file = os.path.join(file, 'requirements.txt')
    logger.debug("Installing: " + file)
    if not os.path.exists(file):
        logger.error("File does not exist. Returning: " + file)
        return
    pip_path = f"{os.path.dirname(sys.executable)}/pip"
    command = [pip_path, 'install', '-r', file, '--disable-pip-version-check']
    my_env = os.environ.copy()
    my_env['PATH'] = '/usr/sbin:/sbin:' + my_env['PATH']
    try:
        validate_cmd_arg(file)
        install_process = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=my_env)
    except Exception as e:
        logger.exception('Error in installing the packages of connector {0} v{1} ERROR:: {2}'.format(connector_name,
                                                                                                     connector_version,
                                                                                                     str(e)))
        if raise_ex:
            raise e
    logger.info('Successfully installed file')


def get_configuration_instance(config_id, conn_id=None):
    if isinstance(config_id, list):
        tags_query = Q()
        for tag in config_id:
            tags_query |= Q(tags__name=tag) if not conn_id else Q(connector=conn_id, tags__name=tag)

        config_instance = ConfigurationModel.objects.annotate(
            matching_tags=Count('tags', filter=tags_query, distinct=True)
        ).filter(matching_tags=len(config_id)).annotate(total_tags=Count('tags', distinct=True)
                                                        ).filter(total_tags=len(config_id))
    else:
        if not conn_id:
            config_instance = ConfigurationModel.objects.filter(
                Q(config_id=config_id) | Q(name=config_id) | Q(tags__name=config_id)).distinct()
        else:
            config_instance = ConfigurationModel.objects.filter(
                Q(connector=conn_id, config_id=config_id) | Q(connector=conn_id, name=config_id) | Q(connector=conn_id,
                                                                                                     tags__name=config_id)).distinct()

    if (len(config_instance) > 1):
        message = xf_integration_48.format(config_id)
        logger.error(message)
        raise ConnectorError(message)
    logger.info("Connector configuration found with input : {0}".format(config_id))
    return config_instance.first()


def get_configuration(conn_id=None, config_id=None, decrypt=False, config_schema={}, convert_to_dict=False,
                      config_instance=None):
    try:
        config_object = []
        many = False

        if not bool(config_schema and config_schema.get('fields', [])) and config_id:
            return {}
        elif not config_schema or not config_schema.get('fields', []):
            return []

        if config_instance:
            pass
        elif config_id:
            config_instance = get_configuration_instance(config_id, conn_id)
            if not config_instance:
                config_instance = ConfigurationModel.objects.filter(connector=conn_id, default=True).first()
        else:
            many = True
            config_instance = ConfigurationModel.objects.filter(connector=conn_id)

        if not many:
            if not config_instance:
                message = xf_integration_47.format(config_id)
                logger.error(message)
                raise ConnectorError(message)
            if config_instance.status == 0:
                message = "The configuration {0} is only partially setup. Cannot be used till the configuration is complete"
                logger.error(message.format(config_instance.name))
                raise ConnectorError(message.format(config_instance.name))

        if config_instance:
            serializer = ConnectorConfigurationSerializer(config_instance, many=many)
            config_object = serializer.data

        if isinstance(config_object, list):
            dict_config = {}
            for v in config_object:
                config = v.get('config', {})
                if decrypt:
                    decrypt_password(config, config_schema)
                config = get_parsed_config(config, config_schema)
                config['config_id'] = v.get('config_id')
                config['name'] = v.get('name')
                config['status'] = v.get('status', 0)
                if convert_to_dict:
                    dict_config[config['config_id']] = config
            if convert_to_dict:
                config_object = dict_config

        else:
            config = config_object.get('config', {})
            if decrypt:
                decrypt_password(config, config_schema)
            config = get_parsed_config(config, config_schema)
            config['config_id'] = config_object.get('config_id')
            config['name'] = config_object.get('name')
            config['status'] = config_object.get('status', 0)
            config_object = config

        return config_object

    except ConnectorError as e:
        raise ConnectorError(str(e))
    except Exception as e:
        logger.exception(
            "Error occurred while retrieving configuration with config id : {0} ERROR :: {1} ".format(config_id,
                                                                                                      str(e)))
        raise ConnectorError("Error occurred while retrieving configuration with config id : {0} ".format(config_id))


def get_operation(name=None, conn_id=None):
    try:
        operation_instance = OperationModel.objects.get(connector=conn_id, operation=name, enabled=True)
        serializer = ConnectorOperationSerializer(operation_instance)
        return serializer.data
    except Exception as e:
        raise ConnectorError(
            "Error occurred while retrieving the connector operation {0} ERROR :: {1}".format(name, str(e)))


def config_type_check(name, type, schema):
    for param in schema:
        if name == param.get('name', ''):
            if not type == param.get('type', ''):
                return False
            break
    return True


def get_nested_config_schema(new_onchange_value, param_value, param_name, old_schema):
    new_nested_config = new_onchange_value.get(str(param_value),
                                               new_onchange_value.get(str(param_value).lower(), []))
    new_nested_schema = {"fields": new_nested_config}
    old_nested_config = []
    for old_schema_value in old_schema:
        if old_schema_value.get('name', None) == param_name:
            old_onchange_value = old_schema_value.get('onchange', None)
            if old_onchange_value:
                old_nested_config = old_onchange_value.get(str(param_value),
                                                           old_onchange_value.get(str(param_value).lower(), []))
            break
    old_nested_schema = {"fields": old_nested_config}
    return new_nested_schema, old_nested_schema


def import_prev_config(new_schema, old_schema, configuration, conn_id, replace=True):
    new_schema = new_schema.get('fields', {})
    old_schema = old_schema.get('fields', {})
    for config_object in configuration:
        valid_config = {}
        config = config_object.get('config', {})
        config_status = 1
        for param in new_schema:
            param_name = param.get('name', '')
            param_value = config.get(param_name, None)
            same_type = config_type_check(param_name, param.get('type'), old_schema)
            new_onchange_value = param.get('onchange', None)
            if new_onchange_value:
                new_nested_schema, old_nested_schema = get_nested_config_schema(new_onchange_value, param_value,
                                                                                param_name, old_schema)
                valid_nested_config = import_prev_config(new_nested_schema, old_nested_schema, [config_object], conn_id)
                config_status = valid_nested_config[0].get('status')
                valid_config.update(valid_nested_config[0].get('config'))
            if param_value is None:
                if param.get('value', param.get('default', None)):
                    valid_config[param_name] = param.get('value', param.get('default'))
                elif param.get('required'):
                    config_status = 0
            else:
                if same_type:
                    valid_config[param_name] = param_value
                elif param.get('required'):
                    config_status = 0

        config_object['config'].update(valid_config)
        config_object['status'] = config_status
        config_object['connector'] = conn_id
        config_object['team_ids'] = config_object.pop('teams', [])
        if not replace or not config_object.get('config_id', None):
            config_object['config_id'] = str(uuid.uuid4())
    return configuration


def schema_validator(schema, input):
    missing_required_keys = []
    for param in schema:
        if param.get('required', False):
            param_name = param.get('name', '')
            param_value = input.get(param_name, None)
            param_title = param.get('title', param_name)
            if param_value or isinstance(param_value, int):
                if isinstance(param_value, str):
                    if not param_value.strip():
                        missing_required_keys.append(param_title)
            else:
                missing_required_keys.append(param_title)
    return missing_required_keys


def is_operation_param_valid(operation_schema=None, params={}):
    param_schema = operation_schema.get('parameters', None)
    if not params:
        params = {}
    if param_schema:
        if not isinstance(params, dict):
            raise ConnectorError(
                "Invalid Inputs :: Params provided for a action are of invalid format, Expected in dictionary format")

        missing_required_title = schema_validator(param_schema, params)
        if len(missing_required_title):
            raise ConnectorError(
                "Invalid params provided :: Parameters {0} have either blank value or not provided".format(
                    ', '.join(missing_required_title)))
    return params


def validate_config(config, config_schema={}):
    if not config_schema and config:
        error_message = "Invalid configuration :: You are trying to insert configuration for the connector with no connfiguration schema"
        logger.error(error_message)
        raise ConnectorError(error_message)

    missing_required_keys = schema_validator(config_schema.get('fields', []), config)

    if len(missing_required_keys):
        error_message = "Invalid configuration params provided :: Parameters {0} have either blank value or not provided".format(
            ', '.join(missing_required_keys))
        logger.error(error_message)
        raise ConnectorError(error_message)


def delete_events(connector_name, connector_version, tenant):
    try:
        if not settings.LW_AGENT:
            event_response = api_call(
                'https://' + settings.WF_HOST + '/api/workflow/events/{0}/{1}/?format=json'.format(
                    connector_name, connector_version),  method='DELETE', body={},
                verify=False, tenant=tenant, appliance_auth=True, all_config=settings.ALL_CONFIG)
            return event_response
    except Exception as e:
        error_message = str(e)
        error_message = get_truncate_message(error_message, 200)
        logger.warn('Error while deleting events: %s'%error_message)


def import_events(connector_path, name, version, *args, **kwargs):
    label = kwargs.get('label')
    image = kwargs.get('image')
    if not settings.LW_AGENT:
        image = _process_images(connector_path, image)
        events_path = os.path.join(connector_path, 'events', 'events.json')
        if not os.path.exists(events_path):
            logger.warn('No Events to import.')
            return []
        try:
            with open(events_path) as json_data:
                events_data = json.load(json_data)
                if events_data is not None and events_data:
                    # Add connector attributes to each event object
                    for event in events_data:
                        event["connector_name"] = name
                        event["connector_label"] = label
                        event["connector_version"] = version
                        event["connector_icon"] = image
        except:
            logger.info('%s Events Import: Error while reading events.json.', name)
            return
        result = []
        try:
            if not settings.LW_AGENT:
                event_response = api_call('https://' + settings.WF_HOST + '/api/workflow/bulk/events/?format=json',
                                          method='POST', body=events_data, verify=False, tenant=kwargs.get('tenant'),
                                          appliance_auth=True, all_config=settings.ALL_CONFIG)
                logger.debug(f'event_response {event_response}')
                result = [{'name': item.get('name'),
                           'description': item.get('description'),
                           '@id': item.get('@id'),
                           'image': item.get('image')
                           } for item in event_response.get('persisted', [])
                          ]
        except Exception as e:
            error_message = str(e)
            error_message = get_truncate_message(error_message, 200)
            logger.warning('%s Event Import: Error while importing events: %s', name, error_message)
        return result


def import_playbook(connector_path, name, label, version, image, override_playbook_info=True, tenant=None):
    if not settings.LW_AGENT:
        playbooks_path = os.path.join(
            connector_path,
            'playbooks',
            'playbooks.json'
        )
        if not os.path.exists(playbooks_path):
            logger.warn('No playbooks to import.')
            return []
        try:
            with open(playbooks_path) as json_data:
                playbook_data = json.load(json_data)
                # ToDo enable after the tags model is made in sealab
                # playbook_exported_tags = playbook_data.get('exported_tags', [])
                playbook_data = playbook_data.get('data')
        except:
            logger.info('%s Playbook Import: Error while reading playbook.json.', name)
            return
        # upload image for PB collection
        try:
            if image and not settings.LW_AGENT:
                image = _process_images(connector_path, image)
                for el in playbook_data:
                    el['image'] = image
        except Exception as e:
            error_message = str(e)
            error_message = get_truncate_message(error_message, 200)
            logger.warn('%s Playbook Import: Error while uploading image for playbook collection %s', name,
                        error_message)

        # Standardise Playbook collection name and description
        if playbook_data and len(playbook_data) == 1 and override_playbook_info:
            playbook_data[0]['name'] = 'Sample - %s - %s' % (label, version)
            playbook_data[0][
                "description"] = 'Sample playbooks for "%s" connector. If you are planning to use any of the sample playbooks in your environment, ensure that you clone those playbooks and move them to a different collection, since the sample playbook collection gets deleted during connector upgrade and delete.' % label

        # ToDo enable after the tags model is made in sealab
        # if len(playbook_exported_tags):
        #     playbook_tags = []
        #     for each_tag in playbook_exported_tags:
        #         playbook_tags.append({'uuid': each_tag})
        #     try:
        #         if not settings.LW_AGENT: make_request('/api/3/bulkupsert/tags', 'POST',
        #                                                {"__data": playbook_tags, "__replace": False,
        #                                                 "__unique": ["uuid"]})
        #     except Exception as e:
        #         logger.warn('%s Playbook Tags Import: Error while importing playbook tags: %s', name, str(e))

        result = []
        try:
            if not settings.LW_AGENT:
                playbook_collection = api_call(
                    'https://' + settings.WF_HOST + '/api/workflow/playbook-collections/?format=json',
                    method='POST', body=playbook_data, verify=False, tenant=tenant,
                    appliance_auth=True, all_config=settings.ALL_CONFIG)
                result = [{'name': item.get('name'),
                           'description': item.get('description'),
                           '@id': item.get('@id'),
                           'image': item.get('image')
                           } for item in playbook_collection.get('persisted', [])
                          ]
        except Exception as e:
            error_message = str(e)
            error_message = get_truncate_message(error_message, 200)
            logger.warning('%s Playbook Import: Error while importing playbooks: %s', name, error_message)
        return result


def import_functions(name, version, operations=[]):
    return []
    if not settings.LW_AGENT:
        fun_import_status = ['Unknown']
        is_contain_function = __import_functions_from_info_json_operations(name, version, operations, fun_import_status)
        is_contain_function_ref = __import_functions_from_functions_directory(name, version, fun_import_status)
        return is_contain_function, is_contain_function_ref, fun_import_status[0]
    return False, False, 'Unknown'


def __import_functions_from_info_json_operations(connector_name, version, operations, fun_import_status) -> bool:
    functions_payload = {'functions': []}
    is_contain_function = False
    for operation in operations:
        if operation.get('include_as_function', False):
            function_category = operation.get('function_category')
            function_param = ""
            for params in operation.get('parameters', []):
                function_param = function_param + params['name'] + "={{params." + params['name'] + "}},"
            if function_param != "":
                function_param = function_param[:-1]

            functions_payload['functions'].append({
                'name': function_category if function_category else "custom_ops",
                'title': function_category if function_category else "Custom",
                'description': '',
                'operations': [
                    {
                        'name': operation.get('title', ''),
                        'jexp': operation['operation'] + "(" + function_param + ")",
                        'description': operation.get('description', ''),
                        'connector_name': connector_name,
                        'action_name': operation.get('operation', ''),
                        'version': version,
                        'params': operation.get('parameters', [])
                    }
                ]
            })
    if functions_payload['functions']:
        is_contain_function = __save_functions(functions_payload, connector_name, fun_import_status)
    return is_contain_function


def __import_functions_from_functions_directory(name, version, fun_import_status) -> bool:
    default_functions_dir = os.path.join(
        get_connector_path(name, version),
        'functions'
    )
    if not os.path.exists(default_functions_dir):
        logger.debug('No functions to import.')
        return False

    functions_payload = {'functions': []}
    for filename in glob(default_functions_dir + '**/*.json'):
        try:
            with open(filename) as json_data:
                model_data = json.load(json_data)
                model_data['name'] = model_data.get('category_name')
                model_data['title'] = model_data.get('category_title')
                for operation in model_data.get('operations', []):
                    operation['connector_name'] = name
                    operation['version'] = version
                functions_payload['functions'].append(model_data)
        except Exception as e:
            fun_import_status[0] = 'Failed'
            logger.error('Error, while forming payload for functions: Error {0}'.format(str(e)))
            return False
    return __save_functions(functions_payload, name, fun_import_status)


def __save_functions(functions, name, fun_import_status) -> bool:
    try:
        headers = {'Content-Type': 'application/json'}
        response = make_request('/api/workflow/expressions/bulk', 'POST',
                                body=functions, headers=headers, validate_response=False)
        if response.status_code == status.HTTP_207_MULTI_STATUS:
            fun_import_status[0] = 'Partial_Success'
            json_response = response.json(strict=False)
        else:
            maybe_json_or_raise(response)
            if fun_import_status[0] == 'Unknown':
                fun_import_status[0] = 'Success'
        return True
    except Exception as e:
        fun_import_status[0] = 'Failed'
        error_message = str(e)
        error_message = get_truncate_message(error_message, 200)
        logger.warn('%s Functions import: Error while importing functions: %s', name, error_message)
        return False


def get_truncate_message(value, chr_limit):
    if not type(value) == str:
        value = str(value)
    if len(value) > chr_limit:
        return value[:chr_limit] + '...'
    return value


def connector_actions_serializer(connector, connector_fields, permissions=[]):
    connector_object = {}
    roles = []
    if permissions:
        connector_object['has_permissions'] = True
        roles_query = connector.roles.all().values('uuid')
        for roles_query_each in roles_query:
            roles.append(roles_query_each.get('uuid'))
        if roles and not list(set(roles) & set(permissions)):
            connector_object['has_permissions'] = False

    for field in connector_fields:
        connector_object[field] = getattr(connector, field, None)
    return connector_object


def insert_connector(info, connector_path='', playbook_collections=[], config={}, isReserved=False,
                     raise_exception=True,
                     prev_operation_role={}, *args, **kwargs):
    # validate file and delete if invalid.
    ack_call = kwargs.get('ack_call', False)
    connector_update = False
    try:
        conn_instance = ConnectorModel.objects.filter(name=info.get('name', None), version=info.get('version', None),
                                                      agent_id=info.get('agent', settings.SELF_ID)).first()
        if conn_instance:
            connector_update = True
        else:
            conn_instance = None
        if not ack_call:
            icon_small, icon_large = process_images(connector_path, info)
        else:
            icon_small = info.get('icon_small')
            icon_large = info.get('icon_large')
        connector_fields = [f.name for f in ConnectorModel._meta.get_fields()]
        connector_data = {}
        info_copy = copy.deepcopy(info)
        for field in connector_fields:
            value = info.get(field, None)
            if value or isinstance(value, bool):
                if field == 'configuration':
                    if value.get('fields'):
                        connector_data['config_schema'] = value
                    else:
                        connector_data['config_schema'] = {}
                        connector_data['config_count'] = -1
                elif field == 'operations':
                    continue
                else:
                    connector_data[field] = value
                info_copy.pop(field, None)
            elif field == 'configuration':
                info_copy.pop(field, None)
                connector_data['config_schema'] = {}
                connector_data['config_count'] = -1

        if conn_instance and connector_data.get('tags'):
            connector_data['tags'].extend(conn_instance.tags)
        connector_data['system'] = isReserved
        connector_data['icon_small'] = icon_small
        connector_data['iconLarge'] = icon_large
        connector_data['playbook_collections'] = playbook_collections
        connector_data['metadata'] = info_copy
        connector_data['migrate'] = True
        connector_data['active'] = True
        connector_data['status'] = 'In-Progress' if settings.LW_AGENT else 'installed'
        connector_data['install_result'] = {}
        connector_data['tenant'] = None
        connector_data['requirements_installed'] = 'In-Progress' if settings.LW_AGENT else 'installed'
        connector_data['category'] = ",".join(connector_data['category']) if isinstance(connector_data['category'],
                                                                                        list) else connector_data[
            'category']
        agent_query = Agent.objects.filter(agentId=connector_data.get('agent', settings.SELF_ID))
        if not agent_query.exists():
            agent_query = Agent.objects.filter(agentId=settings.SELF_ID)
        agent_instance = agent_query.first()
        agent_id = agent_instance.agentId
        connector_data['agent'] = agent_id
        connector_data['install_result']['message'] = 'Success'
        connector_data['install_result']['output'] = ''
        fun_import_status = kwargs.get('fun_import_status', 'Unknown')
        if fun_import_status != 'Unknown':
            connector_data['install_result']['functions_status'] = fun_import_status
        # Only set remote-status here for non-agent installs. This is because the agent still 
        # has to go through install_connector_files()
        if not settings.LW_AGENT:
            remote_status = {'status': 'finished', 'message': 'Success', 'output': ''}
            connector_data['remote_status'] = remote_status

        if connector_update and conn_instance:
            serializer = ConnectorDetailSerializer(conn_instance, data=connector_data, partial=True)
        else:
            if ack_call and not connector_data.get('latestAvailableVersion'):
                # Connectors installed during Agent spinup will not have this attribute set
                try:
                    contenthub_obj = ContentHub.objects.get(name=connector_data['name'])
                    connector_data['latestAvailableVersion'] = contenthub_obj.version
                except ContentHub.DoesNotExist as e:
                    # this will be the case during a custom connector install on Agent
                    pass
            serializer = ConnectorDetailSerializer(data=connector_data)
        serializer.is_valid(raise_exception=True)
        conn_instance = serializer.save()

        if bool(config):
            if isinstance(config, list):
                config_count = len(config)
                for each_config in config:
                    each_config['connector'] = conn_instance.id
                    each_config['agent'] = agent_id
                configuration_serializer = ConnectorConfigurationSerializer(data=config, many=True)
            else:
                config_count = 1
                configuration_serializer = ConnectorConfigurationSerializer(
                    data={
                        'config_id': config.get('config_id'),
                        'name': config.get('name', ''),
                        'default': config.get('default', False),
                        'config': config.get('config'),
                        'connector': conn_instance.id,
                        'agent': agent_id,
                        'tags': config.get('tags', [])
                    }
                )
            configuration_serializer.is_valid(raise_exception=True)
            config_instance = configuration_serializer.save()
            serializer.update(conn_instance, {'config_count': conn_instance.config_count + config_count})

            if kwargs.get('restore_ingestion_mapping'):
                if not isinstance(config_instance, list):
                    config_instance = [conn_instance]
                for config_obj in config_instance:
                    try:
                        data_import = DataImport.objects.filter(previous_config_id=config_obj.config_id).first()
                        data_import.previous_config_id = None
                        data_import.configuration = config_obj
                        data_import.save()
                    except Exception as e:
                        logger.error("Error while updating data ingestion for config {0} ERROR: {1}".format(
                            config_obj.config_id, str(e)))
        operations = []
        operation_fields = [f.name for f in OperationModel._meta.get_fields()]
        for operation in info_copy.get('operations', []):
            operation.pop('id', None)
            operation_object = {}
            if prev_operation_role:
                operation_name = operation.get('operation')
                roles = prev_operation_role.get(operation_name, [])
                operation_object['role_ids'] = roles
            for field in operation_fields:
                value = operation.get(field, None)
                if value or isinstance(value, bool):
                    operation_object[field] = value
                    operation.pop(field, None)
            operation_object['connector'] = conn_instance.id
            operation_object['metadata'] = operation
            operations.append(operation_object)

        if connector_update: OperationModel.objects.filter(connector=conn_instance.id).delete()

        if operations:
            operation_serializer = ConnectorOperationSerializer(data=operations, many=True)
            operation_serializer.is_valid(raise_exception=True)
            operation_serializer.save()
        result = serializer.data
        if kwargs.get('omit_password', True):
            omit_password(result)
        return result

    except ConnectorError as e:
        raise ConnectorError(str(e))

    except Exception as exp:
        # clean up when went wrong.
        if not raise_exception:
            if conn_instance and connector_update:
                install_result = {}
                install_result['message'] = 'Connector Insertion Failed'
                install_result['output'] = get_truncate_message(exp, 2000)
                remote_status = install_result
                remote_status['status'] = 'failed'
                serializer.update(conn_instance, {'status': 'Failed', 'install_result': install_result,
                                                  'remote_status': remote_status})
            elif conn_instance and not connector_update:
                conn_instance.delete()
            logger.exception('{0} ERROR :: {1}'.format(xf_integration_3, str(exp)).format(info.get('name')))
            raise ConnectorError('{0} ERROR :: {1}'.format(xf_integration_3, str(exp)).format(info.get('name')))
        else:
            raise ConnectorError(exp)


def broadcast_to_remote(data, agent_id, action=None, callback_name=None, method=None):
    destination_id = agent_id
    pub_config = get_tenant_publish_action(callback_name, destination_id)
    publisher = Publisher(pub_config)
    if action:
        data.update({'action': action})
    logger.info("Publishing to agent: " + str(destination_id))
    try:
        message = json.dumps(data)
    except Exception as e:
        try:
            message = simplejson.dumps(data)
        except Exception as e:
            message = json.dumps(data, cls=DjangoJSONEncoder)
    publisher.publish(message, destination_id)


def broadcast_message(data, action=None, callback_name=None, method=None, agent_obj=None):
    if not agent_obj:
        logger.info('Broadcast message: Agent obj does not exist. Returning')
        return False
    elif agent_obj.isLocal:
        if not settings.MASTER_ID and Agent.objects.filter(name='Master').exists():
            settings.MASTER_ID = Agent.objects.get(name='Master').agentId
        destination_id = settings.MASTER_ID
    else:
        destination_id = agent_obj.agentId

    if not destination_id:
        logger.info('Broadcast: No destination id found returning, current MASTER ID: %s', settings.MASTER_ID)
        return False
    # force publish- is to publish some instruction that does not consists to remote even if
    # self agent has not allowed remote operation. Eg: publish instruction about allow remote operation
    # switch.
    __force_publish = True if data and data.get('__force_publish') else False
    if not (agent_obj.allowRemoteOperation or __force_publish):
        logger.info('Agent has not allowed remote connector operation.')
        if not agent_obj.isLocal: raise PermissionDenied(xf_integration_16)
        return False
    # When remote connector management is enabled at tenant allow broadcast of all data except when an action is
    # executed for self
    if not data.get('request_id') and action in [execute_action_ack]:
        return False
    pub_config = get_peer_publish_action(callback_name, destination_id)

    message_json = {
        "data": data,
        "status": data.get('status', {}),
        "sourceId": settings.SELF_ID,
        "destinationId": destination_id,
        "action": action,
        "method": method
    }

    try:
        message = json.dumps(message_json)
    except Exception as e:
        try:
            message = simplejson.dumps(message_json)
        except Exception as e:
            message = json.dumps(message_json, cls=DjangoJSONEncoder)

    publisher = Publisher(pub_config)
    publisher.publish(message, destination_id)
    logger.info('Broadcast message: Broadcast Successful for destination %s', destination_id)
    # clean up the file downloaded at agent
    return True


def is_replace(connector_name, compatible_version=None):
    if not compatible_version:
        return True

    compatible_version = compatible_version.strip()

    if compatible_version == 'ALL':
        return True

    if compatible_version == 'NULL':
        return False

    if not re.compile(r"[0-9]+.[0-9]+.[0-9]+$").match(compatible_version):
        raise Exception('Incorrect Compatible Version Information')

    connectors = ConnectorModel.objects.filter(name=connector_name)
    if (connectors.count()) > 0:
        compatible_version_map = tuple(map(int, (compatible_version.split("."))))
        for connector in connectors:
            version = connector.version
            if tuple(map(int, (version.split(".")))) < compatible_version_map:
                return False
        return True
    raise Exception('Failed to validate against installed connectors')


def file_cleanup(env, *args, **kwargs):
    file_dir = settings.TMP_FILE_ROOT
    for file_path, metadata in env.get('files', {}).items():
        try:
            if os.path.isdir(file_path):
                shutil.rmtree(file_path)
            else:
                os.remove(os.path.join(file_dir, file_path))
        except Exception as e:
            logger.error(str(e))


def update_env_files(env, *args, **kwargs):
    for file_path, metadata in env.get('files', {}).items():
        metadata['agent_id'] = settings.SELF_ID


def zip_folder(self, destination_zip_file_path, source_zip_folder_path):
    zip_obj = zipfile.ZipFile(destination_zip_file_path + '.zip', 'w', zipfile.ZIP_DEFLATED)
    root_len = len(source_zip_folder_path) + 1
    for base, dirs, files in os.walk(source_zip_folder_path):
        for file in files:
            fn = os.path.join(base, file)
            zip_obj.write(fn, fn[root_len:])


def check_permission(config_data, operation_data, input_data):
    secret_key = input_data.get('secret_key', '')
    input_roles = input_data.get('rbac_info', {}).get('roles', [])
    input_teams = input_data.get('rbac_info', {}).get('teams', [])
    if isinstance(input_roles, str):
        input_roles = json.loads(input_roles)
    if isinstance(input_teams, str):
        input_teams = json.loads(input_teams)
    try:
        roles = operation_data.get("roles", [])
        teams = config_data.get('teams', [])

        # Checking if it is sealab call by SEALAB_SECRET_KEY
        # Allowed if no roles or team assigined to operation or config
        # Raise error when no role or team found in input role or team
        if roles and not secret_key == SEALAB_SECRET_KEY:
            if not list(set(roles) & set(input_roles)):
                raise Exception("Action {0} is not accessible".format(operation_data.get('operation')))
        if teams and not secret_key == SEALAB_SECRET_KEY:
            if not list(set(teams) & set(input_teams)):
                raise Exception("Configuration {0} is not accessible".format(config_data.get('name')))
    except Exception as e:
        raise Exception(
            "Error occurred while checking the permission for action and configuration ERROR:: {0}".format(str(e)))


def get_or_create_team_or_role(record_type, record_data):
    try:
        record_ids = []
        if record_type.lower() == 'role':
            record_object = Role
        elif record_type.lower() == 'team':
            record_object = Team
        else:
            return record_ids

        if record_data:
            for data in record_data:
                if isinstance(data, dict):
                    uuid = data.get("uuid", data.get("@id"))
                else:
                    uuid = data
                if '/api/3/' in uuid:
                    uuid = uuid.split('/')[-1]
                record_query_set = record_object.objects.filter(uuid=uuid)
                if not record_query_set.exists():
                    data = {"uuid": uuid}
                    team_instance = record_object(**data)
                    team_instance.save()
                else:
                    team_instance = record_query_set.first()
                record_ids.append(team_instance.uuid)
        return record_ids
    except Exception as e:
        logger.error(xf_integration_25.format(record_data, str(e)))
        raise ConnectorError(xf_integration_25.format(record_data, str(e)))


def get_update_list(list_items=[], add_items=[], remove_items=[]):
    for add_item in add_items:
        if not add_item in list_items:
            list_items.append(add_item)
    for remove_item in remove_items:
        if remove_item in list_items:
            list_items.remove(remove_item)
    return list_items


def get_rbac_info(request):
    # Todo we need to remove this depency
    rbac_info = request.data.get('rbac_info', {})

    # if not rbac_info:
    #     try:
    #         url = settings.CRUD_HUB_URL + '/api/3/actors/current?$relationships=true'
    #         auth = request.headers.get('forwarded-authorization') if hasattr(request, 'headers') else None
    #         request_header = {'Authorization': auth}
    #         response = requests.request('GET', url, headers=request_header, json=None, verify=False, )
    #         user_info = maybe_json_or_raise(response)
    #         source = request.headers.get('X-Forwarded-For') or request.headers.get('X-Remote-Addr')
    #         rbac_info = {
    #             'user_iri': user_info.get('uuid'),
    #             'source': source,
    #             'teams': user_info.get('teams'),
    #             'roles': user_info.get('roles'),
    #         }
    #     except Exception as e:
    #         logger.warn('Error occurred while retrieving user info ERROR :: {0}'.format(str(e)))
    return rbac_info


def is_connector_installed(connector_name):
    # No need sudoers permission for /bin/rpm as nginx can run the query on rpm.
    cmd = "/bin/rpm -q " + connector_name + " &>/dev/null"
    ret = os.system(cmd)
    if ret != 0:
        return False
    return True


def identify_if_dependencies_installed(conn_name, conn_version, connector_obj):
    try:
        get_connector_instance(conn_name, conn_version, connector_obj)
        return True
    except ImportError:
        return False
    except Exception as e:
        return False


def delete_workflow_collections(collection_uuids):
    try:
        body = {
            "logic": "AND",
            "filters": [
                {
                    "field": "uuid",
                    "operator": "in",
                    "value": collection_uuids,
                    "type": "object"
                }
            ]
        }
        make_request('/api/3/delete-with-query/workflow_collections', 'DELETE', body=body)
    except Exception as e:
        logger.error('Error while deleting pb collection for collection uuid: %s', collection_uuids)


def delete_functions_associated_with_connector(name, version):
    try:
        url = '/api/workflow/expressions/bulk?connector_name=' + name + '&version=' + version
        make_request(url, 'DELETE', body=None)
    except Exception as e:
        logger.error('Error while deleting functions for connector : %s-%s', name, version)


def get_ingestion_settings(connector_instance, upgrade_in_progress):
    data_ingestion = []
    if upgrade_in_progress:
        config_query = ConfigurationModel.objects.filter(connector=connector_instance)
        data_import = DataImport.objects.filter(configuration__in=config_query)
        for di_obj in data_import:
            data_ingestion.append(copy.deepcopy(di_obj))
    return data_ingestion


def untar_connector_tgz(filepath, conn_name, conn_version='latest'):
    modes = ['r:gz', 'r:', 'r:bz2', 'r:xz']
    error_message = ''
    if not filepath:
        return get_connector_path(conn_name, conn_version)
    tar_extract_dir = os.path.join(tempfile.gettempdir(), 'connectors')
    if not os.path.exists(tar_extract_dir):
        os.mkdir(tar_extract_dir, 0o755)
    for mode in modes:
        try:
            tar = tarfile.open(filepath, mode)
            os.makedirs(tar_extract_dir, exist_ok=True)
            tar.extractall(path=tar_extract_dir)
            tar.close()
            connector_path = os.path.join(tar_extract_dir, conn_name)
            return connector_path
        except Exception as e:
            logger.warn(f"Failed to extract with mode {mode}: {e}")
            error_message = str(e)

    err_msg = "Error while extracting the connector tgz {0}".format(conn_name)
    logger.error("{0} ERROR : {1}".format(err_msg, error_message))
    raise ConnectorError(err_msg)


def get_validated_info_file(connector_path, replace, name, version, isReserved=False, validate=True):
    # traves the file path and validate the content
    connector_files = os.listdir(connector_path)
    # Check for multiple info.json and connector files
    info_count = [filename for filename in connector_files if filename == 'info.json']
    if len(info_count) != 1:
        raise ConnectorError(
            'Invalid connector structure :: info file of connector are either multiple or not provided.')
    connector_count = [filename for filename in connector_files if filename == 'connector.py']
    if len(connector_count) != 1:
        raise ConnectorError(
            'Invalid connector structure :: connector file of connector are either multiple or not provided.')

    # Validate the info content
    with open(os.path.join(connector_path, 'info.json'), 'rb') as f:
        info_content = json.loads(f.read())
        if validate:
            validate_info(info_content, name, version)
        return info_content


def validate_info(info_content, input_name, input_version):
    # Validate Name and version
    name = info_content.get('name', '')
    version = info_content.get('version', '')
    if not re.compile(r'[0-9]+.[0-9]+.[0-9]+$').match(version):
        raise ConnectorError(xf_integration_28)

    if not re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*').match(name):
        raise ConnectorError(xf_integration_27)
    if name != input_name:
        raise ConnectorError(xf_integration_53)
    if version != input_version:
        raise ConnectorError(xf_integration_54)

    # TODO: check in the database instead
    # if not replace and os.path.exists(
    #         get_connector_path(name, version)):
    #     raise ConnectorError('Connector conflict :: Connector with same name is already active')

    # Validate config fields
    config_names = [each_field.get('name') for each_field in info_content.get('configuration', {}).get('fields', [])]
    if 'default' in config_names or 'name' in config_names:
        err_msg = "Reserve keywords [default, name]  is used in configuration fields/param "
        logger.error(err_msg)
        raise ConnectorError(err_msg)


def import_connector(connector_info, replace, isReserved=False, rbac_info={}, *args,
                     **kwargs):
    install_source = kwargs.get('install_source', "")
    connector_path = kwargs.get('connector_path', "")
    config = kwargs.get('config', dict())
    prev_operation_role = dict()

    if not config:
        try:
            # Get the prev connector details in case of upgrade
            config, prev_operation_role = get_previous_version_details(connector_info, replace)
        except Exception as e:
            logger.warning('ERROR :: {0}'.format(e))

    # If Replace, remove all existing connectors
    if replace:
        # ToDo check for reserve connector should be able to delete it on replcae true
        # if connector_info.get('name') in settings.CONNECTORS_RESERVED and install_source == INSTALL_SOURCE_API_UPLOAD:
        # To avoid deletion of reserve connectors
        # When installing through upload tgz
        # raise Exception(cs_integration_20)
        connectors = ConnectorModel.objects.filter(name=connector_info.get('name'), agent_id=settings.SELF_ID,
                                                   status='installed', development=False)
        for connector in connectors:
            connector.metadata.update({'upgrade_in_progress': True})
            connector.save(update_fields=['metadata'])
            data = {
                "name": connector.name,
                "version": connector.version,
                "agent": settings.SELF_ID,
                "tenant": kwargs.get('tenant'),
                "request_data": kwargs.get('request_data')
            }
            delete_connector(data)

    try:
        # Importing sample playbooks of connector
        # Todo remove the dep of token and generate the new token from app keys
        playbook_collections = import_playbook(connector_path,
                                               connector_info.get('name'),
                                               connector_info.get('label'),
                                               connector_info.get('version'),
                                               connector_info.get('icon_large_name'),
                                               connector_info.get('override_playbook_info', True),
                                               tenant=kwargs.get('tenant')
                                               )
        # Importing of Events
        import_events(connector_path, connector_info.get('name'), connector_info.get('version'),
                      label=connector_info.get('label'),
                      image=connector_info.get('icon_small_name'), tenant=kwargs.get('tenant'))
        # Database base entry for all details of connector
        result = insert_connector(info=connector_info, connector_path=connector_path,
                                  playbook_collections=playbook_collections, config=config, isReserved=isReserved,
                                  prev_operation_role=prev_operation_role, agent=kwargs.get('agent'), omit_password=kwargs.get('omit_password', True))
        return result
    except Exception as e:
        raise ConnectorError("Error while inserting connector ERROR: %s" % str(e))


def install_connector_files(connector_name, connector_version, prev_connector_path):
    connector_path = get_connector_path(connector_name, connector_version)
    try:
        shutil.rmtree(connector_path)
    except:
        # just a failsafe delete in case the directory existed from a previous failed attempt
        pass
    # rename version to folder.
    shutil.move(prev_connector_path, connector_path)

    # create symlink for system connector to latest version
    # TODO: check for workspace flag in case of cloning connectors
    if connector_name in settings.CONNECTORS_RESERVED:  # and not workspace_connector:
        logger.info('Symlinking system connector...')
        symlink_system_connector(connector_name, connector_version)
        logger.info('Symlink created')

    if settings.LW_AGENT:
        logger.info('Getting Connector object...')
        conn_instance = ConnectorModel.objects.filter(name=connector_name, version=connector_version,
                                                      agent=settings.SELF_ID).first()
        conn_data = {'status': 'installed', 'remote_status': {"status": REMOTE_STATUS.get('finished')}}
        serializer = ConnectorDetailSerializer(conn_instance, data=conn_data, partial=True)
        serializer.is_valid(raise_exception=True)
        conn_instance = serializer.save()
        logger.info('Connector object saved')

    logger.info('Installing connector packages...')
    install_connector_packages(connector_name, connector_version, raise_ex=True)
    logger.info('Packages installed')


def set_connector_file_permissions(connector_name, connector_version):
    connector_path = get_connector_path(connector_name, connector_version)
    # Set the connector directory permissions appropriately. This is 
    # useful for rpm agents where custom connectors may get created without
    # execute privileges
    os.chmod(connector_path, 0o755)
    for root, dirs, files in os.walk(connector_path):
        for d in dirs:
            os.chmod(os.path.join(root, d), 0o755)
        for f in files:
            os.chmod(os.path.join(root, f), 0o755)


def delete_connector(data, request=None):
    connector_name = data.get('name')
    connector_version = data.get('version')
    conn_instance = ConnectorModel.objects.filter(name=connector_name, version=connector_version,
                                                  agent=settings.SELF_ID).first()
    if conn_instance:
        conn_instance.delete()
        timestamp = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        data.update({"timestamp": timestamp, "status": REMOTE_STATUS.get('uninstall_finished')})
        if settings.LW_AGENT:
            try:
                connector_path = get_connector_path(data.get('name'), data.get('version'))
                shutil.rmtree(connector_path)
            except FileNotFoundError:
                logger.warning(f"Filepath {connector_path} not found")
            send_connector_operations_ack(data, REMOTE_CONNECTOR_OPERATIONS.get('connector_uninstall'),
                                          data.get('tenant'))
        else:
            delete_events(data.get('name'), data.get('version'), data.get('tenant'))
            if request:
                publish_change_event(request, 'integration', 'uninstall', 'connectors', data, timestamp)
        return


def send_connector_operations_ack(data, operation, tenant):
    url = f"https://{settings.ALL_CONFIG.get('application', 'integration_host')}/api/integration/connector-ack/{operation}/"
    logger.info(f'Sending Connector operations Ack request...')
    ack_response = api_call(url, method='POST', body=data, verify=False, auth_type='agent_appliance',
                            appliance_auth=True, all_config=settings.ALL_CONFIG, tenant_id=settings.TENANT_ID)


def download_connector_tgz(name, version):
    connector_base_dir = os.path.join(tempfile.gettempdir(), 'connectors')
    if not os.path.exists(connector_base_dir):
        os.mkdir(connector_base_dir, 0o755)
    base_url = settings.CONNECTOR_REPO_BASE_URL
    folder = name + '-' + str(version) + '/latest'
    filename = name + '.tgz'
    url = base_url + folder + '/' + filename
    requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)
    file_response = requests.get(url=url, timeout=120, verify=False)
    connector_tgz_path = os.path.join(connector_base_dir, 'conn_%s%s' % (
        name, settings.CONNECTORS_EXTENSION))
    with open(connector_tgz_path, 'wb') as f:
        f.write(file_response.content)
        f.close()
    return connector_tgz_path


def reconstruct_request_object(request_data):
    class MyRequestObject:
        def __init__(self, tenantuuid, tenantid, useruuid, username, subtenantid, meta, headers, teams):
            self.tenantuuid = tenantuuid
            self.tenantid = tenantid
            self.useruuid = useruuid
            self.username = username
            self.subtenantid = subtenantid
            self.META = meta
            self.headers = headers
            self.teams = teams


def create_request_from_ack_payload(request_data):
    new_request = HttpRequest()
    for k, v in request_data.items():
        setattr(new_request, k, v)
    return new_request


def extract_audit_data_from_request(request):
    relevant_attributes = {
        "tenantuuid": request.tenantuuid,
        "tenantid": request.tenantid,
        "useruuid": request.useruuid,
        "username": request.username,
        "subtenantid": request.subtenantid,
        # "META": request.meta,
        # "headers": request.headers,
        "teams": request.teams
    }
    relevant_attributes["META"] = {
        "REMOTE_ADDR": request.META.get("REMOTE_ADDR")
    }
    relevant_attributes["headers"] = {
        "Websocket-Session-Id": request.headers.get("Websocket-Session-Id")
    }
    return relevant_attributes


def get_connector_list_data(data):
    # Todo do intersection for upgrade and other data
    return data
    if not data:
        return data
    if isinstance(data, list):
        names = jmespath.search("[*].name", data)
    else:
        names = [data.get('name')]
    with ((schema_context('public'))):
        content_hub_query = ContentHub.objects.filter(name__in=names).values('name', 'version')
        for content_hub in content_hub_query:
            if not isinstance(data, list):
                if data.get('name') == content_hub.get('name'):
                    data['latestAvailableVersion'] = content_hub.get('version') if data.get(
                        'version') != content_hub.get('version') else None
                    break
            for each_data in data:
                if each_data.get('name') == content_hub.get('name'):
                    each_data['latestAvailableVersion'] = content_hub.get('version') if each_data.get(
                        'version') != content_hub.get('version') else None
                    break
    return data


def update_connector(connector, configurations=[]):
    name = connector.get('name')
    version = connector.get('version')
    system = connector.get('system')
    playbook_collections = connector.get('playbook_collections')
    connector_path = get_connector_path(name, version)
    info_path = os.path.join(connector_path, 'info.json')
    try:
        with open(info_path) as json_file:
            info = json.load(json_file)
            if not info.get('name') == name or not info.get('version') == version:
                logger.error('Name or version mismatched in info file and database')
                raise ConnectorError('Name or version mismatched in info file and database')
    except Exception as e:
        logger.error('Error occurred while retrieving the info json ERROR :: {0}'.format(str(e)))
        raise ConnectorError('Error occurred while retrieving the info json ERROR :: {0}'.format(str(e)))

    insert_connector(info, playbook_collections, configurations, isReserved=system)


def validate_agent_health(agent_id):
    agent_object = Agent.objects.get(agentId=agent_id)
    if not agent_object.active:
        msg = f'Agent {agent_id} is not active'
    elif agent_object.healthStatus['status'] != 'connected':
        msg = f'Agent {agent_id} is not in a connected state'
    else:
        return agent_object
    raise ConnectorError({'message': msg})

def validate_cmd_arg(argument, prefix=None, suffix=None, pattern="^[a-zA-Z0-9-_\/\.]*$"):
    if prefix and not argument.startswith(prefix):
        raise Exception(f"Argument {argument} is does not match required prefix {prefix}")
    if suffix and not argument.endswith(suffix):
        raise Exception (f"Argument {argument} does not match required suffix {suffix}")
    if not re.fullmatch(pattern, argument):
        raise Exception(f"Argument {argument} does not match required pattern {pattern}")


class ConnectorConfigurationFilter(FilterSet):
    class Meta:
        model = ConfigurationModel
        fields = ['id', 'name', 'config_id', 'connector', 'agent', 'default', 'tags']
        filter_overrides = {
            TaggableManager: {
                'filter_class': CharFilter,
                'extra': lambda f: {
                    'lookup_expr': 'icontains',
                },
            },
        }
