""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import os.path
from django.conf import settings
from django.db.models import Q
from django.http import HttpResponse
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters
from rest_framework.mixins import ListModelMixin
from rest_framework.parsers import FileUploadParser
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet, ModelViewSet
from packaging.version import Version

from agents.utils.connector_utils import is_connector_installable_on_agent
from connectors.connector_ack_request import ack_operations
from connectors.core.connector import CustomConnectorException, ConnectorError
from connectors.decorators import json_response
from connectors.helper import *
from connectors.models import Operation, Configuration
from connectors.permissions import IsAgentAuthenticated
from connectors.serializers import ConnectorDetailSerializer, ConnectorListSerializer, ConnectorOperationSerializer, \
    ConnectorListConfigurationSerializer, AgentConnectorSerializer, TagSerializer
from connectors.utils import (broadcast_message, identify_if_dependencies_installed, get_configuration, get_operation,
                              is_operation_param_valid, get_connector_or_latest, get_or_create_team_or_role,
                              get_connector_version_or_latest, check_permission, get_parsed_config, get_rbac_info,
                              update_env_files, ConnectorConfigurationFilter)
from constants.error_message import *
from postman.core.exceptions import MessagePublishFailed
from tenant.models import Tenant
from utils.connector_op_utility import get_executable_class
from .utils import get_configuration_instance, get_connector_instance, get_connector_obj
from xfutils.query.filtersets import get_query
from contenthub.models import ContentHub
from taggit.models import Tag


class ConnectorList(ModelViewSet):
    serializer_class = ConnectorListSerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    ordering_fields = ('id', 'name', 'label')
    search_fields = ('$id', '$label')
    filter_fields = ('id', 'name', 'category', 'active', 'ingestion_supported', 'status', 'system', 'agent')
    ordering = ('id',)

    def get_queryset(self):
        filters = {"development": False}
        agent_id = self.request.GET.get('agent')
        if agent_id:
            filters.update({"agent_id": agent_id})
        isLocal = self.request.GET.get('agent__isLocal')
        if isLocal:
            filters.update({"agent__isLocal": isLocal == 'true'})
        queryset = Connector.objects.filter(**filters).defer('iconLarge', 'configuration',
                                                                                        'playbook_collections')
        if 'configured' in self.request.GET:
            configured = self.request.GET.get('configured')
            if configured.lower() == 'true':
                queryset = self.filter_queryset(
                    queryset=Connector.objects.filter(~Q(config_count=0), **filters))
            elif configured.lower() == 'false':
                queryset = self.filter_queryset(queryset=Connector.objects.filter(config_count=0, **filters))
        if CONNECTOR_TAG_FILTER in self.request.GET:
            tags = self.request.GET.get(CONNECTOR_TAG_FILTER)
            queryset = Connector.objects.filter(tags__contains=tags.split(','), **filters)
        return self.filter_queryset(queryset)

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        unique_filter = ['name', 'version']
        if request.GET.get('unique_by'):
            unique_filter = request.GET.get('unique_by').split(',')
        queryset = queryset.order_by('name', '-version').distinct(*unique_filter)
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def connector_list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        filter_data = request.data
        filter_data.update({"distinct": {"field": ['name', 'version']}})
        queryset = get_query(filter_data, queryset)
        page = self.paginate_queryset(queryset)
        if page is not None:
            if filter_data.get('aggregates'):
                return Response(page)
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def connector_action_details(self, request, *args, **kwargs):
        connector_name = self.kwargs.get('connector', None)
        connector_agent = request.GET.get('agent', settings.SELF_ID)
        connector_version = self.kwargs.get('version', None)
        operation_name = request.GET.get('operation', None)
        exclude = request.GET.get('exclude', '').split(',')
        teams = request.data.get('rbac_info', {}).get('teams', [])
        roles = request.data.get('rbac_info', {}).get('roles', [])
        connector_object = {}
        connector_fields = ['name', 'version', 'label', 'iconLarge', 'config_count', 'agent_id', 'publisher',
                            'cs_approved']
        operation_fields = ['operation', 'title', 'visible', 'parameters']
        if isinstance(teams, str):
            teams = json.loads(teams)
        if isinstance(roles, str):
            roles = json.loads(roles)
        connector_querset = Connector.objects.filter(name=connector_name, version=connector_version,
                                                     agent_id=connector_agent,
                                                     development=False)
        if not connector_querset.exists():
            error_message = 'No matching connector by name {0} version {1} exists.'.format(connector_name,
                                                                                           connector_version)
            return Response({"message": error_message}, status=status.HTTP_404_NOT_FOUND)

        connector = connector_querset.first()
        if operation_name:
            operation = connector.operations.filter(operation=operation_name).first()
            operation_data = ConnectorOperationSerializer(operation).data
            if not list(set(operation_data.get('roles')) & set(roles)) and operation_data.get('roles'):
                operation_data['has_permissions'] = False
            else:
                operation_data['has_permissions'] = True

            connector_object['operations'] = [operation_data]

        if 'configuration' not in exclude:
            config_query = connector.configuration.filter(Q(teams__isnull=True) | Q(teams__in=teams)).distinct()
            config_serializer = ConnectorConfigurationSerializer(config_query, many=True)
            connector_object['configurations'] = config_serializer.data

        for field in connector_fields:
            connector_object[field] = getattr(connector, field, None)

        return Response({"data": connector_object, "status": "Success"}, status=status.HTTP_200_OK)

    def connector_actions(self, request, *args, **kwargs):
        connector_agent = request.GET.get('agent', settings.SELF_ID)
        configured = request.GET.get('configured', 'all')
        ingestion_supported = True if request.GET.get('ingestion_supported') and request.GET.get(
            'ingestion_supported').lower() == 'true' else False
        active = True if request.GET.get('active') and request.GET.get('active').lower() == 'true' else False
        exclude = request.GET.get('exclude', '').split(',')
        export = True if request.GET.get('$export') and request.GET.get('$export').lower() == 'true' else False
        connector_fields = ['name', 'version', 'label', 'icon_small', 'config_count', 'agent_id', 'active', 'status']
        operation_fields = ['operation', 'title', 'visible']
        connectors_data = []
        roles = request.data.get('rbac_info', {}).get('roles', [])
        teams = request.data.get('rbac_info', {}).get('teams', [])
        connector_filter = {'development': False}
        distinct_filter = []
        if connector_agent and connector_agent != 'all':
            connector_filter.update({'agent_id': connector_agent})
        else:
            distinct_filter = ['name', 'version']
        if active:
            connector_filter.update({'active': True})
        if ingestion_supported:
            connector_filter.update({'ingestion_supported': True})
            connector_fields.append('iconLarge')
        if isinstance(teams, str):
            teams = json.loads(teams)
        if isinstance(roles, str):
            roles = json.loads(roles)

        if configured.lower() == 'true':
            connectors = Connector.objects.filter(~Q(config_count=0), **connector_filter)
            configured = True
        elif configured.lower() == 'false':
            connectors = Connector.objects.filter(config_count=0, **connector_filter)
            configured = False
        else:
            connectors = Connector.objects.filter(**connector_filter)
            configured = False

        if distinct_filter:
            connectors = connectors.distinct(*distinct_filter)

        for connector in connectors:
            connector_object = {}
            if connector.id:
                connector_object['id'] = int(connector.id)
            if configured or 'configuration' not in exclude:
                config_query = connector.configuration.filter(Q(teams__isnull=True) | Q(teams__in=teams)).distinct()
                if not export:
                    config_serializer = ConnectorListConfigurationSerializer(config_query, many=True)
                else:
                    config_serializer = ConnectorConfigurationSerializer(config_query, many=True)
                connector_object['configuration'] = config_serializer.data

            if 'operation' not in exclude:
                allowed_operation_query = connector.operations.filter(
                    Q(roles__isnull=True) | Q(roles__in=roles)).distinct()
                allowed_operations = ConnectorOperationSerializer(allowed_operation_query, many=True).data
                allowed_operation_ids = []
                for allowed_operation in allowed_operations:
                    allowed_operation['has_permissions'] = True
                    allowed_operation_ids.append(allowed_operation.get('id'))
                denied_operation_query = connector.operations.filter(
                    Q(roles__isnull=False) & ~Q(roles__in=roles) & ~Q(id__in=allowed_operation_ids)).distinct()
                denied_operations = ConnectorOperationSerializer(denied_operation_query, many=True).data
                for denied_operation in denied_operations:
                    denied_operation['has_permissions'] = False
                connector_object['operations'] = allowed_operations + denied_operations

            if export:
                connector_serializer_data = ConnectorListSerializer(connector).data
                connector_serializer_data.pop('configuration', None)
                connector_object.update(connector_serializer_data)
            else:
                for field in connector_fields:
                    connector_object[field] = getattr(connector, field, None)
            if connector_object:
                connectors_data.append(connector_object)

        return Response({"data": connectors_data, "status": "Success"}, status=status.HTTP_200_OK)


class ConnectorOperationView(ModelViewSet):
    model = Operation
    queryset = Operation.objects.all()
    serializer_class = ConnectorOperationSerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    ordering_fields = ('operation',)
    search_fields = ('$operation',)
    filter_fields = ('id', 'operation')

    def operation_output_schema(self, request, *args, **kwargs):
        try:
            connector_name = self.kwargs.get('connector', None)
            connector_version = self.kwargs.get('version', 'latest')
            operation_name = request.data.get('operation', None)
            config_id = request.data.get('config', 'get_default_config')
            agent_id = request.data.get('agent')
            config_instance = None

            if not config_id == 'get_default_config':
                config_instance = Configuration.objects.filter(config_id=config_id).first()
                if config_instance:
                    agent_id = config_instance.agent_id

            if not agent_id:
                agent_id = settings.SELF_ID

            # Getting the connector with the given version or latest if given version not found
            # with respect to given agent or self if agent not provided.
            connector = get_connector_or_latest(connector_name, connector_version, agent_id=agent_id)
            if not connector:
                logger.warn('Connector with name {1} version {0} or latest version not found.'.format(connector_name,
                                                                                                      connector_version))
                return Response({"data": {"output_schema": {}}, "status": "Success"}, status=status.HTTP_200_OK)

            operation = get_operation(name=operation_name, conn_id=connector.id)

            if not config_instance and connector.config_count > -1:
                config_instance = Configuration.objects.get(connector=connector.id, default=True, agent_id=agent_id)

            api_output_schema = operation.get('api_output_schema', None)
            if api_output_schema:
                # For dynamic output schema by calling connector action
                output_schema = {}
                if not agent_id == settings.SELF_ID:
                    logger.warn("API based output schema is not supported for an agent configuration")
                else:
                    params = request.data.get('params', {})
                    input_data = {
                        "connector": connector_name,
                        "version": connector.version,
                        "params": params,
                        "operation": api_output_schema,
                        "config": config_instance.config_id if config_instance else None
                    }
                    try:
                        response, is_binary = ConnectorExecute.execute_connector_operation(input_data)
                        output_schema = response.get("data", {})
                    except Exception as e:
                        error_message = xf_integration_9.format(operation_name, connector_name, connector_version)
                        logger.warn("{0} ERROR: {1}".format(error_message, str(e)))
                response_data = {"output_schema": output_schema}
            elif operation.get('conditional_output_schema'):
                # For conditional output schema defined in info.json file
                response_data = {
                    "conditional_output_schema": operation.get('conditional_output_schema')
                }
            else:
                # For static output schema defined in info.json file
                response_data = {
                    "output_schema": operation.get('output_schema', {})
                }
            return Response({"data": response_data, "status": "Success"}, status=status.HTTP_200_OK)
        except Exception as e:
            error_message = xf_integration_9.format(operation_name, connector_name, connector_version)
            logger.error('{0} ERROR: {1}'.format(error_message, str(e)))
            return Response({'message': error_message}, status=status.HTTP_400_BAD_REQUEST)


class ConnectorConfigurationView(ModelViewSet):
    model = Configuration
    queryset = Configuration.objects.all()
    serializer_class = ConnectorConfigurationSerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    ordering_fields = ('config_id',)
    search_fields = ('$name', '$config_id',)
    filterset_class = ConnectorConfigurationFilter
    filter_fields = ('id', 'name', 'config_id', 'connector', 'agent', 'default', 'tags')
    old_data = {}

    def get_queryset(self):
        if 'connector_name' in self.request.GET and 'connector_version' in self.request.GET:
            name = self.request.GET.get('connector_name')
            version = self.request.GET.get('connector_version', 'latest')
            agent_id = self.request.GET.get('agent', settings.SELF_ID)
            version = get_connector_version_or_latest(name, version, None, agent_id)
            connector = Connector.objects.get(name=name, version=version, agent_id=agent_id)
            return Configuration.objects.filter(connector=connector)
        return super(ConnectorConfigurationView, self).get_queryset()

    def get_object(self):
        pk = self.kwargs.get('pk')
        return get_configuration_obj(pk)

    def config_retrieve(self, request, *args, **kwargs):
        name = kwargs.get('connector_name')
        version = kwargs.get('connector_version', 'latest')
        agent_id = self.request.GET.get('agent', settings.SELF_ID)
        version = get_connector_version_or_latest(name, version, None, agent_id)
        connector = Connector.objects.get(name=name, version=version, agent_id=agent_id)
        queryset =  Configuration.objects.filter(connector=connector)
        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def create(self, request, *args, **kwargs):
        if "tags" not in request.data:
            request.data['tags'] = []
        data = request.data
        agent_id = data.get('agent')
        executable_class = get_executable_class(agent_id)
        func = getattr(
            executable_class,
            executable_class.action_mapping.get(REMOTE_CONNECTOR_OPERATIONS.get('config_create'))
        )
        data.update({'tenant': request.tenantid})
        data = func(data, request)
        return Response(data)

    def update(self, request, *args, **kwargs):
        data = request.data
        instance = self.get_object()
        data['agent'] = instance.agent.agentId
        executable_class = get_executable_class(instance.agent.agentId)
        func = getattr(
            executable_class,
            executable_class.action_mapping.get(REMOTE_CONNECTOR_OPERATIONS.get('config_update'))
        )
        data.update({'tenant': request.tenantid})
        data = func(data, request)
        return Response(data)

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        data = {
            "agent": instance.agent.agentId,
            "config_id": instance.config_id
        }
        executable_class = get_executable_class(instance.agent.agentId)
        func = getattr(
            executable_class,
            executable_class.action_mapping.get(REMOTE_CONNECTOR_OPERATIONS.get('config_delete'))
        )
        data.update({'tenant': request.tenantid})
        data = func(data, request)
        return Response(data)


class ProxySettingView(APIView):
    @json_response
    def post(self, request):
        logger.info('Setting Proxy variables in env')
        # configure_proxy(request.data)
        return Response()


class ConnectorImport(APIView):
    parser_classes = (FileUploadParser,)

    @json_response
    def post(self, request, name, version):
        replace_str = request.GET.get('replace', '')
        replace = True if replace_str.lower() == 'true' else False
        # Todo get some secret key from content hub
        # To identify the API and allow system connector to be replace or reinstalled
        trust_api = request.GET.get('trust')
        # data = request.data
        tags = request.GET.get('tags', [])
        agent_id = request.GET.get('agent', settings.SELF_ID)
        rbac_info = get_rbac_info(request)
        file_obj = request.FILES.get('file')
        #Todo raise error if file not found
        data = {
            'name': name, 'version': version, 'agent': agent_id, 'tags': tags,
            'replace': replace, 'trust_api': trust_api, 'rbac_info': rbac_info, 'file': file_obj,
            'tenant': request.tenantid
        }

        # If connector is already in progress of installation or installed
        installable, message = is_connector_installable_on_agent(name, version, agent_id)
        if not installable:
            return Response(
                {"message": message},
                status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )

        executable_class = get_executable_class(agent_id)
        func = getattr(
            executable_class,
            executable_class.action_mapping.get(REMOTE_CONNECTOR_OPERATIONS.get('connector_install'))
        )
        data = func(data, request)
        return Response(data)


class ConnectorAck(APIView):
    @json_response
    def post(self, request, operation):
        data = request.data
        ack_func = ack_operations.get(operation)
        message = ack_func(data)
        return Response({"message": message})


class ConnectorDetail(ModelViewSet):
    queryset = Connector.objects.all()
    serializer_class = ConnectorDetailSerializer
    permission_classes = [IsAgentAuthenticated]

    def get_object(self):
        pk = self.kwargs.get('pk', None)
        if pk is not None:
            return super(ConnectorDetail, self).get_object()
        name = self.kwargs.get('name', None)
        version = self.kwargs.get('version')
        agent_id = self.request.query_params.get('agent', settings.SELF_ID) if hasattr(self,
                                                                                       'request') else settings.SELF_ID
        try:
            version = get_connector_version_or_latest(name, version, None, agent_id)
            connector = Connector.objects.get(name=name, version=version, agent_id=agent_id)
        except Connector.DoesNotExist:
            raise Http404('No matching connector by name %s exists.' % name)
        return connector

    def retrieve_post(self, request, *args, **kwargs):
        return self._retrieve(request, *args, **kwargs)

    def update(self, request, *args, **kwargs):
        instance = self.get_object()
        data = request.data
        data.update({'name': instance.name,
                     'version': instance.version,
                     'agent': instance.agent.agentId})
        _classObj = get_executable_class(instance.agent.agentId)
        return _classObj.connector_detail_update(data)

    def agents(self, request, *args, **kwargs):
        name = self.kwargs.get('name', None)
        installed = True if request.query_params.get('installed', '').lower() == 'true' else False
        team_uuid = []
        query_dict = {}
        active = request.query_params.get('active', '')
        agent_version__gte = request.query_params.get('agent_version__gte', '0.0.0')

        installed_query = Q(version__gte=agent_version__gte) & Q(isLocal=False) & Q(connectors__name=name)
        if active == 'true': 
            installed_query = installed_query & Q(active=True)
        installable_query = (Q(healthStatus__status='connected') & Q(active=True))
        if not installed:
            agent_qs_filter = installed_query | installable_query
        else:
            agent_qs_filter = installed_query
        agent_qs = Agent.objects.prefetch_related('connectors').filter(agent_qs_filter).distinct()

        page = self.paginate_queryset(agent_qs)
        if page is not None:
            serializer = AgentConnectorSerializer(page, connector_name=name, configured=installed, many=True)
            data = serializer.data
            for each_data in data:
                each_data['connector']=each_data.get('connector')
            return self.get_paginated_response(data)

        serializer = AgentConnectorSerializer(agent_qs, connector_name=name, configured=installed, many=True)
        data = serializer.data
        logger.error(data)
        for each_data in data:
            each_data['connector'] = each_data.get('connector')
        return Response(data)

    def retrieve(self, request, *args, **kwargs):
        return self._retrieve(request, *args, **kwargs)

    def _retrieve(self, request, *args, **kwargs):
        try:
            teams = request.data.get('rbac_info', {}).get('teams', [])
            instance = self.get_object()
            serializer = self.get_serializer(instance, teams=teams)
            result = serializer.data
            # Omit password field from configuration
            omit_password(result)
        except Http404 as err:
            logger.error(err)
            return Response({'message': str(err)},
                            status=status.HTTP_404_NOT_FOUND
                            )
        except Exception as exp:
            logger.exception('Retrieve:')
            return Response({'message': str(exp)},
                            status=status.HTTP_400_BAD_REQUEST
                            )
        return Response(result)

    @json_response
    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.remote_status = {'status': REMOTE_STATUS.get('uninstall_in_progress')}
        instance.save()
        agent_id = instance.agent.agentId
        data = {
            'name': instance.name,
            'version': instance.version,
            'agent': agent_id,
            'tenant': request.tenantid
        }
        executable_class = get_executable_class(agent_id)
        func = getattr(
            executable_class,
            executable_class.action_mapping.get(REMOTE_CONNECTOR_OPERATIONS.get('connector_uninstall'))
        )
        data = func(data, request)
        return Response(data)
    def uninstall(self, request, *args, **kwargs):
        data = request.data
        self.kwargs['name'] = data.get('name')
        self.kwargs['version'] = data.get('version')
        return self.destroy(request, *args, **kwargs)


class ConnectorHealth(ModelViewSet):
    def create(self, request, *args, **kwargs):
        logger.info('checking connector health')
        config_id = request.data.get('config_id', 'get_default_config')
        conn_name = request.data.get('connector_name')
        conn_version = request.data.get('connector_version')
        agent_id = request.data.get('agent', settings.SELF_ID)

        conn_obj = Connector.objects.get(name=conn_name, version=conn_version, agent_id=agent_id)
        config = get_configuration(**{'conn_id': conn_obj.id, 'config_id': config_id, 'decrypt': True,
                                      'config_schema': conn_obj.config_schema})

        action = REMOTE_CONNECTOR_OPERATIONS.get('connector_healthcheck')
        data = {
            'connector': conn_name,
            'version': conn_version,
            'configuration': config,
            'agent': agent_id,
            'tenant': request.tenantid
        }
        executable_class = get_executable_class(agent_id)
        func = getattr(
            executable_class,
            executable_class.action_mapping.get(action)
        )
        data = func(data, request)
        return Response(data, status=status.HTTP_200_OK)
    
    @json_response
    def get(self, request, name, version):
        config_id = request.GET.get('config', None)
        if not config_id:
            return Response({"message": xf_integration_33}, status=status.HTTP_400_BAD_REQUEST)
        try:
            config_object = Configuration.objects.get(config_id=config_id, connector__name=name)
        except Configuration.DoesNotExist:
            message = xf_integration_33.format(config_id, name)
            return Response({"message": message}, status=status.HTTP_404_NOT_FOUND)
        return Response(config_object.health_status, status=status.HTTP_200_OK)
        



class ConnectorDependencies(APIView):
    @json_response
    def get(self, request, name, version):
        agent = request.query_params.get('agent', settings.SELF_ID)
        logger.info('checking dependencies installed')
        conn_instance = Connector.objects.filter(name=name, version=version, agent=agent)
        if conn_instance:
            connector_obj = conn_instance.first()
            requirement_installed = identify_if_dependencies_installed(connector_obj.name, connector_obj.version,
                                                                       connector_obj)
            connector_obj.requirements_installed = 'Completed' if requirement_installed else 'Failed'
            connector_obj.save(update_fields=['requirements_installed'])
            return Response({"dependencies_installed": requirement_installed}, status=status.HTTP_200_OK)

    def post(self, request, name, version):
        logger.info('installing dependencies')
        agent = request.query_params.get('agent', settings.SELF_ID)
        cs_propagated = request.query_params.get('cs_propagated', False)
        conn_instance = Connector.objects.filter(name=name, version=version, agent=agent)
        if conn_instance:
            try:
                _install_connector_package(conn_instance.first(), raise_ex=True)
                if not cs_propagated:
                    self_api = request.META['REQUEST_URI']
                    self_api = self_api + '&cs_propagated=True'
                return Response({'message': 'Dependencies installation completed'}, status=status.HTTP_200_OK)
            except Exception as e:
                return Response({"message": f'Failure {str(e)}'}, status=status.HTTP_504_GATEWAY_TIMEOUT)
        return Response({'message': 'No matching connector found'}, status=status.HTTP_404_NOT_FOUND)


class AgentHealth(APIView):
    @json_response
    def get(self, request, agent):
        logger.info('checking agent health')
        if agent and agent != 'all':
            data = {"agent": agent}
            _classObj = get_executable_class(agent)
            return Response(_classObj.agent_health_check(data), status=status.HTTP_200_OK)
        _classObj = RemoteOperations()
        agent_list = Agent.objects.filter(isLocal=False, active=True)
        for agent in agent_list:
            _classObj.agent_health_check({"agent": agent.agent_id})
        return Response({"message": "Agent health-check call initiated."})


class ConnectorExecute(APIView):
    @staticmethod
    def execute_connector_operation(input_data, name=None, version=None, agent_id=None, request=None):
        if not agent_id: agent_id = settings.SELF_ID
        if not (name or version):
            name = input_data.get('connector')
            version = get_connector_version_or_latest(name=name, version=input_data.get('version'), agent_id=agent_id)
        input_data['version'] = version
        operation = input_data.get('operation', '')
        params = input_data.get('params', {})
        env = input_data.get('env', {})
        input_data.pop('env', None)
        config_id = input_data.get('config', 'get_default_config')

        if not config_id:
            config_id = 'get_default_config'
        conn_obj = Connector.objects.get(name=name, version=version, agent_id=agent_id)

        if conn_obj.active:
            connector = get_connector_instance(name, version, conn_obj)

            operation = get_operation(operation, conn_obj.id)
            is_config_required = operation.get('is_config_required', True)
            config = get_configuration(**{'conn_id': conn_obj.id,
                                          'config_id': config_id,
                                          'decrypt': True,
                                          'config_schema': conn_obj.config_schema if is_config_required else {}})

            if config and not config.get('status', 0):
                message = 'The config {0} is partially configured and cannot be used for execution. ' \
                          'Complete and validate the connector configuration ' \
                          'by providing required inputs for the missing fields in the configuration.'

                logger.error(message.format(config.get('name')))
                raise ConnectorError(message.format(config.get('name')))

            params = is_operation_param_valid(operation, params)
            input_data['config'] = config
            input_data['params'] = params
            # change the handle operation params as there will no input validate check

            if input_data.get('audit', False):
                # audit_data contains the info about executed action
                # rbac_info it is been appended by CH for RBAC, contains user_iri and team_iri
                # audit_info it is add in body by UI for info on which record action was executed, contains record_iri
                audit_data = {
                    'connector': conn_obj.label,
                    'action': operation.get('title'),
                    'config': config.get('name'),
                    'version': version,
                    'params': params
                }
                input_data['rbac_info']['data'] = audit_data
                input_data['rbac_info'].update(input_data.get('audit_info'))

            secret_key = input_data.get('secret_key', '')
            if not secret_key == SEALAB_SECRET_KEY:
                try:
                    input_data['params'] = get_parsed_config(input_data.get('params'))
                except Exception as e:
                    pass
            response, is_binary = connector._handle_operation(input_data,
                                                              env=env,
                                                              request=request
                                                              )
            update_env_files(env)
            return response, is_binary
        else:
            raise ConnectorError(xf_integration_4)


    def execute_connector_action(self, input_data, request=None):
        name = input_data.get('connector')
        version = input_data.get('version')
        env = input_data.get('env')
        config_id = input_data.get('config')
        agent_id = input_data.get('agent')
        operation = input_data.get('operation', '')
        config_obj = None
        connector_obj = None

        try:
            if agent_id and config_id:
                # Get connector based on name, version and agent
                # config_obj is none, would be updated by get_configuration method
                connector_obj = get_connector_obj(name=name, version=version, agent_id=agent_id)
                
            elif not agent_id and config_id:
                # Get connector with name and version, Connector.objects.filter(name=name, version=version)
                # Loop through connector with get_connector_instance. In the loop if there are multiple matching configs get_connector_instance throws error and exit loop
                # If there are multiple connector objects found, throw error - Multiple connectors found matching the configuration, please pass uniqe set of configuration for the connector
                # Set connector_obj, config_obj and config_id
                # Set agent_id based on config_id


                # Get connectors with the specified name and version
                connector_list = [{
                    "connector_obj": connector,
                    "config_obj": get_configuration_instance(config_id=config_id, conn_id=connector.id)
                }
                for connector in Connector.objects.filter(name=name)]

                # Filter connectors with valid configuration
                valid_connectors = [c for c in connector_list if c['config_obj']]
                connector_config_count = len(valid_connectors)

                if connector_config_count == 1:
                    connector = valid_connectors[0]
                    connector_obj = connector.get("connector_obj")
                    config_obj = connector.get("config_obj")
                    config_id = config_obj.config_id
                    agent_id = config_obj.agent.agentId
                        
                if connector_config_count > 1:
                    message = xf_integration_49.format(config_id)
                    logger.error(message)
                    raise ConnectorError(message)
                
                if connector_config_count == 0:
                    message = xf_integration_47.format(config_id)
                    logger.error(message)
                    raise ConnectorError(message)

            elif agent_id and not config_id:
                # Get connector based on name, version and agent
                # config_obj is none
                # Set config_id to default

                config_id = '##get_default_config##'
                connector_obj = get_connector_obj(name=name, version=version, agent_id=agent_id)

            elif not agent_id and not config_id:
                # Set agent_id to self
                # Set config_id to default
                # Get connector based on name, version and agent as self
                
                agent_id = settings.SELF_ID
                config_id = '##get_default_config##'
                connector_obj = get_connector_obj(name=name, version=version, agent_id=agent_id)

            if not connector_obj:
                raise ConnectorError(xf_integration_15.format(name))

            if not connector_obj.active:
                raise ConnectorError(xf_integration_17.format(name))

            operation = get_operation(operation, connector_obj.id)
            is_config_required = operation.get('is_config_required', True)

            logger.info('Connector execution input config_id {0} agent_id {1} connector version {2}'.format(
                config_id, agent_id, connector_obj.version))

            # Validate and decrypt config
            get_config_params = {
                    'conn_id': connector_obj.id,
                    'decrypt': True,
                    'config_schema': connector_obj.config_schema if is_config_required else {},
                    'config_instance': config_obj,
                    'config_id': config_id
            }
            config = get_configuration(**get_config_params)
    
            input_data.update({
                "config": config,
                "operation": operation,
                "agent": agent_id,
                "version": connector_obj.version
            })

            # Calling the respective remote or self class
            executable_class = get_executable_class(agent_id)
            func = getattr(
                executable_class,
                executable_class.action_mapping.get(REMOTE_CONNECTOR_OPERATIONS.get('connector_execution'))
            )
            input_data['tenant'] = request.tenantid
            data, is_binary = func(input_data, request)
            return Response(data)
        except ConnectorError as e:
            logger.error(str(e))
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        except MessagePublishFailed as e:
            return Response({'message': repr(e)}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
        except Exception as e:
            logger.exception(str(e))
            return Response({'message': xf_integration_5}, status=status.HTTP_400_BAD_REQUEST)

    @json_response
    def post(self, request):
        input_data = request.data
        secret_key = request.GET.get('secretKey', '') if request.GET.get('secretKey', '') == SEALAB_SECRET_KEY else ''
        input_data.update({"secret_key": secret_key})
        authorization = request.headers.get('forwarded-authorization') if hasattr(request, 'headers') else None
        if authorization:
            env = input_data.get('env', {})
            env.update({'authorization': authorization})
            input_data['env'] = env
        return self.execute_connector_action(input_data, request=request)

class OperationRoleView(APIView):
    def post(self, request, *args, **kwargs):
        data = request.data
        operation_id = kwargs.get("operation_id")
        roles = data.get("roles", [])
        try:
            operation_query_set = Operation.objects.filter(id=operation_id)
            if not operation_query_set.exists():
                raise Exception("Operation not found")
            operation_instance = operation_query_set.first()
            role_ids = get_or_create_team_or_role('role', roles)
            role_data = {'role_ids': role_ids}
            serializer = ConnectorOperationSerializer(operation_instance, role_data, partial=True)
            serializer.is_valid(raise_exception=True)
            serializer.save()
            return Response(serializer.data, status=status.HTTP_200_OK)
        except Exception as e:
            error_message = "Error occurred while creating mapping of operation and roles"
            logger.error("{0} ERROR: {1}".format(error_message, str(e)))
            return Response(error_message, status=status.HTTP_400_BAD_REQUEST)

    def put(self, request, *args, **kwargs):
        data = request.data
        operation_id = kwargs.get('operation_id')
        roles = data.get("roles")
        link_roles = data.get("__link", {}).get("roles", [])
        unlink_roles = data.get("__unlink", {}).get("roles", [])
        role_ids = []
        try:
            operation_query_set = Operation.objects.filter(id=operation_id)
            if not operation_query_set.exists():
                raise Exception("Operation not found")
            operation_instance = operation_query_set.first()
            if not roles is None:
                operation_instance.roles.clear()
                role_ids = get_or_create_team_or_role('role', roles)
            if unlink_roles or link_roles:
                if link_roles:
                    link_roles = get_or_create_team_or_role('role', link_roles)
                if roles is None:
                    role_id_query = operation_instance.roles.all().values('uuid')
                    for role_id in role_id_query:
                        role_ids.append(role_id.get('uuid'))
                role_ids = get_update_list(role_ids, link_roles, unlink_roles)
            role_data = {'role_ids': role_ids}
            serializer = ConnectorOperationSerializer(operation_instance, role_data, partial=True)
            serializer.is_valid(raise_exception=True)
            serializer.save()
            return Response(serializer.data, status=status.HTTP_200_OK)
        except Exception as e:
            error_message = "Error occurred while updating mapping of operation and roles"
            logger.error("{0} ERROR: {1}".format(error_message, str(e)))
            return Response(error_message, status=status.HTTP_400_BAD_REQUEST)


class ConnectorInstall(ModelViewSet):
    queryset = Connector.objects.all()
    serializer_class = ConnectorDetailSerializer

    def install(self, request, *args, **kwargs):
        data = request.data
        replace_str = request.GET.get('replace', '')
        source = request.GET.get('source', '')
        replace = True if replace_str.lower() == 'true' else False
        rbac_info = get_rbac_info(request)
        agent_id = data.get('agent', settings.SELF_ID)
        version = data.get('version')
        name = data.get('name')
        extra_fields = ['certified', 'featured']
        extra_data = {}

        # Check to see if connector is already installed
        already_installed = Connector.objects.filter(name=name, version=version, agent_id=agent_id)
        if already_installed.exists() and already_installed[0].remote_status.get('status') == REMOTE_STATUS.get('finished'):
            if source and source == 'solutionpack':
                data = {'name': name, 'version': version, 'agent': agent_id}
                return Response(data)
            return Response({'message': f'Connector {name} {version} is already installed'}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        # Todo get some secret key from content hub
        # To identify the API and allow system connector to be replace or reinstalled
        trust_api = request.GET.get('trust')
        contenthub_obj = ContentHub.objects.filter(name=name).first()
        for extra_field in extra_fields:
            extra_data[extra_field] = getattr(contenthub_obj, extra_field)
        if not version:
            version = contenthub_obj.version
        if Version(version) < Version(contenthub_obj.version):
            extra_data['latestAvailableVersion'] = contenthub_obj.version
        data = {
            'name': name, 'version': version, 'agent': agent_id, 'tags': data.get('tags', []), 'replace': replace,
            'trust_api': trust_api, 'rbac_info': rbac_info, 'tenant': request.tenantid, 'extra_data': extra_data
        }
        executable_class = get_executable_class(agent_id)
        func = getattr(
            executable_class,
            executable_class.action_mapping.get(REMOTE_CONNECTOR_OPERATIONS.get('connector_install'))
        )
        data = func(data, request)
        return Response(data)


log_dir = os.path.dirname(settings.APP_LOG_FILE_PATH)



class TagsView(ListModelMixin,
               GenericViewSet):
    """
    A viewset for fetching Tags
    """
    serializer_class = TagSerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    ordering_fields = ('name',)
    search_fields = ('$name',)
    filterset_fields = ('name',)

    def get_queryset(self):
        queryset = Tag.objects.all()
        return queryset