import os
import shutil
import logging
import requests
import uuid
import jmespath
import datetime
from packaging.version import Version
from django.conf import settings
from django.db.models import Q
from django_tenants.utils import schema_context
from urllib3.exceptions import InsecureRequestWarning
from connectors.utils import download_connector_tgz, untar_connector_tgz, get_connector_path
from contenthub.models import ContentHub
from contenthub.serializers import ContentHubSerializer
from tenant.models import Tenant
from connectors.models import Connector

content_hub_file = "manifest.json"

logger = logging.getLogger(__name__)
def load_content_hub_data():
    try:
        requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)
        content_hub_json = requests.get(
            url=settings.CONNECTOR_REPO_BASE_URL + content_hub_file,
            timeout=60,
            verify=False
        )
        return content_hub_json.json()
    except Exception as e:
        raise Exception(e)

def download_connector_files(connectors):
    for connector in connectors:
        name = connector.get('name')
        version = connector.get('version')
        if connector.get('update'):
            connector_path = get_connector_path(name, '*')
            shutil.rmtree(connector_path)
        tgz_path = download_connector_tgz(name, version)
        prev_connector_path = untar_connector_tgz(tgz_path, conn_name=name)
        connector_path = get_connector_path(name, version)
        try:
            shutil.rmtree(connector_path)
        except:
            # just a failsafe delete in case the directory existed from a previous failed attempt
            pass
        # rename version to folder.
        shutil.move(prev_connector_path, connector_path)

def get_serialized_data(data):
    serialized_data = {}
    connector_fields = [f.name for f in ContentHub._meta.get_fields()]
    data_keys = data.keys()
    metadata = {}
    for item in data_keys:
        if item in connector_fields:
            serialized_data[item] = data.get(item)
        else:
            metadata[item] = data.get(item)
    if not data.get('metadata'):
        serialized_data['metadata'] = {}
    serialized_data['metadata'].update(metadata)
    return serialized_data

def sync_connectors():
    content_hub_data = load_content_hub_data()
    content_hub_query = ContentHub.objects.all()
    serializer = ContentHubSerializer(content_hub_query, many=True)
    existing_data = serializer.data
    ignore_fields = ['uuid', 'tenants', 'name', 'lastUpdated', 'publishedDate']
    update_repo_fields = ['iconLarge', 'infoPath']
    content_hub_fields = []
    update_data = []
    new_data = []
    retain_name = []
    download_files = []
    updated_version = {}
    for data in content_hub_data:
        try:
            if data.get('type').lower() == 'connector':
                serialized_data = get_serialized_data(data)
                data_name = serialized_data.get('name')
                retain_name.append(data_name)
                existing_details = jmespath.search(
                    "[?name=='" + data_name + "'].{uuid:uuid,name:name,version:version,build_number:buildNumber} | [0]",
                    existing_data)
                data_uuid = existing_details.get('uuid') if existing_details else None
                for repo_field in update_repo_fields:
                    serialized_data[repo_field] = "https://" + settings.PRODUCT_YUM_SERVER + serialized_data[repo_field] if serialized_data.get(
                        repo_field) else serialized_data.get(repo_field)
                if data_uuid:
                    # last_updated_time = datetime.now() - datetime.timedelta(hours=settings.sync_interval)
                    # if True or last_updated_time < data.get('lastupdated', datetime.now() - datetime.timedelta(hours=settings.sync_interval+1)):
                    # ToDo Update only if current date and last run schedule diff is less than last updated date from repo
                    for field in ignore_fields:
                        d = serialized_data.pop(field, None)
                    content_hub_fields = content_hub_fields + list(set(serialized_data.keys()) - set(content_hub_fields))
                    data_serializer = ContentHubSerializer(data=serialized_data, partial=True)
                    data_serializer.is_valid(raise_exception=True)
                    serialized_data['uuid'] = data_uuid
                    serialized_data['name'] = existing_details.get('name')
                    content_hub_obj = ContentHub(**serialized_data)
                    update_data.append(content_hub_obj)
                    if Version(existing_details.get('version')) < Version(serialized_data.get('version')) or int(existing_details.get(
                            'build_number')) < int(serialized_data.get('buildNumber')):
                        updated_version[existing_details.get('name')] = serialized_data.get('version')
                        download_files.append({'name': data_name, 'version': serialized_data.get('version'), 'update': True})
                else:
                    if not serialized_data.get('uuid'):
                        serialized_data['uuid'] = str(uuid.uuid4())
                    data_serializer = ContentHubSerializer(data=serialized_data)
                    data_serializer.is_valid(raise_exception=True)
                    content_hub_obj = ContentHub(**serialized_data)
                    new_data.append(content_hub_obj)
                    download_files.append({'name': data_name, 'version': serialized_data.get('version'), 'update': False})
        except Exception as e:
            logger.error("Error while sync data {0} Error: {1}".format(data.get('name'), str(e)))

    delete_qs = ContentHub.objects.filter(~Q(name__in=retain_name))
    delete_qs.delete()
    if update_data:
        updated_content_hub = ContentHub.objects.bulk_update(update_data, fields=content_hub_fields)
    if new_data:
        content_hub_fields.extend(ignore_fields)
        created_content_hub = ContentHub.objects.bulk_create(new_data)

    # update all installed connector for update available for a connector
    tenants = Tenant.objects.all().values('tenant_id', 'schema_name')
    connector_names = updated_version.keys()
    for tenant in tenants:
        with schema_context(tenant.get('schema_name')):
            all_connectors = Connector.objects.filter(name__in=connector_names)
            for connector in all_connectors:
                if Version(connector.version) < Version(updated_version[connector.name]):
                    connector.latestAvailableVersion = updated_version[connector.name]
            Connector.objects.bulk_update(all_connectors, fields=['latestAvailableVersion'])
    
    # Uncomment for dev setup to download files of connector
    # process = Thread(target=download_connector_files, args=[download_files])
    # process.start()