""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import json
from functools import reduce

import arrow
from django import forms
from url_filter.filters import CallableFilter, form_field_for_filter


class DatetimeFilter(CallableFilter):
    """
    Improved datetime filter. This builds on the functionality of the url-filter
    datetime filter, adding the ability to search on formatted datetime strings.
    filters.

    By datetime piece:

        * ?field__year=2016
        * ?field__month=4
        * ?field__day=13
        * ?field__hour=6
        * ?field__minute=12
        * ?field__week_day=3

        Note that day of week is the iso version, i.e. it starts with 1 ==
        Sunday and ends with 7 == Saturday.

    By datetime string:

        * ?field=2009-4-13T16:20:01.058008Z -> filter exact
        * ?field__gt=2009-4-13T16:20:01.058008Z -> filter greater than
        * ?field__gte=2009-4-13T16:20:01.058008Z -> filter greater than or\
        equal to
        * ?field__lt=2009-4-13T16:20:01.058008Z -> filter less than
        * ?field__lte=2009-4-13T16:20:01.058008Z -> filter less than or equal\
        to
        * ?field__range=2009-4-13T16:20:01.058008Z,2011-4-13T16:20:01.058008Z \
        -> filter range

    By piece of datetime string:

        * ?field__year=2009-4-13T16:20:01.058008Z == ?field__year=2009
        * ?field__year=2009-4-13T16:20:01.058008Z == ?field__day=13
        * ?field__year=2009-4-13T16:20:01.058008Z == ?field__month=4
        * etc.

    All these filters are convered into the corresponding django datetime
    filter. More info here: \
    https://docs.djangoproject.com/en/1.9/ref/models/querysets/#range
    """

    @form_field_for_filter(forms.CharField())
    def filter_exact_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec, lambda dt: dt.datetime
        )

    # One could probably do some metaprogramming to avoid this repetition but
    # meh. also, dt == datetime which is expected to be an arrow object
    @form_field_for_filter(forms.CharField())
    def filter_gt_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec, lambda dt: dt.datetime
        )

    @form_field_for_filter(forms.CharField())
    def filter_lt_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec, lambda dt: dt.datetime
        )

    @form_field_for_filter(forms.CharField())
    def filter_gte_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec, lambda dt: dt.datetime
        )

    @form_field_for_filter(forms.CharField())
    def filter_lte_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec, lambda dt: dt.datetime
        )

    @form_field_for_filter(forms.CharField())
    def filter_year_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec,
            lambda dt: spec.value if spec.value.isdigit() else dt.year
        )

    @form_field_for_filter(forms.CharField())
    def filter_month_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec,
            lambda dt: spec.value if spec.value.isdigit() else dt.month
        )

    @form_field_for_filter(forms.CharField())
    def filter_day_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec,
            lambda dt: spec.value if spec.value.isdigit() else dt.day
        )

    @form_field_for_filter(forms.CharField())
    def filter_hour_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec,
            lambda dt: spec.value if spec.value.isdigit() else dt.hour
        )

    @form_field_for_filter(forms.CharField())
    def filter_minute_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec,
            lambda dt: spec.value if spec.value.isdigit() else dt.minute
        )

    @form_field_for_filter(forms.CharField())
    def filter_week_day_for_django(self, queryset, spec):
        return self._filter_generic_for_django(
            queryset, spec,
            lambda dt: spec.value if spec.value.isdigit() else dt.isoweekday()
        )

    @form_field_for_filter(forms.CharField())
    def filter_range_for_django(self, queryset, spec):
        # given the filter
        # created__range=2016-07-21T20:28:20.858347Z,2016-07-21T20:29:20.839085Z
        # you will get a tuple like this to filter on:
        # (datetime.datetime(2016, 7, 21, 20, 28, 20, 858347, tzinfo=tzutc()),
        # datetime.datetime(2016, 7, 21, 20, 29, 20, 839085, tzinfo=tzutc()))
        try:
            pieces = spec.value.split(',', 1)

            start = arrow.get(pieces[0])
            end = arrow.get(pieces[1])
        except Exception:
            return queryset

        value = (start.datetime, end.datetime)

        return self._filter_generic_for_django(
            queryset, spec, lambda dt: value
        )

    def _filter_generic_for_django(self, queryset, spec, datetime_resolver):
        """
        datetime_resolver is a function that takes an arrow object, and returns
        an appropriate django filter field. Some filters take datetime objects
        -- such as date, gt, lt, etc --, and some take ints -- such as day,
        month, year, etc.
        """
        # arrow is pretty forgiving when parsing, malformed dates will cause a
        # parse error, but ambiguous ones usually resolve to the unix epoch
        # this is ok, because many of the lambdas in the above filters will
        # forgo this parsing and return int values to filter on
        try:
            parsed_time = arrow.get(spec.value)
        except arrow.parser.ParserError:
            return queryset.none()

        # put the components back together for chained relation filters
        # ['workflow', 'created'] becomes 'workflow__created'
        field = reduce(lambda first, second: first + '__' + second,
                       spec.components)
        field_lookup = field + '__' + spec.lookup

        # get either a native python datetime object or an int to filter on
        value = datetime_resolver(parsed_time)

        # actual filtering
        f = queryset.filter if not spec.is_negated else queryset.exclude
        return f(**{field_lookup: value})


class JSONFilter(CallableFilter):
    """
    Filters on the JSONFields of a model. The value of a field is a 'query
    object' that describes how to construct the filter. The query object is a
    simple key/value json string where the key is the under__score 'path' to the
    field you want to check and the value is the value to compare.

    Consider this json::

        {
            "some": {
                "nested": "value"
            }
        }

    If you make a query that looks like this::

        /model/?field={"some__nested": "value"}

    It will be turned into this query object::

        {
            "field__some__nested": "value"
        }

    Which in turn will be transformed into a queryset filter::

        queryset.filter(field__some__nested="value")

    More about json filtering in django:
    https://docs.djangoproject.com/en/1.9/ref/\
contrib/postgres/fields/#querying-jsonfield
    """

    @form_field_for_filter(forms.CharField())
    def filter_exact_for_django(self, queryset, spec):
        try:
            query_obj = json.loads(spec.value)
        except ValueError:
            return queryset

        try:
            filter_dict = {
                # e.g.: args__nested__key: value
                spec.components[-1] + '__' + key: value
                for key, value in query_obj.items()
            }
        except AttributeError:
            # raised if the query obj isn't actually a dict, such as a string --
            # which is still valid json -- when .items() is called
            return queryset

        # pretty sure that exclude doesn't work with json filters, but maybe
        # they will add it someday and this will just work. Currently it will
        # return an empty queryset instead of the negated queryset
        f = queryset.filter if not spec.is_negated else queryset.exclude
        return f(**filter_dict)

    @form_field_for_filter(forms.CharField())
    def filter_contains_for_django(self, queryset, spec):
        # this is mostly the same as above, but tacks on the extra __contains
        # field. Possible TODO: consolidate more of this logic, as long as it
        # doesn't become too confusing (explicit duplication is better than DRY
        # spaghetti)
        try:
            query_obj = json.loads(spec.value)
        except ValueError:
            return queryset

        try:
            filter_dict = {
                spec.components[-1] + '__' + key + '__contains': value
                for key, value in query_obj.items()
            }
        except AttributeError:
            return queryset
        return queryset.filter(**filter_dict)
