import logging

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import viewsets, status
from rest_framework.response import Response

from xfutils.query.filtersets import query_filters, query_aggregates, query_sort

from agents.serializers import AgentQuerySerializer
from connectors.serializers import ConnectorConfigurationQuerySerializer

logger = logging.getLogger(__name__)

model_map = {
    'agents': 'agents.Agent',
    'connector_configuration': 'connectors.Configuration'
}

serializer_map = {
    'agents': AgentQuerySerializer,
    'connector_configuration': ConnectorConfigurationQuerySerializer
}


class QueryViewSet(viewsets.ReadOnlyModelViewSet):
    def get_queryset(self, *args, **kwargs):
        model = self.kwargs.get('model')
        model = model_map.get(model, None)
        if model:
            model = django_apps.get_model(model, require_ready=False)
            return model.objects.all()
        else:
            raise ObjectDoesNotExist(
                f'The provided model name is not queryable. Provided modle name should be one of: {", ".join(model_map.keys())}')

    def get_serializer_class(self):
        model = self.kwargs.get('model', None)
        return serializer_map.get(model)

    def query(self, request, model, *args, **kwargs):
        """
        This function is accessed with a POST request to /api/wf/{model_name}
            eg. /api/wf/workflows, /api/wf/steps, etc
        It expects the following payload format:
        {
            "logic": {"and", "or"},
            "filters": [
                {
                "operator": {'contains', 'lt', 'gt', 'lte', 'gte', 'isnull', 'in', 'eq', 'neq', 'ncontains'},
                "field": "",
                "value": ""
                },
                ...
            ],
            "aggregates": [],
            "sort": "",
            "limit": 10,
            "fields": []
        }
        """
        try:
            queryset = self.get_queryset()
        except ObjectDoesNotExist as e:
            error_message = f"Error performing query ERROR:: {str(e)}"
            return Response({"message": error_message}, status=status.HTTP_404_NOT_FOUND)

        query = request.data
        if query.get('filters'):
            queryset = query_filters(queryset, query.get('filters'), query.get('logic', 'and'))

        # Commenting this line since Teams are not implemented yet
        # queryset = query_filters(queryset, [{"field": "teams__uuid", "value": request.teams, "operator": "in"}], "and")

        # Commenting this line since SubTenants are not implemented yet
        # queryset = query_filters(queryset, [{"field": "subTenant__id", "operator": "eq", "value": request.subtenantid}], "and")

        if query.get('aggregates'):
            queryset = query_aggregates(queryset, query)

        if query.get('sort'):
            queryset = query_sort(queryset, query)

        if query.get('limit'):
            queryset = queryset[:int(query.get('limit'))]

        page = self.paginate_queryset(queryset)
        if page is not None:
            if query.get('aggregates'):
                return Response({'data': page})
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)
