import glob
import json
import logging
import subprocess
import datetime

from django.conf import settings
from django_tenants.utils import schema_context
from connectors.core.connector import ConnectorError
from connectors.core.constants import REMOTE_CONNECTOR_OPERATIONS
from constants.error_message import *
from agents.models import Agent
from contenthub.models import ContentHub
from utils.connector_op_utility import get_executable_class

logger = logging.getLogger('postman')

def create_self_agent(schema_name, tenant_id):
    with schema_context(schema_name):
        Agent.objects.get_or_create(agentId=tenant_id, defaults={'name': 'Self Agent', 'isLocal': True, 'healthStatus': {'status': 'connected'}})

def load_fixtures(schema_name):
    try:
        fixture_apps = []
        if fixture_apps:
            cmd_load_fixtures = ['python', 'manage.py', 'tenant_command', 'loaddata']
            for each_app in fixture_apps:
                # Use glob to get a list of all JSON fixtures in the app directory
                cmd_load_fixtures.extend(glob.glob("{0}/fixtures/*.json".format(each_app)))

            cmd_load_fixtures.append('--schema=' + schema_name)
            result = subprocess.run(cmd_load_fixtures, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
            if result.returncode != 0:
                logger.error("Error while applying data fixture for {0} ERROR:: {1}".format(fixture_apps, result.stderr))
    except Exception as e:
        msg="Error while applying data fixture for ERROR:: {0}".format(str(e))
        logger.error(msg)
        raise Exception(msg)


def install_system_connectors(schema_name, tenant_id):
    try:
        system_connectors = ContentHub.objects.filter(name__in=settings.CONNECTORS_RESERVED)
        with schema_context(schema_name):
            for connector in system_connectors:
                try:
                    config = settings.ALL_CONFIG.get('default_connectors_config', connector.name, dict())
                    if isinstance(config, str):
                        config = json.loads(config)
                    data = {
                        'name': connector.name, 'version': connector.version, 'agent': settings.SELF_ID,
                        'replace': True, 'trust_api': True, 'config': config, 'tenant': tenant_id
                    }
                    executable_class = get_executable_class(settings.SELF_ID)
                    func = getattr(
                        executable_class,
                        executable_class.action_mapping.get(REMOTE_CONNECTOR_OPERATIONS.get('connector_install'))
                    )
                    data = func(data)
                    logger.info('Successfully installed connector {0}'.format(connector.name))
                except Exception as e:
                    logger.exception('Error while installing system connector {0} ERROR:: {1}'.format(connector.name, str(e)))
    except Exception as e:
        message = 'Error while installing system connectors ERROR:: {0}'.format(str(e))
        logger.exception(message)
        raise Exception(message)