""" Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
from collections import OrderedDict
from urllib.parse import urlparse, urlunparse

from rest_framework.pagination import (
    _divide_with_ceil,
    LimitOffsetPagination
)
from rest_framework.response import Response
from rest_framework.utils.urls import (
    remove_query_param,
    replace_query_param
)


class HydraLimitOffsetPagination(LimitOffsetPagination):
    """
    Basic limit offset paging, but with hydra keywords added.
    """

    # Override super's get_next_link and get_previous_link to build based off
    # the relative (rather than absolute) URL
    def get_next_link(self):
        """
        Build link for next set of results
        """
        if self.offset + self.limit >= self.count:
            return None

        url = self.request.get_full_path()
        url = replace_query_param(url, self.limit_query_param, self.limit)

        offset = self.offset + self.limit
        return replace_query_param(url, self.offset_query_param, offset)

    def get_previous_link(self):
        """
        Build link for previous set of results
        """
        if self.offset <= 0:
            return None

        url = self.request.get_full_path()
        url = replace_query_param(url, self.limit_query_param, self.limit)

        if self.offset - self.limit <= 0:
            return remove_query_param(url, self.offset_query_param)

        offset = self.offset - self.limit
        return replace_query_param(url, self.offset_query_param, offset)

    def get_first_link(self):
        """
        Build link for first page
        """
        url = self.request.get_full_path()
        return remove_query_param(url, self.offset_query_param)

    def get_last_link(self):
        """
        Build link for last page
        """
        url = self.request.get_full_path()

        limit = self.get_limit(self.request)
        totalPages = _divide_with_ceil(self.count, limit)
        last_offset = (totalPages - 1) * limit

        return replace_query_param(url, self.offset_query_param, last_offset)

    def get_paginated_response(self, data):
        """
        Generates all the proper hydra pagination keywords and also will add the
        @context iri. You did add the mixin your viewsets to support this right?
        """
        # FIXME: @context should really be added somewhere else, there isn't
        # enough, uh, context, in this function to decided whether or not to add
        # it for any individual model

        # append context/ to whatever the current route is, and clear out any
        # query params
        url = self.request.get_full_path()
        (scheme, netloc, path, params, query, fragment) = urlparse(url)
        path += 'context/'
        query = {}
        context_url = urlunparse(
            (scheme, netloc, path, params, query, fragment)
        )

        response_page = OrderedDict([
            ('@context', context_url),
            ('hydra:nextPage', self.get_next_link()),
            ('hydra:previousPage', self.get_previous_link()),
            ('hydra:totalItems', self.count),
            ('hydra:itemsPerPage', self.get_limit(self.request)),
            ('hydra:firstPage', self.get_first_link()),
            ('hydra:lastPage', self.get_last_link()),
            ('hydra:member', data),
        ])
        return Response(response_page)

    def get_results(self, data):
        return data['hydra:member']
