import json
import logging
import uuid
from django.db.utils import OperationalError
from django.conf import settings
from rest_framework import authentication
from rest_framework.exceptions import AuthenticationFailed

from agents.models import Agent
from agents.serializers import AgentSerializer
from connectors.core.connector import ConnectorError
from constants.error_message import *
from identity.models import User, Team
from xfutils.utils import api_call

logger = logging.getLogger(__name__)


def init_rbac(request, response):
    # Todo go through while implementing RBAC
    from connectors.utils import get_or_create_team_or_role
    content_type = request.headers.get('Content-Type', '')
    is_content_type_json = 'application/json' in content_type
    roles = response.get('roles', [])
    teams = response.get('teams', [])

    if request.method not in ['GET'] and is_content_type_json:
        # Taking the rbac info if already in request
        rbac_info = json.loads(request.body.decode()).get('rbac_info', {})
        if not rbac_info:
            rbac_info = {'teams': [], 'roles': [], 'secret_key': ''}

    try:
        # Create the Roles if not exist
        role_ids = get_or_create_team_or_role('role', roles)

        # Create the Teams if not exist
        team_ids = get_or_create_team_or_role('team', teams)

        # Set the roles and teams in request body for RBAC
        if request.method not in ['GET'] and is_content_type_json:
            prev_roles = rbac_info.get('roles', [])
            prev_teams = rbac_info.get('teams', [])
            rbac_info['roles'] = prev_roles.extend(role_ids)
            rbac_info['teams'] = prev_teams.extend(team_ids)
            request.body.__add__(json.dumps({'rbac_info': rbac_info}).encode())

    except ConnectorError as e:
        raise ConnectorError(e)
    except Exception as e:
        logger.exception(xf_integration_24.format(str(e)))
        raise ConnectorError(xf_integration_24.format(str(e)))


def verify_token(request):
    token = request.headers.get('Authorization')
    if not token:
        raise AuthenticationFailed(detail='No tenant ID found in the request')

    url = 'https://' + settings.AUTH_HOST + '/api/auth/token/verify/'
    body = {'token': token[7:]}
    headers = {"Authorization": token}
    response = api_call(url, method='POST', body=body, headers=headers, verify=False)
    return response


def get_or_create_user(data):
    user_created = False
    user, user_created = User.objects.get_or_create(
        username=data.get('username'), first_name=data.get('firstname', ''),
        last_name=data.get('lastname', ''), uuid=data.get('user_uuid', uuid.uuid4())
    )
    for team in data.get('teams', []):
        team, team_created = Team.objects.get_or_create(uuid=team)
        if user_created or team_created:
            team.members.add(user)
    return user, {
        'auth_method': 'JWT'
    }

def initialize(request, response):
    request.tenantid = response.get('tenantid')
    request.tenantuuid = response.get('tenantuuid')
    request.subtenantid = response.get('suborgid')
    request.useruuid = response.get('user_uuid')
    request.username = response.get('username')
    request.teams = response.get('teams')
    request.role = response.get('role')
    request.policy = response.get('policy')
    request.META['tenantid'] = response.get('tenantid')
    request.headers.__setattr__('tenantid', response.get('tenantid')) # ToDo: implementation for sub org in this flow
    settings.SELF_ID = request.tenantid

class JWTAuthentication(authentication.BaseAuthentication):

    def authenticate(self, request):
        secretKey = request.GET.get('secretKey')
        if secretKey and secretKey == settings.SEALAB_SECRET_KEY:
            settings.SELF_ID = request.headers.get('TenantId') or request.tenant.tenantid
            request.tenantid = request.headers.get('TenantId') or request.tenant.tenantid
            return get_or_create_user({"username": settings.HMAC_USER})
        try:
            response = verify_token(request)
            # Public API dont have token Ex: swagger:docs
            if not getattr(request, 'public', False):
                user = get_or_create_user(response)
            else:
                user = get_or_create_user({"username": settings.HMAC_USER})
            initialize(request, response)
            return user
        #
        # Need to override this, as we don't want to expose internal details like IP address, port, username etc.
        # django.db.utils.OperationalError: connection to server at "db" (172.21.0.5), port 5432 failed: FATAL:  no pg_hba.conf entry for host "172.21.0.18", user "cyberpgsql", database "integration", no encryption
        #
        except OperationalError as e:
            logger.exception('Error while authentication %s' % str(e))
            raise AuthenticationFailed(detail="Unable to connect to the database")
        except Exception as e:
            logger.exception('Error while authentication %s' % str(e))
            raise AuthenticationFailed(detail=str(e))
